<?php
/**
 * Navigation widget, handles displaying the navigation menus.
 *
 * @package Navigation
 * @since 0.1
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\View;
use Core\Widget;
use Core\Model;
use Core\Router;


class NavigationWidget extends Widget {
	
	public $displaySettings = [
		'image_dimensions' => [
			'type' => Model::ATT_TYPE_STRING,
			'value' => '100x75',
			'title' => 't:STRING_NAVIGATION_WIDGET_IMAGE_DIMENSIONS',
			'description' => 't:MESSAGE_NAVIGATION_WIDGET_IMAGE_DIMENSIONS',
		],
	];
	
	public function view() {
		$view       = $this->getView();
		$m          = NavigationModel::Construct($this->getParameter(0));
		$current    = Router::$DefaultRequest;
		
		// Used to indicate the "active" link.
		$currentBase = Router::$DefaultBaseURL->getBaseURL();
		$currentRewrite = Router::$DefaultBaseURL->getRewriteURL();
		
		if (!$m->checkExists()) {
			return View::STATUS_NOTFOUND;
		}

		// Get the entries for this model as well.
		$entries = $m->getLink('NavigationEntry', 'weight ASC');
		
		// Map the display settings to a simple array for the template to use.
		$displaySettings = [];
		foreach($this->displaySettings as $idx => $val){
			if(isset($val['name'])){
				$displaySettings[ $val['name'] ] = $val['value'];
			}
			else{
				$displaySettings[$idx] = $val['value'];
			}
		}

		// View won't quite just have a flat list of entries, as they need to be checked and sorted
		// into a nested array.
		$sortedentries = [];
		// First level children
		foreach ($entries as $k => $e) {
			if (!$e->get('parentid')) {

				$classes = [];
				$classes[] = Core\str_to_url($e->get('title')) . '-link';
				if($e->get('baseurl') === $currentBase || $e->get('baseurl') === $currentRewrite){
					$classes[] = 'active';
				}
				

				if(\Core\user()->checkAccess($e->getAccessString())){
					// There's a weird bug where sometimes the access cache is empty.
					// In that case, just allow the user to view the page.
					$sortedentries[] = [
						'obj' => $e,
						'children' => [],
						'classes' => $classes
					];
				}

				unset($entries[$k]);
			}
		}

		// One level deep
		if (sizeof($entries)) {
			foreach ($sortedentries as $sk => $se) {
				foreach ($entries as $k => $e) {
					if ($e->get('parentid') == $se['obj']->get('id')) {

						if(\Core\user()->checkAccess($e->getAccessString())){
							// There's a weird bug where sometimes the access cache is empty.
							// In that case, just allow the user to view the page.
							$classes   = [];
							$classes[] = Core\str_to_url($e->get('title')) . '-link';
							if($e->get('baseurl') == $currentBase) {
								$classes[] = 'active';

								// also set active class on the parent so frontenders don't rage :)
								$sortedentries[ $sk ]['classes'][] = 'active';
							}

							// Add the "more" class to the parent.
							$sortedentries[ $sk ]['classes'][] = 'more';

							$sortedentries[ $sk ]['children'][] = [
								'obj'      => $e,
								'children' => [],
								'classes'  => $classes
							];
						}
						unset($entries[ $k ]);
					}
				}
			}
		}

		// Two levels deep
		// this would be so much simpler if the menu was in DOM format... :/
		if (sizeof($entries)) {
			foreach ($sortedentries as $sk => $se) {
				foreach ($se['children'] as $subsk => $subse) {
					foreach ($entries as $k => $e) {
						if ($e->get('parentid') == $subse['obj']->get('id')) {

							if(\Core\user()->checkAccess($e->getAccessString())){
								$classes = [];
								$classes[] = Core\str_to_url($e->get('title')) . '-link';
								if($e->get('baseurl') == $currentBase) {
									$classes[] = 'active';

									// also set active class on the top-most nav parent so frontenders don't rage :)
									$sortedentries[$sk]['children'][$subsk]['class'][] =  'active';
								}

								// Add the "more" class to the parent.
								$sortedentries[$sk]['children'][$subsk]['class'][] = 'more';

								$sortedentries[$sk]['children'][$subsk]['children'][] = [
									'obj' => $e,
									'children' => [],
									'classes' => $classes
								];
							}
							unset($entries[$k]);
						}
					}
				}
			}
		}

		foreach($sortedentries as $k => $el){
			$this->_transposeClass($sortedentries[$k]);
		}

		$view->title        = $m->get('title');
		$view->access       = $m->get('access');
		//$view->templatename = '/widgets/navigation/view.tpl';
		$view->assign('model', $m);
		$view->assign('entries', $sortedentries);
		$view->assign('display_settings', $displaySettings);
	}

	/**
	 * Get the path for the preview image for this widget.
	 *
	 * Should be an image of size 210x70, 210x140, or 210x210.
	 *
	 * @return string
	 */
	public function getPreviewImage(){
		return 'assets/images/previews/templates/widgets/navigation/navigation.png';
	}



	private function _transposeClass(&$el){
		$el['class'] = implode(' ', array_unique($el['classes']));
		if(sizeof($el['children'])){
			foreach($el['children'] as $k => $subel){
				$this->_transposeClass($el['children'][$k]);
			}
		}
	}
}
