"use strict";

let CoreMarkdownEditor = class {
	constructor(element) {
		this.textarea = element;
		this.textarea.editorLoaded = true;
		this.controls = element.closest('.form-element-value').querySelector('.button-toolbar');
		this.fullscreen = false;
		this.preview = element.closest('.form-element-value').querySelector('.markdown-preview-window');

		this.controls.querySelectorAll('.run-action').forEach(el => {
			el.addEventListener('click', evt => {
				this.runAction(evt.target.closest('.run-action').dataset.action);

				this.updateSelectedAction();
				this.updatePreview();
				evt.preventDefault();
			});
		});

		this.controls.querySelectorAll('.run-help').forEach(el => {
			el.addEventListener('click', evt => {
				alert('@TODO');
				evt.preventDefault();
			});
		});
		
		this.textarea.addEventListener('keyup', evt => {
			if(evt.key === 'Enter') {
				// Check if the previous line was something to be continued
				let lastLine = this.getPreviousLine();

				if(lastLine === '* ') {
					this.replacePreviousLine('');
				}
				else if(lastLine.startsWith('* ')) {
					this.simpleInsert('* %');
				}
				else if(lastLine.startsWith('> ')) {
					this.simpleInsert('> %');
				}
				else if(lastLine.match(/^[0-9]*\. $/) !== null) {
					this.replacePreviousLine('');
				}
				else if(lastLine.match(/^[0-9]*\. /) !== null) {
					let lineCounter = lastLine.match(/^([0-9]*)\. /);
					lineCounter = parseInt(lineCounter[1]) + 1;

					this.simpleInsert(lineCounter.toString() + '. %');
				}
			}

			this.updateSelectedAction();
			this.updatePreview();
		});

		this.textarea.addEventListener('click', evt => {
			this.updateSelectedAction();
			this.updatePreview();
		});
	}

	hasSelection() {
		return this.textarea.selectionStart !== this.textarea.selectionEnd;
	}

	atStartOfLine() {
		let previousCharacter = this.getPreviousCharacter();

		return previousCharacter === "\n" || previousCharacter === null;
	}

	/**
	 * Get the character position of the start of the current line
	 *
	 * Returns the position of the newline character
	 *
	 * @returns {number}
	 */
	getCursorPositionStartOfLine() {
		if(this.atStartOfLine()) {
			return this.textarea.selectionStart;
		}

		let pos = this.textarea.value.lastIndexOf('\n', this.textarea.selectionStart - 1);

		return pos === -1 ? 0 : pos + 1;
	}

	/**
	 * Get the character position of the end of the current line
	 *
	 * Returns the position of the newline character
	 *
	 * @returns {number}
	 */
	getCursorPositionEndOfLine() {
		let pos = this.textarea.value.indexOf('\n', this.textarea.selectionStart);

		return pos === -1 ? this.textarea.value.length : pos;
	}

	/**
	 * Get the full previous line from where the cursor is located
	 *
	 * @returns {null|string}
	 */
	getPreviousLine() {
		let pos = this.getCursorPositionStartOfLine(),
			text;

		if(pos === 0) {
			// No previous line
			return null;
		}

		text = this.textarea.value.substr(0, pos - 1).split("\n");
		return text[ text.length - 1 ];
	}

	/**
	 * Get the full current line from where the cursor is located
	 *
	 * @returns {string}
	 */
	getCurrentLine() {
		let pos1 = this.getCursorPositionStartOfLine(),
			pos2 = this.getCursorPositionEndOfLine();

		return this.textarea.value.substr(pos1, (pos2 - pos1));
	}

	/**
	 * Get the previous character behind the cursor position
	 *
	 * @returns {null|string}
	 */
	getPreviousCharacter() {
		if(this.textarea.selectionStart === 0) {
			return null;
		}
		else {
			return this.textarea.value.substr(this.textarea.selectionStart - 1, 1);
		}
	}

	/**
	 * Get the z-index of the editor, useful for being loaded in dialogs.
	 *
	 * @returns number
	 */
	getZIndex() {
		let element = this.textarea,
			zIndex = 0,
			checkValue;

		while(element !== document.body) {

			checkValue = window.document.defaultView.getComputedStyle(element).zIndex;

			if(checkValue && checkValue !== 'auto') {
				zIndex = Math.max(zIndex, parseInt(checkValue));
			}

			element = element.parentElement;
		}

		return zIndex;
	}

	/**
	 * Replace the previous line of text with the provided
	 *
	 * @param text
	 */
	replacePreviousLine(text) {
		let previousText = this.textarea.value.substr(0, this.textarea.selectionStart - 1).split("\n"),
			aheadText = this.textarea.value.substr(this.textarea.selectionStart).split("\n"),
			cpos = this.textarea.selectionStart;

		// Update the cursor position first since we have both data available
		cpos += text.length - previousText[previousText.length - 1].length;

		// Swap the previous line text
		previousText[previousText.length - 1] = text;

		// Write the new merged data to the field and keep the position.
		previousText = previousText.concat(aheadText);
		this.textarea.value = previousText.join("\n") + "\n";

		this.textarea.focus();
		this.textarea.selectionStart = cpos;
		this.textarea.selectionEnd = cpos;
	}


	/**
	 * Replace the current line of text with the provided
	 *
	 * @param text
	 */
	replaceCurrentLine(text) {
		let previousEndPosition = this.getCursorPositionStartOfLine(),
			followingStartPosition = this.getCursorPositionEndOfLine(),
			previousText = this.textarea.value.substr(0, previousEndPosition),
			followingText = this.textarea.value.substr(followingStartPosition),
			currentText = this.getCurrentLine(),
			cpos = this.textarea.selectionStart;

		// Update the cursor position first since we have both data available
		cpos += text.length - currentText.length;

		// Write the new merged data to the field and keep the position.
		this.textarea.value = previousText + text + followingText;

		this.textarea.focus();
		this.textarea.selectionStart = cpos;
		this.textarea.selectionEnd = cpos;
	}

	runAction(action) {
		switch(action) {
			case 'bold':
				this.simpleInsert('**%**');
				break;
			case 'italic':
				this.simpleInsert('_%_');
				break;
			case 'strikethrough':
				this.simpleInsert('~~%~~');
				break;
			case 'ul':
				this.simpleBlockInsert("* %");
				break;
			case 'ol':
				this.simpleBlockInsert("1. %");
				break;
			case 'quote':
				this.simpleBlockInsert("> %");
				break;
			case 'pre':
				this.simpleBlockInsert('```\n%\n```');
				break;
			case 'h1':
				this.doHeading(1);
				break;
			case 'h2':
				this.doHeading(2);
				break;
			case 'h3':
				this.doHeading(3);
				break;
			case 'h4':
				this.doHeading(4);
				break;
			case 'h5':
				this.doHeading(5);
				break;
			case 'hr':
				this.simpleBlockInsert('---\n\n%');
				break;
			case 'link':
				this.doLink();
				break;
			case 'image':
				this.doImage();
				break
			case 'video':
				this.doVideo();
				break;
			case 'fullscreen':
				this.doFullscreen();
				break;
		}
	}

	doImage() {
		let modal = new Core.ModalDialog();

		modal.zIndex = this.getZIndex() + 100;

		modal.pageLoaded = () => {
			window.FileNavigator.openFile = (file) => {

				this.simpleInsert('![%](' + file.dataset.url + ')^');
				this.updatePreview();
				modal.close();
			}
		}
		modal.ajaxLoad(Core.ROOT_URL + 'mediamanagernavigator?ajax=1&navigators=media,public,assets&mime=image/*');
	}

	doVideo() {
		let modal = new Core.ModalDialog();

		modal.zIndex = this.getZIndex() + 100;

		modal.pageLoaded = () => {
			window.FileNavigator.openFile = (file) => {

				let html = '';

				html += '<video';
				if(file.dataset.width) {
					html += ' width="' + file.dataset.width + '"';
				}
				if(file.dataset.height) {
					html += ' height="' + file.dataset.height + '"';
				}
				html += ' controls>\n'; // complete video tag start

				html += '\t<source src="' + file.dataset.url + '" type="' + file.dataset.mimetype + '"/>\n'; // Source

				html += '</video>'; // Closing video tag

				this.simpleBlockInsert(html + '^');
				this.updatePreview();
				modal.close();
			}
		}
		modal.ajaxLoad(Core.ROOT_URL + 'mediamanagernavigator?ajax=1&navigators=media,public,assets&mime=video/*');
	}

	doLink() {
		let url = prompt("Please enter the URL to link, starting with https://");

		if(url === null) {
			// User cancelled out
			return;
		}

		if(this.hasSelection()) {
			this.simpleInsert('[%](' + url + ')^');
			this.updatePreview();
		}
		else {
			this.simpleInsert('[Click Here](' + url + ')^');
			this.updatePreview();
		}
	}

	doHeading(level) {
		let prefix = '#'.repeat(level),
			currentLine = this.getCurrentLine(),
			trim = false;

		if(currentLine.startsWith(prefix + ' ')) {
			// Current line already matches the requested heading level.
			return;
		}
		else if(currentLine.startsWith('# ')) {
			trim = 1;
		}
		else if(currentLine.startsWith('## ')) {
			trim = 2;
		}
		else if(currentLine.startsWith('### ')) {
			trim = 3;
		}
		else if(currentLine.startsWith('#### ')) {
			trim = 4;
		}
		else if(currentLine.startsWith('##### ')) {
			trim = 5;
		}

		if(trim !== false) {
			this.replaceCurrentLine(prefix + currentLine.substr(trim));
		}
		else {
			this.simpleBlockInsert(prefix + ' %');
		}
	}

	doFullscreen() {
		let container = this.textarea.closest('.form-element-value');

		if(container.classList.contains('fullscreen')) {
			container.classList.remove('fullscreen');
			document.body.style.overflow = 'inherit';
			this.fullscreen = false;
			this.preview.innerHTML = '';
			container.style.zIndex = 'inherit';
		}
		else {
			container.classList.add('fullscreen');
			document.body.style.overflow = 'hidden';
			this.fullscreen = true;
			this.preview.innerHTML = marked(this.textarea.value);
			container.style.zIndex = this.getZIndex() + 10;
		}
	}


	updatePreview() {
		if(this.fullscreen) {
			this.preview.innerHTML = marked(this.textarea.value);
		}
	}

	/**
	 * Update the selected action in the toolbar based on the current context.
	 */
	updateSelectedAction() {
		let currentLine = this.getCurrentLine();

		this.controls.querySelectorAll('.active').forEach(el => {
			el.classList.remove('active');
		});

		if(currentLine.startsWith('# ')) {
			this.controls.querySelector('.button[data-action="h1"]').classList.add('active');
		}
		else if(currentLine.startsWith('## ')) {
			this.controls.querySelector('.button[data-action="h2"]').classList.add('active');
		}
		else if(currentLine.startsWith('### ')) {
			this.controls.querySelector('.button[data-action="h3"]').classList.add('active');
		}
		else if(currentLine.startsWith('#### ')) {
			this.controls.querySelector('.button[data-action="h4"]').classList.add('active');
		}
		else if(currentLine.startsWith('##### ')) {
			this.controls.querySelector('.button[data-action="h5"]').classList.add('active');
		}
		else if(currentLine.startsWith('> ')) {
			this.controls.querySelector('.button[data-action="quote"]').classList.add('active');
		}
		else if(currentLine.startsWith('* ')) {
			this.controls.querySelector('.button[data-action="ul"]').classList.add('active');
		}
		else if(currentLine.match(/^[0-9]*\. /) !== null) {
			this.controls.querySelector('.button[data-action="ol"]').classList.add('active');
		}
	}

	simpleBlockInsert(text) {
		if(!this.atStartOfLine()) {
			// If currently on a text line, insert padding below.
			text = "\n\n" + text;
		}
		else if(this.getPreviousLine() !== '') {
			// Previous line is not empty, blocks expect to be padded by enough space.
			text = "\n" + text;
		}

		this.simpleInsert(text);
	}

	simpleInsert(text) {
		let start = this.textarea.selectionStart,
			end = this.textarea.selectionEnd,
			cpos,
			value = this.textarea.value;

		if(text.indexOf('^') === -1) {
			cpos = text.indexOf('%');
		}
		else {
			cpos = text.indexOf('^');
			if(cpos > text.indexOf('%')) {
				// The cursor position is after the text replacement position; this will get shifted.
				cpos--;
			}

			text = text.replace('^', '');
		}

		// Replace linked text
		let preText = value.substr(0, start),
			postText = value.substr(end),
			innerText = value.substr(start, end - start);

		// Fringe case to prevent wrapping newline.
		if(innerText.endsWith("\n")) {
			postText = "\n" + postText;
			innerText = innerText.substr(0, innerText.length - 1);
		}

		this.textarea.value = preText + text.replace('%', innerText) + postText;
		cpos += preText.length + innerText.length;

		/*
		if(start === end) {
			// Append to the end
			if(value === '') {
				this.textarea.value = text.replace('%', '');
			}
			else {
				let lastCharacter = value.substr(value.length - 1);
				if(lastCharacter === ' ' || lastCharacter === "\n") {
					this.textarea.value = value + text.replace('%', '');
					cpos += value.length;
				}
				else {
					this.textarea.value = value + " " + text.replace('%', '');
					cpos += 1 + value.length;
				}
			}
		}
		else {
			// Replace linked text
			let preText = value.substr(0, start),
				postText = value.substr(end),
				innerText = value.substr(start, end - start);

			this.textarea.value = preText + text.replace('%', innerText) + postText;
			cpos += preText.length + innerText.length;
		}*/

		this.textarea.focus();
		this.textarea.selectionStart = cpos;
		this.textarea.selectionEnd = cpos;
	}

	blockInsert(text) {
		let start = this.textarea.selectionStart,
			end = this.textarea.selectionEnd,
			cpos,
			value = this.textarea.value;

		if(text.indexOf('^') === -1) {
			cpos = text.indexOf('%');
		}
		else {
			cpos = text.indexOf('^');
			if(cpos > text.indexOf('%')) {
				// The cursor position is after the text replacement position; this will get shifted.
				cpos--;
			}

			text = text.replace('^', '');
		}

		if(start === end) {
			// Append to the end
			if(value === '') {
				this.textarea.value = text.replace('%', '');
			}
			else {
				this.textarea.value = value + "\n\n" + text.replace('%', '');
				cpos += 2 + value.length;
			}
		}
		else {
			// Replace linked text
			let preText = value.substr(0, start),
				postText = value.substr(end),
				innerText = value.substr(start, end - start);

			this.textarea.value = preText + text.replace('%', innerText) + postText;
			cpos += preText.length + innerText.length;
		}

		this.textarea.focus();
		this.textarea.selectionStart = cpos;
		this.textarea.selectionEnd = cpos;
	}
}
