<?php

use Core\Controller;
use Core\View;

class GettingStartedController extends Controller {
	/**
	 * Display an option to create the first user or state that a user account has been created.
	 *
	 * @throws \Core\AccessDeniedException
	 */
	public function user() {
		$view = $this->getView();

		$usersExist = (UserModel::Count() !== 0);
		$userIsAdmin = \Core\user()->checkAccess('g:admin');

		if($usersExist && !$userIsAdmin) {
			throw new \Core\AccessDeniedException();
		}

		if($usersExist && $userIsAdmin) {
			// User is good, next step is components!
			// This is the entry point for all the getting started system.
			\Core\Session::Set('getting-started', 'user');
			throw new \Core\RedirectException('/gettingstarted/next');
		}

		$view->title = 'Quick Start - Create Account';
		$view->ssl = true;
		$view->addMetaName('robots', 'noindex');
		$view->assign('users_exist', $usersExist);
	}

	/**
	 * Go to the "next" step of the getting started system, whatever that may be
	 *
	 * @core-access-string g:admin
	 */
	public function next() {
		$key = \Core\Session::Get('getting-started');

		if(!$key) {
			throw new \Core\RedirectException('/');
		}

		$next = self::GetNext($key);

		if(!$next) {
			\Core\set_message('You have completed all the steps available in the quick start!', 'success');
			\Core\Session::UnsetKey('getting-started');
			throw new \Core\RedirectException('/');
		}

		\Core\Session::Set('getting-started', $next['key']);
		throw new \Core\RedirectException($next['link']);
	}

	/**
	 * Finish the quickstart guide
	 * 
	 * @core-access-string g:admin
	 */
	public function finish() {
		\Core\set_message('You have completed all the steps available in the quick start!', 'success');
		\Core\Session::UnsetKey('getting-started');
		throw new \Core\RedirectException('/');
	}

	/**
	 * Get the list of links in order
	 *
	 * Each item contains title, link, and description.
	 *
	 * @return array
	 */
	public static function GetLinksArray() : array {
		$links = [
			'user' => [
				'title' => 'Initial User Account',
				'link' => '/gettingstarted/user',
				'description' => 'Create the initial user account',
			],
			'components' => [
				'title' => 'System Updater',
				'link' => '/updater',
				'description' => 'Select and install components available',
			]
		];

		$configureLinks = \Core\HookHandler::DispatchHook('/core/gettingstarted/configurelinks');
		foreach($configureLinks as $key => $link) {
			$links[ 'configure_' . $key ] = [
				'title' => $link['title'],
				'link' => $link['link'],
				'description' => $link['description'],
			];
		}

		return $links;
	}

	public static function GetCurrent(string $linkKey) : ?array {
		$links = self::GetLinksArray();

		foreach($links as $key => $data) {
			if($key === $linkKey) {
				return array_merge(['key' => $key], $data);
			}
		}

		return null;
	}

	public static function GetNext(string $linkKey) : ?array {
		$links = self::GetLinksArray();
		$prev = false;

		foreach($links as $key => $data) {
			if($key === $linkKey) {
				$prev = true;
			}
			elseif($prev) {
				return array_merge(['key' => $key], $data);
			}
		}

		return null;
	}

	public static function HookAttachJS(View $view) : bool {

		if(\Core\user() && \Core\user()->checkExists() && \Core\user()->checkAccess('g:admin') && \Core\Session::IsKeySet('getting-started')) {
			$step = \Core\Session::Get('getting-started');
			$stepData = self::GetCurrent($step);
			$nextStep = self::GetNext($step);

			if($stepData) {
				$url = \Core\resolve_link($stepData['link']);
				$title = $stepData['title'];
			}
			else {
				$url = '';
				$title = '';
			}

			if($nextStep) {
				$next = $nextStep['title'];
			}
			else {
				$next = '';
			}

			$view->addStylesheet('assets/css/gettingstarted.css');
			$view->addScript('assets/js/gettingstarted.js', 'head');
			$view->addScript('<script>renderGettingStarted("' . $title . '", "' . $url . '", "' . $next . '");</script>', 'foot');
		}

		return true;
	}

	public static function HookCatch404(View $view) {
		if (REL_REQUEST_PATH == '/') {

			$view->templatename = 'pages/gettingstarted/index';
			$view->disableCache();

			// Prevent event propagation!
			return false;
		}
	}
}