<?php
/**
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\CLI\CLI;
use Core\Cache;
use Core\Component;
use Core\Controller;
use Core\Core;
use Core\Date\DateTime;
use Core\Dict;
use Core\Filestore\Backends\FileLocal;
use Core\Filestore\Backends\FileRemote;
use Core\Filestore\Contents\ContentTARGZ;
use Core\Filestore\Contents\ContentTGZ;
use Core\Filestore\Directory;
use Core\Filestore\Factory;
use Core\Forms\Form;
use Core\GPG\GPG;
use Core\HealthCheckResult;
use Core\Model;
use Core\NotFoundException;
use Core\RedirectException;
use Core\Templates\Backends\Smarty;
use Core\Theme;
use Core\View;


/**
 * @core-access-string g:admin
 */
class UpdaterController extends Controller {

	/**
	 * Get the controls for the updater.
	 *
	 * @return array|null|void
	 */
	public function getControls(){
		$view = $this->getView();

		$view->addControl(
			[
				'title' => t('STRING_MANAGE_REPOSITORIES'), 
				'link' => '/updater/repos', 
				'icon' => 'cloud']
		);
		$view->addControl(
			[
				'title' => t('STRING_ADD_REPOSITORY_SITE'), 
				'link' => 'updater/repos/add', 
				'icon' => 'add']
		);

		$view->addControl(
			[
				'title' => t('STRING_MANAGE_GPG_KEYS'), 
				'link' => '/updater/keys', 
				'icon' => 'key']
		);
		$view->addControl(
			[
				'title' => t('STRING_IMPORT_GPG_KEY'), 
				'link' => '/updater/keys/import', 
				'icon' => 'add'
			]
		);

		$view->addControl(
			[
				'title' => t('STRING_FIND_NEW_PACKAGES'), 
				'link' => '/updater/browse', 
				'icon' => 'search']
		);
		$view->addControl(
			[
				'title' => t('STRING_MANUALLY_UPLOAD_PACKAGE'),
			    'link' => '/updater/upload',
			    'icon' => 'upload',
			]
		);
	}

	/**
	 * Run the Updater and view information about packages currently installed and their update status
	 *
	 * If ran from the CLI, "--package='type|key|version'" can be provided to auto install
	 * a package from the remote repositories.
	 *
	 * @core-exec-mode BOTH
	 * @core-access-string g:admin
	 */
	public function index() {
		$view = $this->getView();

		if(EXEC_MODE === 'CLI') {
			// This application behaves differently when ran in CLI mode.
			$this->_indexCLI();
			return;
		}

		// @TODO Scan components and detect if they've changed.
		// warn the user if they have.

		$sitecount = UpdateSiteModel::Count();
		$components = array();
		foreach(Core::GetComponents() as $k => $c){
			// Skip the core.
			if($k == 'core') continue;

			$components[$k] = $c;
		}

		// These should really be sorted by name
		ksort($components);

		// Merge in the disabled ones too, (sorted by name as well)
		$disabled = Core::GetDisabledComponents();
		ksort($disabled);
		$components = array_merge($components, $disabled);
		
		// Count how many components are not installed, this will provide the one-click-install for all new components.
		$notinstalled = 0;
		foreach($components as $c) {
			if(!$c->isInstalled()) {
				$notinstalled ++;
			}
		}
		
		// If the theme is disabled, this won't be available.
		$themes = Theme::GetAllThemes();

		$view->title = 't:STRING_SYSTEM_UPDATER';
		$view->assign('sitecount', $sitecount);
		$view->assign('components', $components);
		$view->assign('core', Core::GetComponent('core'));
		$view->assign('themes', $themes);
		$view->assign('not_installed_count', $notinstalled);
	}

	/**
	 * Check for updates controller
	 *
	 * This is just a very simple json function to return true or false on if there are updates for currently installed components.
	 *
	 * This is so simple because its sole purpose is to just notify the user if there is an update available.
	 * For more full-featured update scripts, look at the getupdates page; that actually returns the updates.
	 *
	 * @core-access-string g:admin
	 */
	public function check() {
		$view = $this->getView();



		$updates = UpdaterHelper::GetUpdates();

		$view->contenttype = View::CTYPE_JSON;
		// Will get overwrote if found to be true.
		$view->jsondata = false;
		$view->record = false;


		if(isset($updates['core']) && $updates['core']['status'] == 'update'){
			$view->jsondata = true;
			return;
		}

		// Same for components and themes, (only a little more depth)
		foreach($updates['components'] as $up){
			if($up['status'] == 'update'){
				$view->jsondata = true;
				return;
			}
		}
		foreach($updates['themes'] as $up){
			if($up['status'] == 'update'){
				$view->jsondata = true;
				return;
			}
		}
	}


	/**
	 * Get the list of updates from remote repositories, (or session cache).
	 *
	 * @core-access-string g:admin
	 */
	public function getupdates() {
		$view = $this->getView();
		$req  = $this->getRequest();

		$updater = new \Core\Updater();
		$allPackages = $updater->getAllPackages();
		$allUpdates = [];

		foreach($allPackages as $package) {
			// We only want to look at packages currently installed and active.
			if(!$package->checkIsInstalledLocally()) {
				continue;
			}

			$allUpdates[] = [
				'type' => $package->type,
				'key' => $package->key,
				'name' => $package->name,
				'update' => $package->checkIsUpdateAvailable(),
				'version' => $package->version
			];
		}

		$view->contenttype = View::CTYPE_JSON;
		$view->jsondata = [
			'updates' => $allUpdates
		];
	}

	/**
	 * Similar to getupdates, but this returns 'packages' and only not-locally-installed packages.
	 *
	 * @core-access-string g:admin
	 */
	public function getpackages() {
		$view = $this->getView();
		$req  = $this->getRequest();

		$updater = new \Core\Updater();
		$allPackages = $updater->getAllPackages();
		$allUpdates = [];

		foreach($allPackages as $package) {
			// Skip locally installed packages; we only want new.
			if($package->checkIsInstalledLocally()) {
				continue;
			}

			$allUpdates[] = [
				'type'        => $package->type,
				'key'         => $package->key,
				'name'        => $package->name,
				'version'     => $package->version,
				'logo'        => $package->logo,
				'url'         => $package->url,
				'description' => $package->description,
				'packager'    => $package->packager,
				'email'       => $package->email,
				'released'    => DateTime::FormatString($package->released, DateTime::FORMAT_SHORTDATE),
			];
		}

		$view->contenttype = View::CTYPE_JSON;
		$view->jsondata = [
			'packages' => $allUpdates
		];
	}

	/**
	 * Sites listing controller, displays all update sites and links to manage them.
	 *
	 */
	public function repos() {
		$view = $this->getView();

		if(!is_dir(GPG_HOMEDIR)){
			// Try to create it?
			if(is_writable(dirname(GPG_HOMEDIR))){
				// w00t
				mkdir(GPG_HOMEDIR);
			}
			else{
				\Core\set_message(GPG_HOMEDIR . ' does not exist and could not be created!  Please fix this before proceeding!', 'error');
			}
		}
		elseif(!is_writable(GPG_HOMEDIR)){
			\Core\set_message(GPG_HOMEDIR . ' is not writable!  Please fix this before proceeding!', 'error');
		}


		$sites = UpdateSiteModel::Find();

		$view->title = 't:STRING_MANAGE_REPOSITORIES';
		$view->assign('sites', $sites);

	}

	/**
	 * Add a repository to the site.
	 * This will also handle the embedded keys, (as of 2.4.5).
	 *
	 * This contains the first step and second steps.
	 */
	public function repos_add() {
		$request = $this->getRequest();
		$view    = $this->getView();

		$site = new UpdateSiteModel();

		$form = Form::BuildFromModel($site);
		$form->set('action', \Core\resolve_link('/updater/repos/add'));
		$form->addElement('submit', array('value' => 'Next'));

		$view->title = 'Add Repo';
		// Needed because dynamic pages do not record navigation.
		$view->addBreadcrumb('Repositories', 'updater/repos');

		$view->assign('form', $form);

		if(!is_dir(GPG_HOMEDIR)){
			// Try to create it?
			if(is_writable(dirname(GPG_HOMEDIR))){
				// w00t
				mkdir(GPG_HOMEDIR);
			}
			else{
				\Core\set_message(GPG_HOMEDIR . ' does not exist and could not be created!  Please fix this before proceeding!', 'error');
				$form = null;
			}
		}
		elseif(!is_writable(GPG_HOMEDIR)){
			\Core\set_message(GPG_HOMEDIR . ' is not writable!  Please fix this before proceeding!', 'error');
			$form = null;
		}


		// This is the logic for step 2 (confirmation).
		// This is after all the template logic from step 1 because it will fallback to that form if necessary.
		if($request->isPost()){
			$url      = $request->getPost('model[url]');
			$username = $request->getPost('model[username]');
			$password = $request->getPost('model[password]');

			// Validate and standardize this repo url.
			// This is because most people will simply type corepl.us.
			if(strpos($url, '://') === false){
				$url = 'https://' . $url;
			}

			// Lookup that URL first!
			if(UpdateSiteModel::Count(array('url' => $url)) > 0){
				\Core\set_message($url . ' is already used!', 'error');
				return;
			}

			// Load up a new Model, that's the easiest way to pull the repo data.
			$model = new UpdateSiteModel();
			$model->setFromArray([
				'url' => $url,
				'username' => $username,
				'password' => $password,
			]);

			// From here on out, populate the previous form with this new model.
			$form = Form::BuildFromModel($model);
			$form->set('action', \Core\resolve_link('/updater/repos/add'));
			$form->addElement('submit', array('value' => 'Next'));
			$view->assign('form', $form);

			/** @var FileRemote $remote */
			$remote = $model->getFile();

			if($remote->requiresAuthentication()){
				if(!$username){
					\Core\set_message($url . ' requires authentication!', 'error');
					return;
				}
				else{
					\Core\set_message('Invalid credentials for ' . $url, 'error');
					return;
				}
			}

			if(!$model->isValid()){
				\Core\set_message($url . ' does not appear to be a valid repository!', 'error');
				return;
			}

			$repo = new RepoXML();
			$repo->loadFromFile($remote);
			
			// Make sure the keys are good
			if(!$repo->validateKeys()){
				\Core\set_message('There were invalid/unpublished keys in the repo!  Refusing to import.', 'error');
				return;
			}

			// The very final bit of this logic is to look and see if there's a "confirm" present.
			// If there is, the user clicked accept on the second page and I need to go ahead and import the data.
			if($request->getPost('confirm')){
				$model->set('description', $repo->getDescription());
				$model->save();
				$keysimported = 0;
				$keycount     = sizeof($repo->getKeys());
				$gpg          = new GPG();

				foreach($repo->getKeys() as $keyData){
					if(!$keyData['installed']){
						try{
							if($keyData['contents']){
								// Local import! :)
								$gpg->importKey($keyData['contents']);
							}
							else{
								$gpg->importKey($keyData['key']);
							}
							++$keysimported;
						}
						catch(Exception $e){
							\Core\set_message('Unable to import key [' . $keyData['key'] . '] from keyserver!' . $e->getMessage(), 'error');
						}
					}
					else{
						// Flag already-imported keys as good.
						++$keysimported;
					}
				}

				if(!$keycount){
					\Core\set_message('Added repository site successfully!', 'success');
				}
				elseif($keycount != $keysimported){
					\Core\set_message('Added repository site, but unable to import ' . ($keycount-$keysimported) . ' key(s).', 'info');
				}
				else{
					\Core\set_message('Added repository site and imported ' . $keysimported . ' key(s) successfully!', 'success');
				}

				\core\redirect('/updater/repos');
			}

			$view->templatename = 'pages/updater/repos_add2.tpl';
			$view->assign('description', $repo->getDescription());
			$view->assign('keys', $repo->getKeys());
			$view->assign('url', $url);
			$view->assign('username', $username);
			$view->assign('password', $password);
		}
	}

	public function repos_edit() {
		$request = $this->getRequest();
		$view    = $this->getView();

		$site = UpdateSiteModel::Construct($request->getParameter(0));
		if(!$site->exists()){
			return View::STATUS_NOTFOUND;
		}

		$form = Form::BuildFromModel($site);
		$form->set('callsmethod', 'UpdaterController::_SaveRepo');
		$form->addElement('submit', array('value' => 'Update'));

		$view->title = 'Update Repo';
		// Needed because dynamic pages do not record navigation.
		$view->addBreadcrumb('Repositories', 'updater/repos');
		$view->assign('form', $form);
	}

	/**
	 * Page to remove a repository.
	 */
	public function repos_delete(){
		$request = $this->getRequest();
		if(!$request->isPost()){
			return View::STATUS_BADREQUEST;
		}

		$model = new UpdateSiteModel($request->getParameter(0));
		if(!$model->exists()){
			return View::STATUS_NOTFOUND;
		}

		$model->delete();
		\Core\set_message('Removed repository successfully', 'success');
		\core\redirect('/updater/repos');
	}

	/**
	 * Browse the repositories for a components and themes
	 *
	 * @core-access-string g:admin
	 */
	public function browse() {
		$view = $this->getView();
		$view->title = 't:STRING_SYSTEM_PACKAGE_BROWSE';
	}

	public function keys() {
		$view = $this->getView();

		// Get the existing keys.
		$gpg = new GPG();
		$keys = $gpg->listKeys();
		
		$managerAvailable = Core::IsComponentAvailable('gpg-key-manager');

		$view->title = "GPG Keys";
		$view->assign('directory', GPG_HOMEDIR);
		$view->assign('keys', $keys);
		$view->assign('manager_available', $managerAvailable);
	}

	public function keys_import() {
		$view  = $this->getView();
		$req   = $this->getRequest();
		$error = null;

		if(!is_writable(GPG_HOMEDIR)){
			$error = 'Please ensure that ' . GPG_HOMEDIR . ' is writable!';
		}

		if($req->isPost()){
			// Receive public key from a keyserver.
			if($_POST['pubkeyid']){
				$id = strtoupper(preg_replace('/[^a-zA-Z0-9]*/', '', $_POST['pubkeyid']));
				exec('gpg --keyserver-options timeout=6 --homedir "' . GPG_HOMEDIR . '" --no-permission-warning --keyserver "hkp://pool.sks-keyservers.net" --recv-keys "' . $id . '"', $output, $result);
				if($result != 0){
					$error = 'Unable to lookup ' . $id . ' from keyserver.';
				}
				else{
					\core\redirect('/updater/keys');
				}
			}
			elseif($_POST['pubkey']){
				$tmp = Factory::File('tmp/importkey-' . \Core\random_hex(2) . '.gpg');
				$tmp->putContents($_POST['pubkey']);
				exec('gpg --homedir "' . GPG_HOMEDIR . '" --no-permission-warning --import "' . $tmp->getFilename() . '"', $output, $result);
				$tmp->delete();
				if($result != 0){
					$error = 'Unable to import requested key.';
				}
				else{
					\core\redirect('/updater/keys');
				}
			}
			elseif($_FILES['pubkeyfile']){
				$tmp = Factory::File($_FILES['pubkeyfile']['tmp_name']);
				$tmp->putContents($_POST['pubkey']);
				exec('gpg --homedir "' . GPG_HOMEDIR . '" --no-permission-warning --import "' . $tmp->getFilename() . '"', $output, $result);
				$tmp->delete();
				if($result != 0){
					$error = 'Unable to import requested key.';
				}
				else{
					\core\redirect('/updater/keys');
				}
			}
		}

		$view->addBreadcrumb('GPG Keys', 'updater/keys');
		$view->title = 'Import Key';
		$view->assign('error', $error);
	}

	public function keys_delete() {
		$view = $this->getView();
		$req = $this->getRequest();

		// This is a post-only page!
		if(!$req->isPost()){
			$view->status = View::STATUS_BADREQUEST;
			return;
		}

		$key = $req->getParameter(0);
		if(!$key){
			$view->status = View::STATUS_BADREQUEST;
			return;
		}

		$key = strtoupper(preg_replace('/[^a-zA-Z0-9]*/', '', $key));

		exec('gpg --homedir "' . GPG_HOMEDIR . '" --no-permission-warning --batch --yes --delete-key "' . $key . '"', $output, $result);
		if($result != 0){
			\Core\set_message('Unable to remove key ' . $key, 'error');
		}
		\core\redirect('/updater/keys');
	}

	/**
	 * Install or upgrade a component from a remote repository
	 *
	 * @core-access-string g:admin
	 */
	public function component_install() {
		$view = $this->getView();
		$req = $this->getRequest();

		$name    = $req->getParameter(0);
		$version = $req->getParameter('version');

		$this->_performInstall('component', $name, $version);
	}

	/**
	 * Admin page to kick off the installation or upgrade of themes.
	 *
	 * @core-access-string g:admin
	 */
	public function theme_install() {
		$view = $this->getView();
		$req  = $this->getRequest();

		$name    = $req->getParameter(0);
		$version = $req->getParameter('version');

		$this->_performInstall('theme', $name, $version);
	}

	/**
	 * Admin page to kick off the installation or upgrade of the core.
	 *
	 * @core-access-string g:admin
	 */
	public function core_install() {
		$view = $this->getView();
		$req  = $this->getRequest();

		$version = $req->getParameter('version');

		$this->_performInstall('core', 'core', $version);
	}

	/**
	 * Admin page to do exactly as it states; update everything possible.
	 *
	 * @core-access-string g:admin
	 */
	public function update_everything(){
		$view = $this->getView();
		$req  = $this->getRequest();

		if(!$req->isPost()){
			return View::STATUS_BADREQUEST;
		}

		$view->mode = View::MODE_NOOUTPUT;
		$view->render();

		CLI::PrintHeader('Loading Updates');
		CLI::PrintProgressBar(2);
		$everything = UpdaterHelper::GetUpdates();
		$updates = [];

		// Core not updateable? remove them.
		if($everything['core']['status'] == 'update'){
			$updates[] = $everything['core'];
		}

		// Check each component too.
		foreach($everything['components'] as $c => $dat){
			if($dat['status'] == 'update'){
				$updates[] = $everything['components'][$c];
			}
		}

		// And the themes.
		foreach($everything['themes'] as $c => $dat){
			if($dat['status'] == 'update'){
				$updates[] = $everything['themes'][$c];
			}
		}
		CLI::PrintProgressBar(5);

		// Calculate the progress amount for each iteration based on the number of total updates to apply.
		$progressEa = '+' . (95 / (sizeof($updates) * 2));

		CLI::PrintHeader('Installing ' . sizeof($updates) . ' Updates');
		foreach($updates as $thing){
			try{
				CLI::PrintLine('Installing ' . $thing['type'] . ' ' . $thing['name'] . ' ' . $thing['version']);

				$return = UpdaterHelper::PerformInstall($thing['type'], $thing['name'], $thing['version'], true, false);
				CLI::PrintProgressBar($progressEa);

				if($return['status'] == 1){
					// Good to install!
					$return = UpdaterHelper::PerformInstall($thing['type'], $thing['name'], $thing['version'], false, false);
					CLI::PrintProgressBar($progressEa);

					if($return['status'] != 1){
						CLI::PrintWarning($return['message']);
					}
				}
				else{
					CLI::PrintWarning($return['message']);
				}
			}
			catch(Exception $e){
				CLI::PrintError($e->getMessage());
			}
		}
	}


	/**
	 * Helper function called by the *_install views.
	 *
	 * @param $type
	 * @param $name
	 * @param $version
	 */
	private function _performInstall($type, $name, $version){
		$view = $this->getView();
		$req  = $this->getRequest();

		// If this is a GET request, just do the checks.
		// If POST, then it's confirmed to be requested.
		$dryRun = (EXEC_MODE === 'WEB' && !$req->isPost()) || EXEC_MODE === 'CLI';

		$view->mode = View::MODE_STREAM;
		$view->render();


		// Check if this user can write to the local directory.
		// If not, then automatic installation cannot proceed!
		if(!is_writable(ROOT_PDIR)) {
			$user = posix_getpwuid(posix_geteuid());
			CLI::PrintError(ROOT_PDIR . ' is not writable by the user ' . $user['name'] . ', unable to proceed with automatic installation');

			if(get_current_user() !== $user['name']) {
				CLI::PrintLine('This application was probably installed by a local user, try the command line interface instead.');
			}
			return;
		}

		CLI::PrintActionStart('Downloading Repository Package Metas');
		$updater = new \Core\Updater();
		$updater->getAllPackages();
		CLI::PrintActionStatus(true);

		// Look for this specific update, since this is from the updater, it needs to be from an update site.
		$packageMeta = $updater->getPackage($type, $name);
		if($packageMeta === null) {
			CLI::PrintError($name . ' could not be located in the repositories!');
			return;
		}

		if($packageMeta->version !== $version) {
			CLI::PrintError($name . ' ' . $version . ' could not be located in the repositories!  Please go back and check the version.');
			return;
		}


		// Get the queue of packages to install; this may be required
		// because a given package may require additional packages to be installed prior
		// to it, and the updater will resolve those checks.
		CLI::PrintActionStart('Resolving Dependencies');
		try {
			$queue = $updater->getInstallQueue($type, $name);
		}
		catch(Exception $ex) {
			CLI::PrintActionStatus(false);
			CLI::PrintError($ex->getMessage());
			return;
		}

		// Scan the GPG keys of all the new packages too!
		// This will at least just verify if they're available.
		foreach($queue as $packageMeta) {
			if(!$packageMeta->checkIsGPGKeyAvailable()) {
				CLI::PrintActionStatus(false);
				CLI::PrintError('GPG Key ' . $packageMeta->gpgKey . ' is not present!  Unable to verify package integrity.');
				return;
			}
		}
		CLI::PrintActionStatus(true);
		

		if($dryRun) {
			CLI::PrintHeader('Pending Changes To By Applied');
			foreach($queue as $packageMeta) {
				CLI::PrintLine('Install ' . $packageMeta->name . ' ' . $packageMeta->version);
			}

			if(EXEC_MODE === 'WEB') {
				$url = \Core\resolve_link('/updater/' . $type . '/install/' . $name . '?version=' . $version);

				echo <<<EOD
<hr/>
<form action="$url" method="POST">
<input type="submit" value="Confirm Changes and Proceed"/>
</form>
EOD;
				return;
			}
			else {
				$confirm = CLI::PromptUser('Confirm Changes and Proceed?', 'bool', false);
				if(!$confirm) {
					return;
				}
			}
		}


		// If here, then dryrun was set to false, (or otherwise conditions are met to proceed)
		foreach($queue as $packageMeta) {
			CLI::PrintActionStart('Downloading ' . $packageMeta->name . ' ' . $packageMeta->version);
			$package = $packageMeta->downloadPackage();
			CLI::PrintActionStatus(true);

			// From here, we have the full Package to work with!
			// This gives us all the data from within the package, not just meta information.

			CLI::PrintActionStart('Verifying Package');
			try {
				$tarball = $package->getExtractedPackage();
				CLI::PrintActionStatus(true);
			}
			catch(Exception $ex) {
				CLI::PrintActionStatus(false);
				CLI::PrintError($ex->getMessage());
				return;
			}

			CLI::PrintActionStart('Extracting Files');
			try {
				/** @var ContentTARGZ $fileObject */
				$fileObject = $tarball->getContentsObject();
				$fileObject->extract(ROOT_PDIR);
				CLI::PrintActionStatus(true);
			}
			catch(Exception $ex) {
				CLI::PrintActionStatus(false);
				CLI::PrintError($ex->getMessage());
				return;
			}

			if($packageMeta->type === 'component' || $packageMeta->type === 'core') {

				// Clear Model caching, as the database structure may change here.
				Model::$_ModelCache = [];
				Model::$_ModelFindCache = [];
				Model::$_ModelSchemaCache = [];

				CLI::PrintActionStart('Loading Component XML');
				try {
					// Core uses a different path than most components, so take that into account
					$file = $packageMeta->type === 'component' ? 'component.xml' : 'core/component.xml';
					$component = new Component($packageMeta->getBaseDirectory() . $file);
					$component->loadAllFromXML();
					$component->loadAllFromDB();
					CLI::PrintActionStatus(true);
				}
				catch(Exception $ex) {
					CLI::PrintActionStatus(false);
					CLI::PrintError($ex->getMessage());
					return;
				}


				if($packageMeta->checkIsInstalledLocally()) {
					// Already installed, perform an upgrade.
					CLI::PrintActionStart('Upgrading ' . $packageMeta->name);
					$changes = $component->upgrade();
					CLI::PrintActionStatus(count($changes['errors']) ? 'warn' : true);
					CLI::PrintLine($changes['changes']);
					if(count($changes['errors'])) {
						CLI::PrintError($changes['errors']);
					}
				}
				else {
					CLI::PrintActionStart('Installing ' . $packageMeta->name);
					$changes = $component->install();
					$component->enable();
					CLI::PrintActionStatus(count($changes['errors']) ? 'warn' : true);
					CLI::PrintLine($changes['changes']);
					if(count($changes['errors'])) {
						CLI::PrintError($changes['errors']);
					}
				}

				// Make this new or updated component available immediately.
				Core::RegisterComponent($component);
			}
		}

		// Clear the entire system cache now that all procedures are complete.
		Cache::Flush();
		Smarty::FlushCache();

		CLI::PrintSuccess('All Actions Applied');

		if(EXEC_MODE === 'WEB') {
			echo '<a href="' . \Core\resolve_link('/updater') . '">Go Back</a>';
		}
	}

	private function _indexCLI() {
		$view = $this->getView();

		$view->mode = View::MODE_STREAM;
		$view->render();


		// The fetching of update sites isn't required, but is here for informational reasons.
		CLI::PrintActionStart('Fetching and Checking Update Repositories');
		$updateSitesValid = [];
		$updateSitesInvalid = [];
		$updateSites = UpdateSiteModel::Find();
		foreach($updateSites as $site) {
			/** @var UpdateSiteModel $site */

			if (!$site->checkIsValid()) {
				$updateSitesInvalid[] = $site;
			}
			else {
				$updateSitesValid[] = $site;
			}
		}
		CLI::PrintActionStatus(true);

		CLI::PrintLine('Found ' . count($updateSites) . ' repository(s)');
		foreach($updateSitesValid as $site) {
			// Use the action start/status because we like the formatting.
			CLI::PrintActionStart($site->url, '');
			CLI::PrintActionStatus(true);
		}
		foreach($updateSitesInvalid as $site) {
			// Use the action start/status because we like the formatting.
			CLI::PrintActionStart($site->url, '');
			CLI::PrintActionStatus('fail');
		}

		CLI::PrintActionStart('Downloading Package Meta Information');
		$updater = new \Core\Updater();
		$allPackages = $updater->getAllPackages();
		CLI::PrintActionStatus(true);

		// Allow a package to be specified manually, useful for quick installs.
		if($this->request->getParameter('package')) {
			$package = $this->request->getParameter('package');
			$found = false;
			list($type, $name, $version) = explode('|', $package);

			foreach($allPackages as $package) {
				if($package->type === $type && $package->key === $name && $package->version === $version) {
					$this->_performInstall($type, $name, $version);
					$found = true;
				}
			}

			if(!$found) {
				CLI::PrintError('Requested package not found in repositories!');
			}
			return;
		}

		$table = new \Core\CLI\CLITable();
		$table->addColumn('#', 'right');
		$table->addColumn('Type');
		$table->addColumn('Component / Theme');
		$table->addColumn('Version');
		$table->addColumn('Status', 'center');
		$table->addColumn('Repo Version');
		$table->addColumn('Repo Status', 'center');

		$allUpdates = [];
		foreach($allPackages as $package) {
			// We only want to look at packages currently installed and active.
			if(!$package->checkIsInstalledLocally()) {
				continue;
			}

			$allUpdates[$package->type . '-' . $package->key] = [
				'update' => $package->checkIsUpdateAvailable(),
				'version' => $package->version
			];
		}

		// Compile all the current sources into a single array.
		$allComponents = [];
		foreach(Core::GetComponents() as $component) {
			$group = $component->getKeyName() === 'core' ? 'a' : 'b';
			$allComponents[ $group . '-' . $component->getName()] = $component;
		}

		foreach(Core::GetDisabledComponents() as $component) {
			$group = $component->getKeyName() === 'core' ? 'a' : 'b';
			$allComponents[ $group . '-' . $component->getName()] = $component;
		}

		foreach(Theme::GetAllThemes() as $theme) {
			$group = 'c';
			$allComponents[ $group . '-' . $theme->getName()] = $theme;
		}

		// Sort them for convenience.
		ksort($allComponents);

		// Add all data to the table
		$options = [];
		$counter = 0;
		foreach($allComponents as $component) {
			if($component instanceof Theme) {
				$pkgType = 'theme';
				$type = 'Theme';
			}
			elseif($component->getKeyName() === 'core') {
				$pkgType = 'core';
				$type = 'Core';
			}
			else {
				$pkgType = 'component';
				$type = 'Component';
			}

			$pkgKey = $pkgType . '-' . $component->getKeyName();
			$update = $allUpdates[$pkgKey] ?? ['update' => false, 'version' => false];
			++$counter;

			if(!$component->isEnabled()) {
				$repoStatus = 'N/A';
				$repoStatusColor = null;
			}
			elseif($update['update']) {
				$repoStatus = 'Update Available!';
				$repoStatusColor = COLOR_SUCCESS;
			}
			elseif($update['version']) {
				$repoStatus = 'Available In Repo';
				$repoStatusColor = null;
			}
			else {
				$repoStatus = 'Not In Repo';
				$repoStatusColor = COLOR_ERROR;
			}

			if($component->isEnabled()) {
				// If it's enabled, it's also installed.
				$status = 'Installed / Enabled';
			}
			elseif($component->isInstalled() && !$component->isEnabled()) {
				// Components can be installed, but disabled.
				$status = 'Installed / Disabled';
			}
			else {
				$status = 'Not Installed';
			}

			$options[$counter] = [
				'component' => $component,
				'package' => $update,
			];

			$table->addRecord([
				$counter,
				$type,
				$component->getName(),
				$component->getVersion(),
				$status,
				$update['version'],
				[
					$repoStatus,
					'color' => $repoStatusColor,
				]
			]);
		}

		$selection = CLI::PromptUser('Current Components and Update Status', $table);
		if(isset($options[$selection])) {
			// Next "page"!
			// User selected a component to manage, so display options for that component.
			$this->_indexCLIComponentSelected($options[$selection]['component'], $options[$selection]['package']['version']);
		}
	}

	private function _indexCLIComponentSelected(Component $component, $updateVersion) {
		// Display some information about this package
		CLI::PrintHeader($component->getName() . ' ' . $component->getVersion());
		CLI::PrintLine($component->getDescription());

		if($component->isInstalled()) {
			$installed = COLOR_SUCCESS . 'Yes' . COLOR_RESET;
		}
		else {
			$installed = 'No';
		}

		if($component->isEnabled()) {
			$enabled = COLOR_SUCCESS . 'Enabled' . COLOR_RESET;
		}
		else {
			$enabled = COLOR_ERROR . 'Disabled' . COLOR_RESET;
		}

		if($updateVersion) {
			$repo = COLOR_SUCCESS . 'Available in repository' . COLOR_RESET;
			$repoVersion = $updateVersion;
		}
		else {
			$repo = 'No, not available in any repository';
			$repoVersion = 'N/A';
		}

		if($component instanceof Theme) {
			$type = 'theme';
		}
		else {
			$type = $component->getKeyName() === 'core' ? 'core' : 'component';
		}

		CLI::PrintLine('Installed:        ' . $installed);
		CLI::PrintLine('Status:           ' . $enabled);
		CLI::PrintLine('Update Available: ' . $repo);
		CLI::PrintLine('Update Version:   ' . $repoVersion);

		// Create a dictionary of actions the user can take on this Component.
		$options = new Dict();

		// General install/enable/disable options
		if($type === 'component') {
			if(!$component->isInstalled()) {
				$options->addItem('install', '_I_nstall and Activate');
			}
			elseif(!$component->isEnabled()) {
				$options->addItem('enable', '_E_nable');
			}
			else {
				$options->addItem('disable', '_D_isable');
				$options->addItem('uninstall', '_R_emove and Uninstall');
			}
		}
		elseif($type === 'core') {
			if(!$component->isInstalled()) {
				$options->addItem('install', '_I_nstall and Activate');
			}
		}
		elseif($type === 'theme') {
			if(!$component->isInstalled()) {
				$options->addItem('install', '_I_nstall and Activate');
			}
		}


		// Updater options
		if($updateVersion) {
			$options->addItem('upgrade', '_U_pgrade From Repo');
		}

		if($options->count()) {
			$ans = CLI::PromptUser('Perform An Action', $options);
		}
		else {
			$ans = null;
		}


		switch($ans) {
			case 'upgrade':
				$this->_performInstall($type, $component->getKeyName(), $updateVersion);
				break;
			case 'install':
				$component->install();
				$component->enable();
				CLI::PrintSuccess('Installed ' . $component->getName());
				break;
			case 'enable':
				$component->enable();
				CLI::PrintSuccess('Enabled ' . $component->getName());
				break;
			case 'disable':
				$component->disable();
				CLI::PrintSuccess('Disabled ' . $component->getName());
				break;
			case 'uninstall':
				if(CLI::PromptUser(
					'This will completely remove associated data with this component!  Continue?',
					bool,
					false
				)) {
					$component->uninstall();
					CLI::PrintSuccess('Uninstalled ' . $component->getName());
				}
				break;
		}
	}

	public static function _SaveRepo(Form $form) {
		try{
			$model = $form->getModel();
			$model->save();
		}
		catch(Exception $e){
			\Core\set_message($e->getMessage(), 'error');
			return false;
		}

		\Core\set_message('Updated repository successfully', 'success');
		return '/updater/repos';
	}

	/**
	 * Call to check for updates as part of the health checking system in Core.
	 * 
	 * @return array
	 */
	public static function _HealthCheckHook(){
		// SERVER_ADMIN_EMAIL
		
		$checks      = [];

		// Check that the update sites are good and ready.
		$updateSites = UpdateSiteModel::Find();
		if(!count($updateSites)) {
			$checks[] = HealthCheckResult::ConstructWarn('No update repositories available!');
		}
		else {
			foreach($updateSites as $site) {
				if($site->checkIsValid()) {
					$checks[] = HealthCheckResult::ConstructGood($site->url . ' update repository is available');
				}
				else {
					$checks[] = HealthCheckResult::ConstructWarn($site->url . ' update repository is not available or invalid');
				}
			}
		}

		// Pull the actual package updates
		$updater = new \Core\Updater();
		$allPackages = $updater->getAllPackages();

		foreach($allPackages as $package) {
			if($package->checkIsUpdateAvailable()) {
				switch ($package->type) {
					case 'core':
						$checks[] = HealthCheckResult::ConstructWarn(
							t('STRING_WARNING_UPDATER_CORE_OUTDATED'),
							t('MESSAGE_WARNING_UPDATER_CORE_OUTDATED_S_AVAILABLE', $package->version),
							'/updater'
						);
						break;
					case 'component':
						$checks[] = HealthCheckResult::ConstructWarn(
							t('STRING_WARNING_UPDATER_COMPONENT_S_OUTDATED', $package->name),
							t('MESSAGE_WARNING_UPDATER_COMPONENT_S_OUTDATED_S_AVAILABLE', $package->version),
							'/updater'
						);
						break;
					case 'theme':
						$checks[] = HealthCheckResult::ConstructWarn(
							t('STRING_WARNING_UPDATER_THEME_S_OUTDATED', $package->name),
							t('MESSAGE_WARNING_UPDATER_THEME_S_OUTDATED_S_AVAILABLE', $package->version),
							'/updater'
						);
						break;
				}
			}
		}

		return $checks;
	}
}

