<?php
/**
 * Main controller for the user system
 *
 * Provides both admin functions and front-end user functions.
 *
 * @package Core\User
 * @since 1.9
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\AccessDeniedException;
use Core\ConfigHandler;
use Core\Controller;
use Core\Date\DateTime;
use Core\Date\Timezone;
use Core\EmailAPI\Email;
use Core\Filestore\Contents\ContentCSV;
use Core\Filestore\Factory;
use Core\Forms\AccessStringInput;
use Core\Forms\Form;
use Core\HookHandler;
use Core\ListingTable\Table;
use Core\RedirectException;
use Core\Session;
use Core\User;
use Core\User\Helper;
use Core\UserNotActivatedException;
use Core\View;
use Core\ViewControls;
use Core\FilterForm;
use Core\User\UserTable;
use Core\Dict;
use Core\Model;
use Core\ViewControl;
use Core\NotFoundException;



/**
 * Main controller for the user system
 *
 * This controller is only responsible for Core user functions.
 * Authentication-specific functions must be contained on the specific auth driver or its respective controller.
 *
 * @package Core\User
 */
class UserController extends Controller{

	/**
	 * Admin listing of all the users
	 * 
	 * @core-access-string p:/user/users/manage
	 *
	 * @return null|int
	 */
	public function admin(){
		$view = $this->getView();

		//$enableAvatars = ConfigHandler::Get('/user/enableavatar');
		
		$table = new UserTable();

		$table->loadFromRequest();
		$table->addControlToPage(ViewControl::FactoryTLI('t:STRING_ADD_USER', '/user/admin/add', ViewControl::ICON_ADD));
		$table->addControlToPage(ViewControl::FactoryTLI('t:STRING_IMPORT_USERS', '/user/import', ViewControl::ICON_IMPORT));

		$view->mastertemplate = 'admin';
		$view->title = 't:STRING_USERS';
		$view->assign('listing', $table);
	}
	
	/**
	 * Administratively create a new user
	 * 
	 * @core-access-string p:/user/users/manage
	 */
	public function admin_add() {
		$view          = $this->getView();
		
		$form = new Form();
		$model = new UserModel();
		
		$model->set('active', '0');
		$form->set('callsmethod', 'Core\\User\\Helper::RegisterHandler');
		$form->addModel($model, 'user');
		
		//$form->removeElement('user[active]');
		//$form->removeElement('user[admin]');
		
		$form->addElement('submit', ['value' => 't:STRING_CREATE_USER']);

		$view->mastertemplate = 'admin';
		$view->autoBreadcrumbs('/user/admin');
		$view->title = 't:STRING_CREATE_USER';
		$view->assign('form', $form);
	}

	/**
	 * This view has been removed and functionality merged into /user/view
	 *
	 * @deprecated 20190720
	 *
	 * @throws RedirectException
	 * @throws AccessDeniedException
	 */
	public function me(){
		$user    = \Core\user();

		if(!$user->checkExists()){
			throw new AccessDeniedException();
		}
		else {
			throw new RedirectException('/user/view/' . $user->id);
		}
	}

	/**
	 * Show a user's profile
	 *
	 * @throws NotFoundException
	 * @throws AccessDeniedException
	 * @throws UserNotActivatedException
	 */
	public function view() {
		$view    = $this->getView();
		$request = $this->getRequest();

		$user = UserModel::Construct($request->getParameter(0));
		$currentUser = \Core\user();

		$isManager = $currentUser->checkAccess('p:/user/users/manage');

		if(!$user->checkExists()){
			throw new NotFoundException();
		}

		if($user->id !== $currentUser->id && !$isManager) {
			// If it's a different user account, the current user needs manager access.
			throw new AccessDeniedException();
		}

		if(!$user->isActive()) {
			if($isManager) {
				\Core\set_message('This account is not active!', 'warning');
			}
			else {
				throw new UserNotActivatedException();
			}
		}

		if($request->isJSON()){
			$view->mode = View::MODE_PAGEORAJAX;
			$view->contenttype = View::CTYPE_JSON;

			$view->jsondata['user'] = $user->asArray();

			// Massage some user data a bit, and remove things that don't need to be exported.
			unset($view->jsondata['user']['password']);

			return;
		}
		
		// Pull the user groups for this user to display on the page.
		$groupIDs = $user->getGroups();
		$groups = [];
		foreach($groupIDs as $gid){
			$groups[] = UserGroupModel::Construct($gid);
		}
		
		// Retrieve all the attributes for this User
		$attributes = [];
		$schema = $user->getColumns();
		foreach($schema as $s) {
			if($s->formatter !== null) {
				$attributes[ $s->field ] = $s;
			}
		}

		// Retrieve a list of current user sessions and saved device logins.
		$sessions = SessionModel::Find(['user_id = ' . $user->id], null, 'updated DESC');
		$devices = UserDeviceModel::Find(['user = ' . $user->id], null, 'updated DESC');

		$view->controls = ViewControls::DispatchModel($user);

		$view->templatename = 'pages/user/view.tpl';

		if($isManager) {
			$view->mastertemplate = 'admin';
			$view->addBreadcrumb('User Administration', '/user/admin');
			$view->title = t('STRING_USER_PROFILE_S', $user->getDisplayName());
		}
		else {
			$view->mastertemplate = 'user';
			$view->title = t('STRING_MY_PROFILE');
		}
		$view->assign('user', $user);
		$view->assign('profiles', $user->get('external_profiles'));
		$view->assign('groups', $groups);
		$view->assign('attributes', $attributes);
		$view->assign('sessions', $sessions);
		$view->assign('devices', $devices);
		$view->assign('title', $view->title);
	}


	/**
	 * View to edit the user account, both administratively and from within the user's profile.
	 */
	public function edit(){

		$view          = $this->getView();
		$req           = $this->getRequest();
		$userid        = $req->getParameter(0);
		$manager       = \Core\user()->checkAccess('p:/user/users/manage'); // Current user an admin?
		$groupmanager  = \Core\user()->checkAccess('p:/user/groups/manage');
		$contextnames  = [];
		$contexts      = [];
		$usecontexts   = false;

		if($userid === null){
			// Default to current user.
			$userid = \Core\user()->get('id'); 
		}

		// Only allow this if the user is either the same user or has the user manage permission.
		if($userid !== \Core\user()->get('id') && !$manager){
			throw new AccessDeniedException();
		}

		/** @var UserModel $user */
		$user = UserModel::Construct($userid);
		if(!$user->checkExists()){
			throw new \Core\NotFoundException();
		}
		
		$form = new Form();
		$form->set('callsmethod', 'Core\\User\\Helper::UpdateHandler');
		$form->addModel($user, 'user');

		if($groupmanager){
			$contextgroups = UserGroupModel::Find(['context != '], null, 'name');
			foreach($contextgroups as $group){
				/** @var UserGroupModel $group */

				$ckey = $group->get('context');
				$gkey = $group->get('id');
				$contextnames[ $group->get('name') ] = $gkey;

				// I need to load *all* those models into the system so they're available to the UI.
				$fac = new ModelFactory($ckey . 'Model');
				$all = [];
				foreach($fac->get() as $m){
					/** @var Model $m */
					$all[$m->getPrimaryKeyString()] = $m->getLabel();
				}
				$contexts[$gkey] = $all;

				$usecontexts = true;
			}
		}
		
		$form->addElement('submit', ['value' => 't:STRING_SAVE']);



		$view->controls = ViewControls::DispatchModel($user);

		self::SetBreadcrumbsUserPage($view, $user);
		$view->title = 't:STRING_EDIT';

		$view->assign('form', $form);
		$view->assign('contextnames_json', json_encode($contextnames));
		$view->assign('contextnames', $contextnames);
		$view->assign('contexts_json', json_encode($contexts));
		$view->assign('use_contexts', $usecontexts);
		$view->assign('user', $user);

		// Breadcrumbs! (based on access permissions)
		if($manager){
			//$view->addBreadcrumb('User Administration', '/user/admin');
		}
	}

	/**
	 * Function to edit the user's connected profiles.
	 */
	public function connectedprofiles(){
		$view    = $this->getView();
		$req     = $this->getRequest();
		$userid  = $req->getParameter(0);
		$manager = \Core\user()->checkAccess('p:/user/users/manage'); // Current user an admin?

		if($userid === null) $userid = \Core\user()->get('id'); // Default to current user.

		// Only allow this if the user is either the same user or has the user manage permission.
		if(!($userid == \Core\user()->get('id') || $manager)){
			return View::STATUS_ACCESSDENIED;
		}

		/** @var UserModel $user */
		$user = UserModel::Construct($userid);
		// I will be dealing with only one custom field...
		$profiles = $user->get('external_profiles');

		if($req->isPost()){
			// Update the new profiles.... yay
			$error = false;
			$profiles = array();
			foreach($_POST['type'] as $k => $v){
				// Check that this looks like a URL.
				if(!(
					// Links should be a valid URL (starting with HTTP or HTTPS).
					preg_match(Model::VALIDATION_URL_WEB, $_POST['url'][$k]) ||
					// Or twitter-style links using the '@' prefix.
					$_POST['url'][$k][0] == '@'
				)){
					$error = true;
					\Core\set_message($_POST['url'][$k] . ' does not appear to be a valid URL!  Please ensure that it starts with http:// or https://', 'error');
				}

				$profiles[] = array(
					'type' => $_POST['type'][$k],
					'url' => $_POST['url'][$k],
					'title' => $_POST['title'][$k],
				);
			}

			if(!$error){
				$user->set('external_profiles', $profiles);
				$user->save();
				\Core\set_message('Updated profiles successfully', 'success');
				\Core\go_back();
			}
		}

		if($manager){
			$view->addBreadcrumb($user->getDisplayName(), '/user/view/' . $user->get('id'));
		}
		else{
			$view->addBreadcrumb($user->getDisplayName(), '/user/me');
		}
		$view->title = 'Edit Connected Profiles';
		$view->assign('profiles_json', json_encode($profiles));
		$view->assign('profiles', $profiles);
	}

	/**
	 * Display the login page for whatever drivers may happen to be installed.
	 *
	 * @return View
	 */
	public function login(){
		$request = $this->getRequest();
		$view    = $this->getView();

		// Is the user already logged in?
		// Set the access permissions for this page as anonymous-only.
		if(\Core\user()->exists()){
			\core\redirect('/user/me');
		}

		$auths = Helper::GetEnabledAuthDrivers();

		$view->ssl = true;
		$view->mastertemplate = 'user-login';
		$view->assign('drivers', $auths);
		$view->assign('allowregister', ConfigHandler::Get('/user/register/allowpublic'));
		// Google has no business indexing user-action pages.
		$view->addMetaName('robots', 'noindex');
		return $view;
	}

	/**
	 * Ajax page to allow for quickly linking the current user to a facebook account from a strictly javascript interface.
	 *
	 */
	public function linkfacebook(){
		$request = $this->getRequest();
		$view    = $this->getView();

		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;

		if(!$request->isAjax()){
			return View::STATUS_BADREQUEST;
		}
		if(!$request->isPost()){
			return View::STATUS_BADREQUEST;
		}

		$user = \Core\user();
		if(!$user->exists()){
			return View::STATUS_BADREQUEST;
		}

		// Necessary fields are id and token.
		$user->set('facebook_access_token', $_POST['token']);
		$user->set('facebook_id', $_POST['id']);
		$user->save();

		$view->jsondata = array(
			'id' => $user->get('id'),
			'access_token' => $user->get('facebook_access_token'),
			'facebook_id' => $user->get('facebook_id'),
		);
	}

	/**
	 * Display the register page for new users.
	 *
	 * @return int
	 */
	public function register(){

		$view          = $this->getView();
		$manager       = \Core\user()->checkAccess('p:/user/users/manage'); // Current user an admin?

		// Anonymous users should have access to this if it's allow public.
		if(!\Core\user()->exists() && !ConfigHandler::Get('/user/register/allowpublic')){
			return View::STATUS_BADREQUEST;
		}

		// Authenticated users must check the permission to manage users.
		if(\Core\user()->exists() && !$manager){
			return View::STATUS_ACCESSDENIED;
		}

		$auths = Helper::GetEnabledAuthDrivers();

		$view->title = 'Register';
		$view->ssl = true;
		$view->assign('drivers', $auths);
		// Google has no business indexing user-action pages.
		$view->addMetaName('robots', 'noindex');

		// Breadcrumbs! (based on access permissions)
		if($manager){
			$view->addBreadcrumb('User Administration', '/user/admin');
		}
		else {
			$view->mastertemplate = 'user-login';
		}
	}

	/**
	 * The actual Core registration page.
	 *
	 * This renders all the user's configurable options at registration.
	 */
	public function register2(){
		$view    = $this->getView();
		$request = $this->getRequest();
		$manager = \Core\user()->checkAccess('p:/user/users/manage'); // Current user an admin?

		// Anonymous users should have access to this if it's allow public.
		if(!\Core\user()->checkExists() && !ConfigHandler::Get('/user/register/allowpublic')){
			return View::STATUS_BADREQUEST;
		}

		// Authenticated users must check the permission to manage users.
		if(\Core\user()->checkExists() && !$manager){
			return View::STATUS_ACCESSDENIED;
		}

		/** @var NonceModel $nonce */
		$nonce = NonceModel::Construct($request->getParameter(0));
		if(!$nonce->isValid()){
			\Core\set_message('Invalid nonce token, please try again.', 'error');
			\Core\go_back();
		}
		$nonce->decryptData();
		$data = $nonce->get('data');

		if(!isset($data['user']) || !($data['user'] instanceof UserModel)){
			if(DEVELOPMENT_MODE){
				\Core\set_message('Your nonce does not include a "user" key.  Please ensure that this is set to a non-existent UserModel object!', 'error');
			}
			else{
				\Core\set_message('Invalid login type, please try again later.', 'error');
			}
			\Core\go_back();
		}

		/** @var UserModel $user */
		$user = $data['user'];

		$form = new Form();
		$form->set('callsmethod', 'Core\\User\\Helper::RegisterHandler');
		$form->addModel($user, 'user');
		
		// If the total number of form elements here are only 2, then only the user object and submit button are present.
		// Instead of showing the form, auto-submit to that destination.
		if(sizeof($form->getElements()) <= 2){
			$user->setDefaultGroups();
			$user->setDefaultMetaFields();
			$user->setDefaultActiveStatuses();
			$user->generateNewApiKey();
			$user->save();

			// User created... make a log of this!
			SystemLogModel::LogSecurityEvent('/user/register', 'User registration successful', null, $user->get('id'));

			// Send a thank you for registering email to the user.
			try{
				$user->sendWelcomeEmail();
			}
			catch(Exception $e){
				\Core\log_exception($e);
				\Core\set_message('t:MESSAGE_ERROR_CANNOT_SEND_WELCOME_EMAIL');
			}

			// "login" this user if not already logged in.
			if(!\Core\user()->checkExists()){
				if($user->get('active')){
					$user->set('last_login', DateTime::Now('U', Timezone::TIMEZONE_GMT));
					$user->save();
					Session::SetUser($user);
				}
				\Core\set_message('t:MESSAGE_SUCCESS_CREATED_USER_ACCOUNT');

				if(($overrideurl = HookHandler::DispatchHook('/user/postlogin/getredirecturl'))){
					// Allow an external script to override the redirecting URL.
					$url = $overrideurl;
				}
				elseif($form->getElementValue('redirect')){
					// The preferred default redirect method.
					// This is set from /user/register2, which is in turn passed in, (hopefully), by the original callee registration page.
					$url = $form->getElementValue('redirect');
				}
				elseif(strpos(REL_REQUEST_PATH, '/user/register') === 0){
					// If the user came from the registration page, get the page before that.
					$url = '/';
				}
				else{
					// else the registration link is now on the same page as the 403 handler.
					$url = REL_REQUEST_PATH;
				}

				\Core\redirect($url);
			}
			// It was created administratively; redirect there instead.
			else{
				\Core\set_message('t:MESSAGE_SUCCESS_CREATED_USER_ACCOUNT');
				\Core\redirect('/user/admin');
			}
		}
		
		$form->addElement('system', ['name' => 'redirect', 'value' => $data['redirect']]);
		$form->addElement('submit', ['value' => 'Register']);

		$view->title = 'Complete Registration';
		$view->assign('form', $form);
	}

	public function logout(){
		// I need to access *something* in the Session in order to get the session ID,
		// otherwise Session::DestroySession() doesn't do anything because no session is started yet.
		\Core\user();

		// Kill the device cookie too!
		$device = UserDeviceModel::GetCurrentDevice();
		if($device) {
			$device->delete();
		}
		UserDeviceModel::UnsetCookie();

		Session::DestroySession();

		\core\redirect('/');
	}



	/**
	 * Simple controller to activate a user account.
	 * Meant to be called with json only.
	 */
	public function activate(){
		$req    = $this->getRequest();
		$view   = $this->getView();
		$userid = $req->getPost('user') ? $req->getPost('user') : $req->getParameter('user');
		$active = ($req->getPost('status') !== null) ? $req->getPost('status') : $req->getParameter('status');
		if($active === '') $active = 1; // default.


		if(!\Core\user()->checkAccess('p:/user/users/manage')){
			return View::STATUS_ACCESSDENIED;
		}

		if(!$req->isPost()){
			return View::STATUS_BADREQUEST;
		}

		if(!$userid){
			return View::STATUS_BADREQUEST;
		}


		$user = UserModel::Construct($userid);

		if(!$user->exists()){
			return View::STATUS_NOTFOUND;
		}

		$user->set('active', $active);
		$user->save();

		// Send an activation notice email to the user if the active flag is set to true.
		if($active){
			try{
				$email = Email::Factory('/emails/user/activation.tpl');

				if(!$user->get('password')){
					// Generate a Nonce for this user with the password reset.
					// Use the Nonce system to generate a one-time key with this user's data.
					$nonce = NonceModel::Generate(
						'1 week',
						['type' => 'password-reset', 'user' => $user->get('id')]
					);
					$setpasswordlink = \Core\resolve_link('/datastoreauth/forgotpassword?e=' . urlencode($user->get('email')) . '&n=' . $nonce);
				}
				else{
					$setpasswordlink = null;
				}

				$email->assign('user', $user);
				$email->assign('sitename', SITENAME);
				$email->assign('rooturl', ROOT_URL);
				$email->assign('loginurl', \Core\resolve_link('/user/login'));
				$email->assign('setpasswordlink', $setpasswordlink);
				$email->setSubject('Welcome to ' . SITENAME);
				$email->setTo($user->get('email'));

				// TESTING
				//error_log($email->renderBody());
				$email->send();
			}
			catch(Exception $e){
				\Core\ErrorManagement\exception_handler($e);
			}
		}

		if($req->isJSON()){
			$view->mode = View::MODE_AJAX;
			$view->contenttype = View::CTYPE_JSON;

			$view->jsondata = array(
				'userid' => $user->get('id'),
				'active' => $user->get('active'),
			);
		}
		else{
			\Core\go_back();
		}
	}

	/**
	 * Permanently delete a user account and all configuration options attached.
	 *
	 * @return int
	 */
	public function delete(){
		$view  = $this->getView();
		$req   = $this->getRequest();
		$id    = $req->getParameter(0);
		$model = UserModel::Construct($id);

		if(!\Core\user()->checkAccess('p:/user/users/manage')){
			return View::STATUS_ACCESSDENIED;
		}

		if(!$req->isPost()){
			return View::STATUS_BADREQUEST;
		}

		// Users are now a standard model, deleting a user account will automatically propagate down the stack.
		$model->delete();
		\Core\set_message('Removed user successfully', 'success');
		\Core\go_back();
	}

	/**
	 * View to sudo as another user.
	 */
	public function sudo(){
		$view  = $this->getView();
		$req   = $this->getRequest();
		$id    = $req->getParameter(0);


		if($id){
			$model = UserModel::Construct($id);

			if(!\Core\user()->checkAccess('p:/user/users/sudo')){
				return View::STATUS_ACCESSDENIED;
			}

			if(!$req->isPost()){
				return View::STATUS_BADREQUEST;
			}

			if(!$model->exists()){
				return View::STATUS_NOTFOUND;
			}

			Session::Set('user_sudo', $model);
		}
		elseif(Session::Get('user_sudo') !== null){
			Session::UnsetKey('user_sudo');
		}

		\Core\redirect('/');
	}

	/**
	 * Import a set of users from a CSV file.
	 */
	public function import(){
		$view = $this->getView();

		if(!\Core\user()->checkAccess('p:/user/users/manage')){
			throw new AccessDeniedException();
		}

		$view->addBreadcrumb('User Administration', '/user/admin');
		$view->title = 'Import Users';

		if(Session::Get('user-import/log') !== null){
			// Counts array is present... show the results page.
			$this->_import3();
		}
		elseif(Session::Get('user-import/file') !== null && file_exists(Session::Get('user-import/file'))){
			// The file is set, that's step two.
			$this->_import2();
		}
		else{
			$this->_import1();
		}
	}

	/**
	 * Link to abort the import process.
	 */
	public function import_cancel(){
		Session::UnsetKey('user-import/*');
		\core\redirect('/user/import');
	}

	/**
	 * User "Dashboard" page, similar to the admin dashboard, but suitable for user-level operations.
	 *
	 * @core-access-string g:authenticated
	 */
	public function dashboard() {
		$view = $this->getView();

		$view->mastertemplate = 'user';
		$view->title = 't:STRING_USER_DASHBOARD';
	}

	/**
	 * Send the welcome email to a user
	 *
	 * @core-request-method POST
	 * @core-access-string p:/user/users/manage
	 */
	public function sendwelcome() {
		$request = $this->getRequest();
		$userID  = $request->getParameter(0);
		$user = UserModel::Construct($userID);

		if(!$user->checkExists()) {
			throw new NotFoundException();
		}

		if($user->active !== '1') {
			throw new NotFoundException();
		}
		
		try{
			$email = Email::Factory('emails/user/activation');
			$email->assign('user', $user);
			$email->assign('sitename', SITENAME);
			$email->assign('rooturl', ROOT_URL);
			$email->assign('loginurl', \Core\resolve_link('/user/login'));
			$email->setSubject('Welcome to ' . SITENAME);
			$email->setTo($user->get('email'));

			// TESTING
			//error_log($email->renderBody());
			$email->send();
		}
		catch(Exception $e){
			\Core\ErrorManagement\exception_handler($e);
		}
	}

	/**
	 * Display the initial upload option that will kick off the rest of the import options.
	 */
	private function _import1(){
		$view = $this->getView();
		$request = $this->getRequest();

		$form = new Form();
		$form->set('callsmethod', 'Core\\User\\ImportHelper::FormHandler1');
		$form->addElement(
			'file',
			[
				'name' => 'file',
				'title' => 'File To Import',
				'basedir' => 'tmp/user-import',
				'required' => true,
				'accept' => '.csv',
			]
		);
		$form->addElement('submit', ['name' => 'submit', 'value' => 'Next']);


		$view->templatename = '/pages/user/import1.tpl';
		$view->assign('form', $form);
	}

	/**
	 * There has been a file selected; check that file for headers and what not to display something useful to the user.
	 */
	private function _import2(){
		$view = $this->getView();
		$request = $this->getRequest();

		$filename = Session::Get('user-import/file');
		$file = Factory::File($filename);
		$contents = $file->getContentsObject();

		if(!$contents instanceof \Core\Filestore\Contents\ContentCSV){
			\Core\set_message($file->getBaseFilename() . ' does not appear to be a valid CSV file!', 'error');
			Session::UnsetKey('user-import/file');
			\Core\reload();
		}

		$hasheader = $contents->hasHeader();
		$data = $contents->parse();

		// Since I don't want to display the entire dataset in the preview...
		if($hasheader){
			$header = $contents->getHeader();
		}
		else{
			$header = array();
			$i=0;
			foreach($data[0] as $k => $v){
				$header[$i] = 'Column ' . ($i+1);
				$i++;
			}
		}

		$form = new Form();
		$form->set('callsmethod', 'Core\\User\\ImportHelper::FormHandler2');
		$form->addElement('system', ['name' => 'key', 'value' => Session::Get('user-import/key')]);
		$form->addElement(
			'checkbox',
			[
				'name' => 'has_header',
				'title' => 'Has Header',
				'value' => 1,
				'checked' => $hasheader,
				'description' => 'If this CSV has a header record on line 1, check this to ignore that line.'
			]
		);

		$form->addElement(
			'checkbox',
			[
				'name' => 'merge_duplicates',
				'title' => 'Merge Duplicate Records',
				'value' => 1,
				'checked' => true,
				'description' => 'Merge duplicate records that may be found in the import.'
			]
		);

		// Only display the user groups if the current user has access to manage user groups.
		$usergroups = UserGroupModel::Find(['context = ']);
		if(sizeof($usergroups) && \Core\user()->checkAccess('p:/user/groups/manage')){
			$usergroupopts = array();
			foreach($usergroups as $ug){
				$usergroupopts[$ug->get('id')] = $ug->get('name');
			}
			$form->addElement(
				'checkboxes',
				[
					'name' => 'groups[]',
					'title' => 'User Groups to Assign',
					'options' => $usergroupopts,
					'description' => 'Check which groups to set the imported users to.  If merge duplicate records is selected, any found users will be set to the checked groups, (and consequently unset from any unchecked groups).',
				]
			);
		}
		else{
			$form->addElement('hidden', ['name' => 'groups[]', 'value' => '']);
		}
		
		// Build the list of available columns.
		$schema = UserModel::GetSchema();
		$maptos = new Dict();
		$maptos->autoI18N = true;
		$maptos[''] = '-- Do Not Map --';

		foreach($schema->definitions as $s) {
			if($s->importer !== false) {
				$maptos[ $s->field ] = $s->getTitle();
			}
		}
		foreach($schema->metas as $s) {
			if($s->importer !== false) {
				$maptos[ $s->field ] = $s->getTitle();
			}
		}

		foreach($header as $key => $title){
			$value = '';
			if(isset($maptos[$key])) {
				// This header exists as a key
				$value = $key;
			}
			elseif(($match = $maptos->search($title)) !== null) {
				$value = $match;
			}

			$form->addElement(
				'select',
				[
					'name' => 'mapto[' . $key . ']',
					'title' => $title,
					'options' => $maptos,
					'value' => $value
				]
			);
		}


		$view->templatename = '/pages/user/import2.tpl';
		$view->assign('has_header', $hasheader);
		$view->assign('form', $form);
	}

	private function _import3(){
		$view = $this->getView();
		$request = $this->getRequest();

		$view->templatename = '/pages/user/import3.tpl';
		$view->assign('log', Session::Get('user-import/log'));

		Session::UnsetKey('user-import/*');
	}



	/**
	 * This is a helper controller to expose server-side data to javascript.
	 *
	 * It's useful for currently logged in user and what not.
	 * Obviously nothing critical is exposed here, since it'll be sent to the useragent.
	 */
	public function jshelper(){
		$request = $this->getRequest();

		// This is a json-only page.
		if($request->ctype != View::CTYPE_JSON){
			\core\redirect('/');
		}

		// The data that will be returned.
		$data = array();

		$cu = Core::User();

		if(!$cu->exists()){
			$data['user'] = array(
				'id' => null,
				'displayname' => ConfigHandler::Get('/user/displayname/anonymous'),
				//'email' => null,
			);
			$data['accessstringtemplate'] = null;
		}
		else{
			$data['user'] = array(
				'id' => $cu->get('id'),
				'displayname' => $cu->getDisplayName(),
				//'email' => $cu->get('email'),
			);

			// Templated version of the access string form system, useful for dynamic permissions on the page.
			$templateel = new AccessStringInput(array(
				'title' => '##TITLE##',
				'name' => '##NAME##',
				'description' => '##DESCRIPTION##',
				'class' => '##CLASS##',
				'value' => 'none'
			));
			$data['accessstringtemplate'] = $templateel->render();
		}

		$this->getView()->jsondata = $data;
		$this->getView()->contenttype = View::CTYPE_JSON;
	}


	/**
	 * Hook to catch 403 pages; renders the login or register forms for guest users
	 *
	 * @param View $view
	 * @return bool|View
	 * @throws Exception
	 */
	public static function _HookHandler403(View $view){

		if(EXEC_MODE !== 'WEB'){
			// This only applies to pages in WEB mode.
			return true;
		}
		
		if(\Core\user()->checkExists()) {
			// User is already logged in... I can't do anything.
			return true;
		}

		// I need to replace the current page with this one so that the previous controller never executes.
		$newcontroller = new self();
		$newcontroller->overwriteView($view);
		//$view->baseurl = '/user/login';

		$auths = Helper::GetEnabledAuthDrivers();

		$view->title = 't:STRING_AUTHENTICATION_REQUIRED';
		$view->ssl = true;
		$view->status = View::STATUS_ACCESSDENIED;
		$view->allowerrors = true;
		$view->mastertemplate = 'user-login';
		$view->templatename = 'pages/user/guest403';
		$view->assign('drivers', $auths);
		$view->assign('allowregister', ConfigHandler::Get('/user/register/allowpublic'));
		// Google has no business indexing user-action pages.
		$view->addMetaName('robots', 'noindex');

		$loginform = new Form();
		$loginform->set('callsMethod', 'Core\\User\\Helper::LoginHandler');

		$loginform->addElement('text', array('name' => 'email', 'title' => 'Email', 'required' => true));
		$loginform->addElement('password', array('name' => 'pass', 'title' => 'Password', 'required' => true));
		$loginform->addElement('submit', array('value' => 'Login'));

		if(ConfigHandler::Get('/user/register/allowpublic')){
			$registerform = Helper::GetRegistrationForm();
		}
		else{
			$registerform = null;
		}


		$error = false;



		$view->assign('error', $error);
		$view->assign('backends', ConfigHandler::Get('/user/backends'));
		$view->assign('loginform', $loginform);
		$view->assign('registerform', $registerform);
		$view->assign('allowregister', ConfigHandler::Get('/user/register/allowpublic'));


		return $view;
	}

	public static function SetBreadcrumbsUserPage(View $view, UserModel $user = null) {
		if($user === null) {
			$user = \Core\user();
		}

		if(\Core\user()->checkAccess('p:/user/users/manage')) {
			$view->mastertemplate = 'admin';
			$view->addBreadcrumb('User Administration', '/user/admin');
			$view->addBreadcrumb(t('STRING_USER_PROFILE_S', $user->getDisplayName()), '/user/view/' . $user->id);
		}
		else {
			$view->addBreadcrumb('t:STRING_MY_PROFILE', '/user/view/' . $user->id);
		}
	}

}


