<?php
/**
 * Main loading file for the Caching system.
 * 
 * @package Core
 * @since 1.9
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;

/**
 * Class Cache
 * This is the main public interface that will be used by external scripts.
 *
 * @package Core
 */
class Cache {
	/**
	 * @var null|string The backend type
	 */
	private static $_Backend = null;

	/**
	 * Get a cached value based on its key
	 *
	 * @param string $key
	 *
	 * @return mixed
	 */
	public static function Get(string $key){
		return self::_Factory()->read($key);
	}

	/**
	 * Set a cached value on a given key
	 *
	 * @param string $key
	 * @param mixed  $value
	 * @param int    $expires
	 *
	 * @return bool
	 */
	public static function Set(string $key, $value, int $expires = 7200) {
		return self::_Factory()->write($key, $value, $expires);
	}

	/**
	 * Delete a given key from the system cache
	 *
	 * @param $key
	 *
	 * @return bool
	 */
	public static function Delete(string $key){
		return self::_Factory()->delete($key);
	}

	/**
	 * Flush the entire system cache
	 *
	 * @return bool
	 */
	public static function Flush(){
		return self::_Factory()->flush();
	}
	
	public static function Ls(){
		return self::_Factory()->ls();
	}

	/**
	 * Factory to retrieve the current system-default Caching system
	 * 
	 * @return Cache\CacheInterface
	 */
	private static function _Factory(){

		if(self::$_Backend === null){
			// Load the backend from the site configuration.xml
			$cs = ConfigHandler::LoadConfigFile("configuration");
			$type = $cs['cache_type'];
			
			switch($type){
				case 'memcache':
				case 'memcached':
					self::$_Backend = new Cache\Memcache();
					break;
				case 'file':
				default:
					self::$_Backend = new Cache\File();
					break;
			}
			
			// Make sure it's ready and fallback to File if necessary.
			if(!self::$_Backend->isReady()){
				self::$_Backend = new Cache\File();
				self::$_Backend->isReady();
			}
		}
		
		return self::$_Backend;
	}
}
