<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core;
use Exception;

/**
 * Description of ComponentRequirement
 *
 * @author charlie
 */
class ComponentRequirement {
	private $_type;
	private $_name;
	private $_version;
	private $_operation;

	const CHECK_TYPE_NATIVEBINARY = 'binary';
	const CHECK_TYPE_COMPONENT    = 'component';
	const CHECK_TYPE_FUNCTION     = 'function';
	const CHECK_TYPE_JSLIBRARY    = 'jslibrary';
	const CHECK_TYPE_LIBRARY      = 'library';
	const CHECK_TYPE_PHPEXT       = 'phpextension';
	
	public function __construct($type, $name, $version = null, $operation = null){
		$this->_type = $type;
		$this->_name = $name;
		$this->_version = $version;
		$this->_operation = $operation;
	}
	
	/**
	 * Check and see if this requirement is met by the current conditions of the system
	 * 
	 * @throws Exception Check failed due to unmet dependencies, reason will be in the message
	 * 
	 * @return bool Return will be TRUE on successful check and FALSE if the TYPE is unknown
	 */
	public function check(){
		switch($this->_type){
			case self::CHECK_TYPE_NATIVEBINARY:
				$this->_checkNativeBinary();
				break;

			case self::CHECK_TYPE_COMPONENT:
				$this->_checkComponent();
				break;
			
			case self::CHECK_TYPE_FUNCTION:
				$this->_checkFunction();
				break;
			
			case self::CHECK_TYPE_JSLIBRARY:
				$this->_checkJSLibrary();
				break;
				
			case self::CHECK_TYPE_LIBRARY:
				$this->_checkLibrary();
				break;
				
			case self::CHECK_TYPE_PHPEXT:
				$this->_checkPHPExtension();
				break;
				
			default:
				return false;
		}
		
		return true;
	}
	
	public function getName() : string {
		return $this->_name;
	}
	
	/**
	 * @return string Type of the check, component, library, phpextension, etc.
	 */
	public function getType() : string {
		return $this->_type;
	}

	public function getVersion() : ?string {
		return $this->_version;
	}

	public function getOperation() : ?string {
		return $this->_operation;
	}
	
	public function getLabel(){
		$v = ($this->_version && $this->_operation) ? ' ' . $this->_getVersionOperationString() : '';
			
		switch($this->_type){
			case self::CHECK_TYPE_NATIVEBINARY:
				return 'Native System Binary ' . $this->_name . ' required';

			case self::CHECK_TYPE_COMPONENT:
				return 'Core component ' . $this->_name . $v . ' required';
			
			case self::CHECK_TYPE_FUNCTION:
				return 'PHP function ' . $this->_name . ' required';
			
			case self::CHECK_TYPE_JSLIBRARY:
				return 'Core javascript library ' . $this->_name . $v . ' required';
				
			case self::CHECK_TYPE_LIBRARY:
				return 'Core library ' . $this->_name . $v . ' required';
				
			case self::CHECK_TYPE_PHPEXT:
				return 'PHP function ' . $this->_name . $v . ' required';
				
			default:
				return 'Unknown check';
		}
	}

	/**
	 * Execute the check to look for a native system binary
	 *
	 * @throws Exception
	 */
	private function _checkNativeBinary() {
		$check = exec('which ' . $this->_name . ' 2>/dev/null');
		if($check == '') {
			throw new Exception('Native binary ' . $this->_name . ' is missing, please install it');
		}
	}
	
	/**
	 * Execute the checks for a specific Core component present
	 * 
	 * @throws Exception
	 */
	private function _checkComponent(){
		if (!Core::IsComponentAvailable($this->_name)) {
			throw new Exception('Core component ' . $this->_name . ' not installed');
		}
		
		if(
			$this->_version && 
			$this->_operation && 
			!Core::IsComponentAvailable($this->_name, $this->_version, $this->_operation)
		){
			throw new Exception('Core component ' . $this->_name . ' ' . $this->_getVersionOperationString() . ' required');
		}
	}
	
	/**
	 * Execute the check for the presence of a PHP function, probably for extended functionality like LDAP.
	 * 
	 * @throws Exception
	 */
	private function _checkFunction(){
		if(!function_exists($this->_name)){
			throw new Exception('Missing PHP function ' . $this->_name);
		}
	}
	
	private function _checkJSLibrary(){
		if (!Core::IsJSLibraryAvailable($this->_name)) {
			throw new Exception('Core JS library ' . $this->_name . ' not installed');
		}
	}
	
	private function _checkLibrary(){
		if (!Core::IsLibraryAvailable($this->_name)) {
			throw new Exception('Core library ' . $this->_name . ' not installed');
		}
		
		if(
			$this->_version && 
			$this->_operation && 
			!Core::IsLibraryAvailable($this->_name, $this->_version, $this->_operation)
		){
			throw new Exception('Core library ' . $this->_name . ' ' . $this->_getVersionOperationString() . ' required');
		}
	}
	
	private function _checkPHPExtension(){
		if(extension_loaded($this->_name) === false){
			throw new Exception('Missing PHP Extension ' . $this->_name);
		}
		
		if(
			$this->_version && 
			$this->_operation && 
			!version_compare(phpversion($this->_name), $this->_version, $this->_operation)
		){
			throw new Exception('PHP Extension ' . $this->_name . ' ' . $this->_getVersionOperationString() . ' required');
		}
	}
	
	private function _getVersionOperationString(){
		$opMap = [
			'>=' => 'at least',
			'ge' => 'at least',
			'>'  => 'newer than',
			'gt' => 'newer than',
			'<=' => 'at most',
			'le' => 'at most',
			'<'  => 'older than',
			'lt' => 'older than',
			'='  => 'at',
			'eq' => 'at',
		];
		
		return (isset($opMap[ $this->_operation ]) ? $opMap[ $this->_operation ] : $this->_operation) . ' version ' . $this->_version;
	}
}
