<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core;

use ReflectionMethod;
use ReflectionParameter;

/**
 * Description of ControllerMethodDescriber
 *
 * @author charlie
 */
class ControllerMethodDescriber {
	public $title;
	public $description        = '';
	public $requestMethod      = '*';
	public $requestMode        = '*';
	public $accessString       = '';
	public $execMode           = 'WEB';
	public $parameters         = [];
	public $execDevEnvironment = false;
	
	public function __construct(ReflectionMethod $method) {
		
		foreach($method->getParameters() as $p){
			$this->_parseParameter($p);
		}
		
		// Scan the parent controller documentation for hints
		$parentDoc = $method->getDeclaringClass()->getDocComment();
		if($parentDoc){
			$this->_parseDoc($parentDoc);
		}
		
		// Scan the method's documentation for hints
		$doc = $method->getDocComment();
		if($doc){
			$this->_parseDoc($doc, true);
		}
	}
	
	/**
	 * Check if this Method can be accessed by the current user/request.
	 * 
	 * Will return true if everything is good, and an exception otherwise.
	 * 
	 * @throws BadRequestException
	 * @throws AccessDeniedException
	 * 
	 * @return bool
	 */
	public function checkCanAccess() {
		
		$request = Router::$DefaultRequest;
		
		if($this->requestMethod !== '*') {
			// Check for the request method, usually @core-request-method POST, but could be any available verb.
			$methods = array_map('trim', explode('|', strtoupper($this->requestMethod)));
			if(!in_array($request->method, $methods)) {
				throw new BadRequestException('Page does not accept ' . $request->method . ' methods.');
			}
		}

		if($this->requestMode !== '*') {
			// Check for the request mode, @core-request-mode AJAX | PAGE.
			// (currently only those two are supported).
			$modes = array_map('trim', explode('|', strtoupper($this->requestMode)));
			$mode = $request->isAjax() ? 'AJAX' : 'PAGE';
			if(!in_array($mode, $modes)) {
				throw new BadRequestException('Page does not accept ' . $mode . ' requests.');
			}
		}
		
		if(
			$this->execMode !== 'BOTH' &&
			$this->execMode !== EXEC_MODE
		){
			// Allows scripts to be WEB, CLI, or both.
			throw new BadRequestException('Page does not accept ' . EXEC_MODE . ' connections.');
		}
		
		if(
			$this->accessString &&
			!\Core\user()->checkAccess($this->accessString)
		){
			// Basic user role check
			throw new AccessDeniedException();
		}
		
		if(
			$this->execDevEnvironment &&
			!DEVELOPMENT_MODE
		) {
			// Restrict anything listed as @core-development-mode to DEV_MODE ONLY!
			// These are usually tools and pages useful only for developers.
			throw new AccessDeniedException('This page is a development-only page!  To access it, please set DEVELOPMENT_MODE to true in your configuration.');
		}
		
		return true;
	}
	
	private function _parseDoc(string $doc, bool $parseText = false) {
		// Scan the method's phpdoc for hints about how it should be used
		// and other developer-set hints for security and sanitization.
		$lines = array_map('trim', explode("\n", $doc));
		
		$paramsStarted = false;
		foreach($lines as $line){
			if(strpos($line, '/*') !== false){
				// Skip the header
				continue;
			}
			elseif(strpos($line, '* @core-request-method ') === 0){
				$paramsStarted = true;
				$this->requestMethod = substr($line, 23);
			}
			elseif(strpos($line, '* @core-request-mode ') === 0){
				$paramsStarted = true;
				$this->requestMode = substr($line, 21);
			}
			elseif(strpos($line, '* @core-access-string ') === 0){
				$paramsStarted = true;
				$this->accessString = substr($line, 22);
			}
			elseif(strpos($line, '* @core-exec-mode ') === 0){
				$paramsStarted = true;
				$this->execMode = substr($line, 18);
			}
			elseif(strpos($line, '* @core-development-mode') === 0) {
				$paramsStarted = true;
				$this->execDevEnvironment = true;
			}
			elseif(strpos($line, '* @param') === 0){
				$paramsStarted = true;
				$this->_parseParameterDoc($line);
			}
			elseif(strpos($line, '* @') === 0){
				$paramsStarted = true;
			}
			elseif($parseText && !$paramsStarted && !$this->title && strlen($line) > 2){
				$this->title = substr($line, 2);
			}
			elseif($parseText && !$paramsStarted){
				$this->description .= substr($line, 2) . "\n";
			}
		}
		
		$this->description = trim($this->description);
	}
	
	private function _parseParameterDoc($line){
		$param = preg_replace('/.*\$([a-zA-Z0-9_]*).*/', '\1', $line);
		$desc  = trim(substr($line, strpos($line, '$' . $param) + strlen($param) + 1));
		foreach($this->parameters as $idx => $dat){
			if($dat['name'] === $param){
				$this->parameters[ $idx ]['description'] = $desc;
			}
		}
	}
	
	private function _parseParameter(ReflectionParameter $param){
		$type     = $param->hasType() ? $param->getType()->getName() : null;
		$required = !$param->isOptional();
		
		$this->parameters[] = [
			'name'        => $param->name,
			'type'        => $type,
			'required'    => $required,
			'description' => '',
		];
	}
}
