<?php
/**
 * Core model systems, including Model, ModelException, ModelFactory, and ModelValidationException.
 *
 * This file makes up a major component of the underlying datastore system of Core+.  It offers rather high
 * level of accessing and manipulating data in the datastore, essentially providing as the entirity of the
 * "M" part in MVC.
 *
 * In order to build custom models in this system, they must all extend the Model class.  By doing such,
 * they inherit this class's core abilities and functionality essential for operation, not to mention
 * they are picked up as valid Models in the core.
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;

use ArrayAccess;
use Core\Datamodel\Columns\SchemaColumn_enum;
use Core\Forms\FormElement;
use Core\Search\SearchIndex;
use Error;
use Exception;
use ReflectionClass;
use Core\Datamodel\Columns\SchemaColumn;
use Core\Datamodel\Dataset;
use Core\Datamodel\DatasetStream;
use Core\Datamodel\DatasetWhereClause;
use Core\Templates\Template;
use UserModel;

/**
 * Core Model object, main class responsible for the "M" component in MVC.
 *
 * Every Model in the system be it core models or user-created components, MUST extend this
 * class in order for proper functioning.
 */
class Model implements ArrayAccess {

	/**
	 * Short set of characters, usually < 256 in length.
	 */
	const ATT_TYPE_STRING = 'string';
	/**
	 * Any length of string, (up to approximately 64kb or so of text)
	 */
	const ATT_TYPE_TEXT = 'text';
	/**
	 * Any binary-safe data.
	 */
	const ATT_TYPE_DATA = 'data';
	/**
	 * 4-byte Integer, no precision; From −2,147,483,648 to 2,147,483,647
	 */
	const ATT_TYPE_INT = 'int';
	/**
	 * 8-byte Integer, no precision; −9,223,372,036,854,775,808 to 9,223,372,036,854,775,807
	 */
	const ATT_TYPE_LONG = 'long';
	/**
	 * Number with decimal places.
	 */
	const ATT_TYPE_FLOAT = 'float';
	/**
	 * True or false value, approximate values are translated appropriately, (such as "yes" or "0").
	 */
	const ATT_TYPE_BOOL = 'boolean';
	/**
	 * Value from a set of options
	 */
	const ATT_TYPE_ENUM = 'enum';
	/**
	 * This is a unique id for a table, but instead of relying on auto-increment, it's based on Core's GenerateUUID method.
	 */
	const ATT_TYPE_UUID = '__uuid';
	/**
	 * A column that is a "foreign key" to a UUID column.
	 */
	const ATT_TYPE_UUID_FK = '__uuid_fk';
	/**
	 * The auto-incrementing integer of a table
	 */
	const ATT_TYPE_ID = '__id';
	/**
	 * A column that is a "foreign key" to an ID column.
	 */
	const ATT_TYPE_ID_FK = '__id_fk';
	/**
	 * Date this record was last updated, updated automatically
	 */
	const ATT_TYPE_UPDATED = '__updated';
	/**
	 * Date this record was originally created, set automatically
	 */
	const ATT_TYPE_CREATED = '__created';
	/**
	 * Date this record was deleted, (if deleted)
	 * Will be updated automatically on delete() instead of actually deleting the record.
	 *
	 * Calling delete() a second time on a "deleted" record will actually delete it from the system.
	 */
	const ATT_TYPE_DELETED = '__deleted';
	/**
	 * Site ID for the current site.
	 * Has no function outside of enterprise/multisite mode.
	 */
	const ATT_TYPE_SITE = '__site';

	/**
	 * Alias of another column in the record.
	 * Useful for column renames where the data needs to persist to the new column name.
	 */
	const ATT_TYPE_ALIAS = '__alias';

	/**
	 * Meta field, not actually in the database, but present in the logic for some reason or another.
	 * This is useful for defining fields that are calculated for listing tables.
	 */
	const ATT_TYPE_META = '__meta';

	/**
	 * ISO 8601 datetime formatted field.
	 *
	 * This stores the data as `YYYY-mm-dd HH:ii:ss` in the database and
	 * is the default `datetime` format for MySQL.
	 *
	 * This format is discouraged in Core+ as it does not contain necessary timezone information
	 * and is ambiguous when considering full l10n, but is allowed.
	 */
	const ATT_TYPE_ISO_8601_DATETIME = 'ISO_8601_datetime';

	/**
	 * ISO 8601 date formatted field.
	 *
	 * This stores the data as `YYYY-mm-dd` in the database and
	 * is the default `date` format for MySQL.
	 *
	 * This format is discouraged in Core+ as it does not contain necessary timezone information
	 * and is ambiguous when considering full l10n, but is allowed.
	 */
	const ATT_TYPE_ISO_8601_DATE = 'ISO_8601_date';

	const ATT_TYPE_SEARCH_INDEX_STR = '__search_index_str';
	const ATT_TYPE_SEARCH_INDEX_PRI = '__search_index_pri';
	const ATT_TYPE_SEARCH_INDEX_SEC = '__search_index_sec';

	/**
	 * Validation for any non-blank value.
	 */
	const VALIDATION_NOTBLANK = "/^.+$/";

	/**
	 * Validation for a valid email address.
	 *
	 * Optionally can use DNS lookups to perform a more indepth validation of the domain.
	 */
	const VALIDATION_EMAIL = 'Core\\ModelValidator::Email';

	/**
	 * Validation for a valid protocol-based URL.
	 *
	 * This includes http, https, ftp, ssh, and any other protocol.
	 */
	const VALIDATION_URL = '#^[a-zA-Z]+://.+$#';

	/**
	 * Validation for a valid web URL.
	 *
	 * This will only match http:// or https://.
	 */
	const VALIDATION_URL_WEB = '#^[hH][tT][tT][pP][sS]{0,1}://.+$#';

	/**
	 * Validation for GT 0 ints.
	 */
	const VALIDATION_INT_GT0 = 'Core\\ModelValidator::IntGT0';

	/**
	 * Simple validation for any whole numeric digits.
	 *
	 * 0, 1, 1337, 100000 are all valid whole numbers
	 */
	const VALIDATION_NUMBER_WHOLE = "/^[0-9]*$/";

	/**
	 * Simple validation for USD currencies
	 */
	const VALIDATION_CURRENCY_USD = '#^(\$)?[,0-9]*(?:\.[0-9]{2})?$#';

	/**
	 * Definition for a model that has exactly one child table as a dependency.
	 *
	 * !WARNING! This gets deleted automatically if the parent is deleted!
	 */

	const LINK_HASONE  = 'one';
	/**
	 * Definition for a model that has several records from a child table for dependencies.
	 *
	 *  !WARNING! These get deleted automatically if the parent is deleted!
	 */
	const LINK_HASMANY = 'many';

	/**
	 * Definition for a model that *BELONGS* to one table.
	 *
	 * This is important because changes do not propagate up to the parent, as deleting the child
	 * should have no effect on the parent!
	 */
	const LINK_BELONGSTOONE = 'belongs_one';

	/**
	 * Definition for a model that *BELONGS* to many tables.
	 *
	 * A good example of this would be a linking table, that contains a Many-to-Many relationship
	 * between two models.
	 */
	const LINK_BELONGSTOMANY = 'belongs_many';

	/** Data encoding for base64 data, only available on ATT_TYPE_DATA columns! */
	const ATT_ENCODING_BASE64 = 'base64';
	/** Data encoding for JSON data, only available on ATT_TYPE_DATA columns! */
	const ATT_ENCODING_JSON = 'json';
	/** Data encoding for serialized data, only available on ATT_TYPE_DATA columns! */
	const ATT_ENCODING_SERIALIZE = 'serialize';
	/** Data encoding for compressed data, only available on ATT_TYPE_DATA columns! */
	const ATT_ENCODING_GZIP = 'gzip';
	/** Default encoding for strings in the database, available for any STRING-based columns. */
	const ATT_ENCODING_UTF8 = 'utf8';
	/** @var string NEW default encoding for strings, supports extended unicode (aka emojis) */
	const ATT_ENCODING_UTF8MB4 = 'utf8mb4';

	/** Pending creation */
	const STATUS_PENDING_CREATE = 'pending_creation';
	/** Pending delete */
	const STATUS_PENDING_DELETE = 'pending_deletion';
	/** Pending update */
	const STATUS_PENDING_UPDATE = 'pending_update';
	/** Exists and no changes pending */
	const STATUS_PUBLISHED      = '';


	/**
	 * Which DataModelInterface should this model execute its operations with.
	 * 99.9% of the time, it's fine to leave this as null, which will use the
	 * system DMI.  If however you want to utilize a Model with Memcache,
	 * (say for session information), it can be useful.
	 *
	 * @var \Core\Datamodel\BackendInterface
	 */
	protected $_interface = null;

	/**
	 * Allow data to get overloaded onto models.
	 * This is common with Controllers tacking on extra data for templates to better handle the model.
	 * This data is not saved and does not effect the dirty flags.
	 *
	 * @var array
	 */
	protected $_dataother = [];

	/**
	 * @var SchemaColumn[] Set of columns along with the data represented therein.
	 */
	protected $_columns;

	/**
	 * @var array Array of aliases defined on this model, used because they can work from getters and setters.
	 */
	protected $_aliases;

	/**
	 * @var SchemaColumn[] Any columns that are set as *meta* types
	 */
	protected $_metas;

	/**
	 * @var boolean Flag to signify if this record exists in the database.
	 */
	protected $_exists = false;

	/** @var ModelLink Set of linked children models on this model, populated by the Constructor and *link methods. */
	protected $_linked = [];

	/** @var array Cache of link names to their index, used to speed up repeated queries. */
	protected $_linkIndexCache = [];

	protected $_cacheable = true;

	/**
	 * @var array The schema as per defined in the extending model.
	 */
	public static $Schema = [];

	/**
	 * @var array Any indexes that are required on this data.  Must be defined in the extending model if used.
	 */
	public static $Indexes = [];

	/**
	 * @var bool Set to true if this model is searchable, (and auto-create the necessary search index fields).
	 */
	public static $HasSearch = false;

	/**
	 * @var bool Set to true if this model has a created timestamp, (and auto-create the necessary search index fields).
	 */
	public static $HasCreated = false;

	/**
	 * @var bool Set to true if this model has an updated timestamp, (and auto-create the necessary search index fields).
	 */
	public static $HasUpdated = false;

	/**
	 * @var bool Set to true if this model has a deleted timestamp, (and auto-create the necessary search index fields).
	 */
	public static $HasDeleted = false;

	/**
	 * @var null|string Set to a string name to group this Model in the export/import system
	 */
	public static $ExportGroup = null;

	public static $_ModelCache = [];

	/**
	 * @var array Cache for blank Find statements.
	 *
	 * ONLY used with a completely blank Find() query!!!
	 */
	public static $_ModelFindCache = [];

	/**
	 * @var array Array of rendered Schemas for each named model.
	 *
	 * Used to accelerate GetSchema on multiple calls for the same model type.
	 */
	public static $_ModelSchemaCache = [];

	/**
	 * Used with the defer save option to bulk-insert commands when possible.
	 * This is used to speed up bulk INSERT statements.
	 *
	 * @var array
	 */
	protected static $_DeferInserts = [];

	/**
	 * List of models that provide supplemental functionality on the base model.
	 *
	 * Used for GetSchema, GetIndexes, and the various Model-based hooks.
	 *
	 * @var array
	 */
	public static $ModelSupplementals = [];


	// <editor-fold defaultstate="collapsed" desc="Constructor, load, and save PUBLIC methods">

	/**
	 * Create a new instance of the requested model.
	 *
	 * @param null $key
	 *
	 * @throws Exception
	 */
	public function __construct(...$keys) {
		// If an extending model set links in that constructor, reformat them to ensure they're set correctly.
		if(sizeof($this->_linked)){
			$clone = $this->_linked;
			$this->_linked = [];
			foreach($clone as $model => $dat){
				$dat['model'] = $model;
				$this->_linked[] = new ModelLink(null, $dat);
			}
		}

		// Update the _data array based on the schema.
		$schema = self::GetSchema(true);
		$this->_columns = [];
		$this->_aliases = [];

		foreach($schema->definitions as $k => $column) {
			$column->parent = $this;
			$this->_columns[ $k ] = $column;

			if($column->link){
				if(!isset($column->link->model)){
					throw new Exception('Required attribute [model] not provided on link [' . $k . '] of model [' . get_class($this) . ']');
				}
				if(!isset($column->link->type)){
					throw new Exception('Required attribute [type] not provided on link [' . $k . '] of model [' . get_class($this) . ']');
				}

				// And populate the linked array with this link data.
				$this->_linked[] = $column->link;
			}
		}


		// Aliases reference other columns.  The other column must exist.
		foreach($schema->aliases as $a => $k) {
			if(!isset($this->_columns[ $k ])){
				throw new Exception('Aliased column ' . $k . ' does not exist on model ' . get_class($this) . '!  (Referenced from ' . $a . ')');
			}

			$this->_aliases[ $a ] = $k;
		}


		// Metas!
		foreach($schema->metas as $k => $column) {
			$column->parent = $this;

			$this->_metas[ $k ] = $column;
		}

		$pks       = isset($schema->indexes['primary']) ? $schema->indexes['primary'] : [];
		$countPK   = count($pks);
		$countKeys = count($keys);

		if( $countPK && $countKeys ){
			// There is at least one primary key and at least one parameter requested.

			if(
				in_array('site', $pks) &&
				$countPK === $countKeys + 1
			) {
				// This model has a "site" field and it was omitted from the request.
				// This is a magic field that gets populated in when not set so to allow
				// non-site aware applications to utilize data that's constrained to multisites.
				$this->_columns['site']->setValueFromApp(
					Core::IsComponentAvailable('multisite') && \MultiSiteHelper::IsEnabled() ?
					\MultiSiteHelper::GetCurrentSiteID() :
					0
				);

				// Remove and reindex the PKSs now that site is removed from the scenario.
				unset($pks[array_search('site', $pks)]);
				$pks = array_values($pks);
				$countPK--;
			}

			/*var_dump($keys, $pks);
			echo '<pre>';
			debug_print_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS);
			echo '</pre>';*/

			if($countPK === $countKeys) {
				// Keys match up 1-to-1, load them!
				foreach($pks as $idx => $key) {
					$this->_columns[ $key ]->setValueFromApp( $keys[ $idx ]);
				}
			}

			$this->load();
		}
	}

	/**
	 * Delete this record from the datastore.
	 *
	 * Will IMMEDIATELY remove the record!
	 *
	 * If this model has a "deleted" column that is set as a zero value, that record is set to the current timestamp instead.
	 * This functionality is meant for advanced record tracking such as those in use in sync systems.
	 *
	 * @return bool
	 * @throws Exception
	 */
	public function delete() {

		if(!$this->checkExists()) {
			// Models that do not exist cannot be deleted.
			// This should safely be able to abort without throwing anything,
			// as a request to delete something that does not exist in the first place
			// will still yield the same results.
			return false;
		}

		$classname = get_called_class();
		// Allow all supplemental models to tap into this too!
		if(isset(self::$ModelSupplementals[$classname])){
			foreach(self::$ModelSupplementals[$classname] as $supplemental){
				if(class_exists($supplemental)){
					$ref = new ReflectionClass($supplemental);
					if($ref->hasMethod('PreDeleteHook')){
						$ref->getMethod('PreDeleteHook')->invoke(null, $this);
					}
				}
			}
		}

		foreach ($this->_columns as $c) {
			/** @var \Core\Datamodel\Columns\SchemaColumn $c */

			// Certain key types have certain functions.
			if($c->type === Model::ATT_TYPE_DELETED) {
				// Is this record already set as deleted?
				// If it is, then proceed with the delete as usual.
				// Otherwise, update the record instead of deleting it.
				if(!$c->value){
					$nv = Time::GetCurrentGMT();
					$this->set($c->field, $nv);
					return $this->save();
				}
				else{
					// I can safely break out of the foreach statement at this stage.
					break;
				}
			}
		}

		// Blank out any dependent records based on links.
		// Since relationships may have various levels, I need to execute delete on each of them instead of just
		// issuing a broad delete request.
		foreach ($this->_linked as $k => $l) {

			switch($l->type) {
				case Model::LINK_HASONE:
					// Delete this child, (and any of its subordinates).
					$child = $this->getLink($k);
					$child->delete();
					break;
				case Model::LINK_HASMANY:
					// Request all the children and issue a delete on them.
					$children = $this->getLink($k);
					foreach($children as $child){
						/** @var Model $child */
						$child->delete();
					}
					break;
				// There is no default behaviour... other than to ignore it.
			}

			$this->_linked[$k]->records = null;
		}


		if ($this->exists()) {

			$schemas = $this->GetSchema();
			$pri = $schemas->indexes['primary'] ?: [];

			// Check and see if this model extends another model.
			// If it does, then create/update that parent object to keep it in sync!
			if(($class = get_parent_class($this)) != 'Model'){
				if(
					count($pri) === 1 &&
					isset($schemas->definitions[ $pri[0] ]) &&
					$schemas->definitions[ $pri[0] ]->type === Model::ATT_TYPE_UUID
				) {
					$refp = new ReflectionClass($class);
					$refm = $refp->getMethod('Construct');
					/** @var Model $parent */
					$parent = $refm->invoke(null, $this->get( $pri[0] ));
					$parent->delete();
				}
			}

			$n = $this->_getTableName();
			$i = self::GetIndexes();
			// Delete the data from the database.
			$dat = new Dataset();

			$dat->table($n);

			if (count($pri) === 0) {
				throw new Exception('Unable to delete model [ ' . get_class($this) . ' ] without any primary keys.');
			}

			foreach ($pri as $k) {
				$dat->where([$k => $this->get($k)]);
			}

			$dat->limit(1)->delete();

			if ($dat->execute($this->_interface)) {
				$this->_exists = false;
			}
		}

		return true;
	}

	/**
	 * Load this record from the datastore.
	 *
	 * Generally not needed to be called directly, but can be if required.
	 */
	public function load() {

		// If there is no associated table, do not load anything.
		if (!self::GetTableName()) {
			return;
		}

		$schema = self::GetSchema();
		$keys = [];


		// I need to check the pks first.
		// If they're not set I can't load anything from the database.
		if(isset($schema->indexes['primary'])){
			foreach($schema->indexes['primary'] as $k){
				$v = $this->get($k);
				if($v !== null){
					$keys[ $k ] = $v;
				}
			}
		}

		if(!count($keys)) {
			// Skip loading of data if there are no keys requested!
			return;
		}

		// If the enterprise/multimode is set and enabled and there's a site column here,
		// that should be enforced at a low level.
		// This has an additional check where the site could be global as well.
		// This preserves support for global admin pages, for example.
		/** @noinspection PhpUndefinedClassInspection */
		if(
			isset($schema->definitions['site']) &&
			$schema->definitions['site']->type === Model::ATT_TYPE_SITE &&
			array_key_exists('site', $keys)
		){
			$site = $keys['site'];

			if($site === null) {
				// Drop in for the current site.
				$site = Core::IsComponentAvailable('multisite') && MultiSiteHelper::IsEnabled() ? MultiSiteHelper::GetCurrentSiteID() : 0;
			}

			unset($keys['site']);
		}
		else{
			$site = null;
		}

		$ds = Dataset::Init()
			->select('*')
			->table(self::GetTableName())
			->where($keys);

		if($site !== null) {
			// This allows for a multi-match for the site based on global OR requested.
			if($site == -1) {
				$ds->where('site = -1');
			}
			else{
				$ds->getWhereClause()->addWhereSub(
					'OR',
					[
						'site = -1',
						'site = ' . $site
					]
				);
			}
		}


		$data = $ds->execute($this->_interface);

		if ($data->num_rows) {
			$this->_loadFromRecord($data->current());
			$this->_exists = true;
		}
		else {
			$this->_exists = false;
		}

		return;
	}

	/**
	 * Load this model from an associative array, or record.
	 * This is meant to be called from the Factory system, and the data passed in
	 * MUST be sanitized and valid!
	 *
	 * @param array $record
	 */
	public function _loadFromRecord(array $record) {

		foreach($record as $k => $v){
			if($this->checkHasColumn($k)) {
				// Only load data that's present on this Model.
				// THis is a fix for legacy DB data that contains columns that are no longer used.
				try{
					$this->getColumn($k)->setValueFromDB($v);
				}
				catch (Exception $ex) {
					// Ignore exceptions here
				}
			}
			else {
				// This table does not have a column for this type, so attach the value to an 'other' field.
				$this->_dataother[$k] = $v;
			}
		}
		$this->_exists = true;
	}

	/**
	 * Save this Model into the datastore.
	 *
	 * Return true if saved successfully, false if no change required,
	 * and will throw a DMI_Exception if there was an error.
	 *
	 * As of 5.0.0, bulk inserts can be performed by passing TRUE as the one argument.
	 * If this is done, you MUST call CommitSaves() after all data has been stored!
	 *
	 * @param bool $defer Set to true to batch-save this data as a BULK INSERT.
	 *
	 * @return boolean
	 * @throws DMIException
	 */
	public function save($defer = false) {

		$classname = get_called_class();


		// Dispatch the pre-save hook for models.
		// This allows utilities to hook in and modify the model or perform some other action.
		// Allow all supplemental models to tap into this too!
		if(isset(self::$ModelSupplementals[$classname])){
			foreach(self::$ModelSupplementals[$classname] as $supplemental){
				if(class_exists($supplemental)){
					$ref = new ReflectionClass($supplemental);
					if($ref->hasMethod('PreSaveHook')){
						$ref->getMethod('PreSaveHook')->invoke(null, $this);
					}
				}
			}
		}
		HookHandler::DispatchHook('/core/model/presave', $this);


		// Only do the same operation if it's been changed.
		// This is actually a little more in depth than simply seeing if it's been modified, as I also
		// have to look into the linked classes and see if it exists
		$save = false;
		// new models always get saved.
		if(!$this->_exists){
			$save = true;
		}
		// Models that have changed content always get saved.
		elseif($this->checkIsChanged()){
			$save = true;
		}
		else{
			foreach($this->_linked as $l){
				if($l->records !== null || $l->purged !== null) {
					// If there are any linked models in the records array, trigger the save.
					$save = true;
					break;
				}
			}
		}

		if(!$save){
			return false;
		}


		// NEW in 2.8, I need to run through the linked models and see if any local key is a foreign key of a linked model.
		// If it is, then I need to save that foreign model so I can get its updated primary key, (if it changed).
		foreach($this->_linked as $l){
			// If this link has a 1-sized relationship and the local node is set as a FK, then read it as such.

			if($l->type === Model::LINK_HASONE && count($l->on) === 1){
				reset($l->on);
				$remotek = key($l->on);
				$localk  = $l->on[$remotek];

				try{
					$local   = $this->getColumn($localk);
				}
				catch (Exception $ex) {
					// No schema?  Ok, nothing to save to!
					continue;
				}

				if($local->type !== Model::ATT_TYPE_UUID_FK){
					// If this is not a FK, then I don't care!
					continue;
				}

				// OTHERWISE..... ;)
				if($l->records instanceof Model){
					/** @var Model $model */
					$model = $l->records;
				}
				elseif( $local->value instanceof Model ) {
					// The alternative location for this linked model to be.
					// This can happen when the link is established in the Schema data and the parent record doesn't exist yet.
					/** @var Model $model */
					$model = $local->value;
				}
				else{
					// No valid model found... :/
					continue;
				}

				$model->save();
				$this->set($localk, $model->get($remotek));
			}
		}


		if ($this->_exists){
			$changed = $this->_saveExisting();
		}
		else{
			// Inserts can be deferred!
			$this->_saveNew($defer);
			$changed = true;
		}


		// Go through any linked tables and ensure that they're saved as well.
		foreach($this->_linked as $k => $l){
			switch($l->type){
				case Model::LINK_HASONE:
					$models = $l->records !== null ? [ $l->records ] : null;
					$deletes = $l->purged;
					break;
				case Model::LINK_HASMANY:
					$models = $l->records;
					$deletes = $l->purged;
					break;
				default:
					// There is no default behaviour... other than to ignore it.
					$models = null;
					$deletes = null;
					break;
			}

			// Are there deletes requested?
			// Deletes MUST happen before creates because if there is a record that overrides an existing record,
			// but that existing record is set to be deleted, the delete operation must be before the create operation.
			//
			// This happens on the page updates with meta fields.
			// The meta fields are completely re-created, and saved at one time only.
			// Since the incoming data has the same PK as the existing data, (that's already marked as deteled),
			// the operation fails if these orders are reversed.
			if($deletes){
				foreach($deletes as $model){
					$model->delete();
					$changed = true;
				}

				$l->purged = null;
			}

			// Are there saves requested?
			if($models){
				foreach($models as $model){
					/** @var $model Model */
					// Ensure all linked fields still match up.  Something may have been changed in the parent.
					$model->setFromArray($this->_getLinkWhereArray($k));
					if($model->save()){
						$changed = true;
					}
				}
			}
		}

		// Commit all the columns
		$this->_exists   = true;
		foreach($this->_columns as $c){
			/** @var \Core\Datamodel\Columns\SchemaColumn $c */
			$c->commit();
		}

		// Check and see if this model extends another model.
		// If it does, then create/update that parent object to keep it in sync!
		if(($class = get_parent_class($this)) !== 'Core\\Model'){
			$idx = self::GetIndexes();
			if(isset($idx['primary']) && sizeof($idx['primary']) == 1){
				$schema = $this->getKeySchema($idx['primary'][0]);
				if($schema['type'] == Model::ATT_TYPE_UUID){
					$refp = new ReflectionClass($class);
					$refm = $refp->getMethod('Construct');
					/** @var Model $parent */
					$parent = $refm->invoke(null, $this->get($idx['primary'][0]));

					// Populate the parent with this child's data.
					// Any non-existent field will simply be ignored.
					$parent->setFromArray($this->getAsArray());
					if($parent->save()){
						$changed = true;
					}
				}
			}
		}

		if($changed){
			// Allow all supplemental models to tap into this too!
			if(isset(self::$ModelSupplementals[$classname])){
				foreach(self::$ModelSupplementals[$classname] as $supplemental){
					if(class_exists($supplemental)){
						$ref = new ReflectionClass($supplemental);
						if($ref->hasMethod('PostSaveHook')){
							$ref->getMethod('PostSaveHook')->invoke(null, $this);
						}
					}
				}
			}
			HookHandler::DispatchHook('/core/model/postsave', $this);
			// Indicate that something happened.
			return true;
		}
		else{
			// Nothing happened!
			return false;
		}
	}

	/**
	 * Handle data validation for keys.
	 *
	 * This will lookup if any "validation" is set on the schema, and check it if it exists.
	 * This will not actually do any setting, simply return true or throw an exception if there was a problem.
	 *
	 * @param string $k The key to validate
	 * @param mixed  $v The value to validate with
	 *
	 * @return bool
	 *
	 * @throws ModelValidationException
	 */
	public function validate($k, $v) : bool {
		// Is there validation available for this key?

		try {
			$def = $this->getColumn($k);
		}
		catch(Exception $ex) {
			// This happens when no column or column alias exists for the given key.
			// in this case, we have nothing to perform an automatic validation against, so simply allow it to continue.
			return true;
		}

		// If this key is not required and not filled in, skip validation.... it's valid.
		if(
			($v == '' || $v === null) &&
			!$def->required
		){
			return true;
		}

		if($def instanceof SchemaColumn_enum) {
			// ENUMs have options that dictate what can be saved in the database.
			if(is_array($def->options) && !in_array($v, $def->options)) {
				throw new ModelValidationException('Unsupported value for ' . $k . ', "' . $v . '" not an allowed value');
			}
		}


		if($def->validation) {
			// Validation exists... check it.
			$check = $def->validation;

			// Special "this" method.
			if (is_array($check) && sizeof($check) == 2 && $check[0] == 'this') {
				call_user_func([$this, $check[1]], $v);
			}
			// Method-based validation.
			elseif (strpos($check, '::') !== false) {
				// the method can either be true, false or a string.
				// Only if true is returned will that be triggered as success.
				call_user_func($check, $v);
			}
			// regex-based validation.  These don't have any return strings so they're easier.
			elseif (
				($check[0] == '/' && !preg_match($check, $v)) ||
				($check[0] == '#' && !preg_match($check, $v))
			) {
				throw new ModelValidationException($k . ' fails validation due to regex mismatch');
			}
		}
		elseif($def->typecast === 'int' && !$def->autoinc) {
			// Default validation for INTs.
			if(!(
				is_int($v) ||
				//ctype_digit($v) ||
				is_numeric($v) ||
				(is_float($v) && strpos($v, '.') === false)
			)){
				throw new ModelValidationException($k . ' does not appear to be an integer');
			}
		}

		return true;
	}

	// </editor-fold>


	// <editor-fold defaultstate="collapsed" desc="GETTER PUBLIC methods">

	/**
	 * Get the requested column value for this object/record.
	 *
	 * This automatically looks up columns, "other" data fields, aliases, and links.
	 *
	 * ### Content Format Requested
	 *
	 * If the optional second parameter `$format` is passed in, the CTYPE can be defined and
	 * if supported by the underlying dataset, can be rendered into that format natively.
	 *
	 * Please note, not all keys and backends support alternate rendering, and may ignore this parameter.
	 *
	 * @param string      $key
	 * @param string|null $format
	 *
	 * @return mixed
	 */
	public function get($key, $format = null) {
		if($key === '__CLASS__') {
			// Magic key
			return get_called_class();
		}
		elseif($key === '__PRIMARYKEY__') {
			// Magic key
			return $this->getPrimaryKeyString();
		}
		elseif(isset($this->_columns[$key])) {
			// It's a standard column object!
			// Try to utilize the underlying Column object for format handling.
			$col = $this->_columns[$key];

			if($col->formatter){
				return $col->callFormatter($format);
			}
			elseif(
				$format === 'text/html' &&
				$col->type === Model::ATT_TYPE_ENUM &&
				$col->valueTranslated
			) {
				// No formatter was provided, but it's an ENUM, so the options probably have I18N strings associated!
				return t(
					'VALUE_' . $col->getI18NKey() . '_' . \Core\I18N\I18NLoader::KeyifyString($col->valueTranslated)
				);
			}
			else{
				// No formatter was registered, passthru.
				return $col->valueTranslated;
			}
		}
		elseif(isset($this->_aliases[$key])) {
			// It's an alias of another key, re-call this method on that key instead.
			return $this->get($this->_aliases[$key]);
		}
		elseif(isset($this->_metas[$key])) {
			// Metas are also standard column objects, mostly.
			// Try to utilize the underlying Column object for format handling.
			$col = $this->_metas[$key];

			if($col->formatter){
				return $col->callFormatter($format);
			}
			else {
				// Meta fields only support formatters, otherwise they don't have any values.
				return null;
			}
		}
		elseif (array_key_exists($key, $this->_dataother)) {
			// Check if this data was set from the "set" command on a non-tracked column.
			return $this->_dataother[$key];
		}
		elseif($this->getLink($key)){
			// Check if this data is actually a linked model
			return $this->getLink($key);
		}
		else {
			return null;
		}
	}

	/**
	 * Get an array of all changed columns on this Model
	 *
	 * @return SchemaColumn[]
	 */
	public function getChangedColumns() : array {
		$ret = [];
		foreach($this->_columns as $col) {
			if($col->changed()) {
				$ret[] = $col;
			}
		}

		return $ret;
	}

	/**
	 * Get all columns as the raw SchemaColumn for this object
	 *
	 * @return SchemaColumn[]
	 */
	public function getColumns() : array {
		return $this->_columns;
	}

	/**
	 * Get the list of CSS classnames for this record, by default just the model's name.
	 *
	 * @return string
	 */
	public function getClass() : string {
		return strtolower(get_class($this));
	}

	/**
	 * Get the "keyname" for this Model.
	 *
	 * This will modify "SomethingFooModel" to "something_foo"
	 *
	 * @return string
	 */
	public function getClassKeyName() : string {
		$class = get_class($this);

		if(strrpos($class, 'Model') === (strlen($class) - 5) ) {
			// Trim off the suffix 'Model' that should be present.
			$class = substr($class, 0, -5);
		}

		$class = strtolower(preg_replace('/([A-Z])/', '_$1', $class));

		if($class[0] === '_') {
			// This probably (should), start with an underscore, as the class should be capitalized.
			$class = substr($class, 1);
		}

		return $class;
	}

	/**
	 * Get an array of control links for this model.
	 *
	 * Please call array_merge($results, parent::getControlLinks())
	 * in any extending method to retain the supplemental model functionality.
	 *
	 * The returned data MUST be either an empty array or an index array of arrays.
	 * Each internal array should have link, title, icon, and any other parameter supported by the ViewControl
	 *
	 * @return array[]
	 */
	public function getControlLinks() : array {
		$ret = [];

		$classname = get_class($this);
		$baseurl = '/' . strtolower($classname);

		// Allow all supplemental models to tap into the schema too!
		if(isset(self::$ModelSupplementals[$classname])){
			foreach(self::$ModelSupplementals[$classname] as $supplemental){
				if(class_exists($supplemental)){
					$ref = new ReflectionClass($supplemental);
					if($ref->hasMethod('GetControlLinks')){
						$supplementalRet = $ref->getMethod('GetControlLinks')->invoke(null, $this);

						// Include some error handling for supplemental models that are not setup fully.
						if(!is_array($supplementalRet)){
							trigger_error($supplemental . '::GetControlLinks must return an array!', E_USER_NOTICE);
						}
						else{
							$ret = array_merge($ret, $supplementalRet);
						}
					}
				}
			}
		}

		// Allow hooks to be used to register control links too
		$additionallinks = HookHandler::DispatchHook('/core/controllinks' . $baseurl, $this);
		if(is_array($additionallinks)) {
			$ret = array_merge($ret, $additionallinks);
		}

		return $ret;
	}

	/**
	 * Get the column schema for a given key or alias.
	 *
	 * @param string $key
	 *
	 * @throws Exception If the column is not located
	 *
	 * @return SchemaColumn
	 */
	public function getColumn(string $key) : SchemaColumn {
		if( isset($this->_columns[ $key ]) ){
			// Return the original column if it's located.
			return $this->_columns[ $key ];
		}
		elseif( isset($this->_aliases[ $key ]) ){
			// Allow aliases to be requested here as well; will return the aliased Column.
			return $this->_columns[ $this->_aliases[ $key] ];
		}
		else{
			throw new Exception($key . ' does not exist within ' . get_class($this));
		}
	}

	/**
	 * Get a hash of data tags for this Model.
	 *
	 * By default, nothing is returned, but extend this method to set your own data-* tags.
	 *
	 * @return array
	 */
	public function getDataTags() : array {
		return [];
	}

	/**
	 * Get all "data-*" tags as a flat string that can be injected into templates
	 *
	 * @return string
	 */
	public function getDataTagsAsString() : string {
		$dataTags = $this->getDataTags();
		$out      = '';

		foreach($dataTags as $tag => $value){
			$out .= ' data-' . $tag . '="' . str_replace('"', '&quot;', $value) . '"';
		}

		return $out;
	}

	/**
	 * Get the draft status of this model.
	 *
	 * @return string
	 */
	public function getDraftStatus() : string {
		if(!$this->exists()){
			return self::STATUS_PENDING_CREATE;
		}
		elseif($this->hasDraft() && $this->get('___auditmodel')->get('data') == '[]'){
			return self::STATUS_PENDING_DELETE;
		}
		elseif($this->hasDraft()){
			return self::STATUS_PENDING_UPDATE;
		}
		else{
			return self::STATUS_PUBLISHED;
		}
	}

	/**
	 * Convenience method to get the original "DB" value of a given column.
	 * Useful for listing what the previous value was before a change.
	 *
	 * If no change occurred, simply the DB value will be returned.
	 *
	 * @param string $key
	 * @return mixed
	 */
	public function getOriginalValue(string $key){
		try{
			return $this->getColumn($key)->valueDB;
		}
		catch (Exception $ex) {
			return null;
		}
	}

	/**
	 * Get the human-readable label for this record.
	 *
	 * By default, it will sift through the schema looking for keys that appear to be human-readable terms,
	 * but for best results, please extend this method and have it return what's necessary for the given Model.
	 *
	 * @return string
	 */
	public function getLabel(){

		if(($name = $this->get('name')) !== null) {
			return $name;
		}
		elseif(($name = $this->get('title')) !== null) {
			return $name;
		}
		elseif(($name = $this->get('key')) !== null) {
			return $name;
		}
		else {
			return get_class($this) . ' ' . $this->getPrimaryKeyString();
		}
	}

	/**
	 * Get the search index representation for this Model
	 *
	 * Provides more advanced information beyond simply text.
	 *
	 * @return SearchIndex
	 */
	public function getSearchIndex() : SearchIndex {
		$index = new SearchIndex();

		// The default behaviour is to sift through the records on this model itself.
		foreach($this->_columns as $column) {
			if(!$column->searchable){
				// Skip any columns that are marked as not searchable
				continue;
			}

			$val = $this->get( $column->field );

			if(preg_match('/^[0-9\- \.\(\)]*$/', $val)) {
				// If this is a numeric-based value, compress all the numbers without formatting.
				// This is to support phone numbers that may have arbitrary formatting applied.
				$val = preg_replace('/[ \-\.\(\)]/', '', $val);
			}

			$index->addText($val, $column->searchWeight);
		}

		return $index;
	}

	/**
	 * Get a textual representation of this Model as a flat string.
	 *
	 * Used by the search systems to index the model, (or multiple models into one).
	 *
	 * @return string
	 */
	public function getSearchIndexString() : string {
		return $this->getSearchIndex()->getString();
	}

	/**
	 * Get the friendly version of the ID of this order,
	 * in this case, it's the last 8 digits of the ID uppercase.
	 *
	 * This could lead to duplicate IDs, but with entropy depth of 16 ^ 8,
	 * but 4.2B unique combinations should be sufficient for most cases.
	 *
	 * If the model does not contain an "id" column, this will return an empty string.
	 *
	 * @return string
	 */
	public function getFriendlyID() : string {
		if($this->id === null) {
			return '';
		}
		elseif(strlen($this->id) < 8) {
			// Just in case the ID is a basic int.
			return wordwrap(str_pad($this->id, 8, '0', STR_PAD_LEFT), 4, '-', true);
		}
		else {
			$id = str_replace('-', '', $this->id);

			// Trim to just the last 8 characters and split the string with a hyphen
			$id = wordwrap(substr($id, -8), 4, '-', true);

			// and all uppercase them
			return strtoupper($id);
		}
	}

	// </editor-fold>


	// <editor-fold defaultstate="collapsed" desc="SETTER PUBLIC Methods">

	/**
	 * Set a value of a specific key.
	 *
	 * The data is validated automatically as per the specific Model specifications.
	 *
	 * This supports data overloading.
	 *
	 * @param string $k The key to set
	 * @param mixed  $v The value to set
	 *
	 * @throws ModelValidationException
	 */
	public function set(string $k, $v) {
		if($v instanceof Model && $this->_getLinkIndex($k) !== null){
			// Allow setting a linked Model via set().
			// This allows a full Model to be passed in from let's say a form system.
			$this->setLink($k, $v);
			return;
		}

		try{
			$column = $this->getColumn($k);
		}
		catch (Exception $ex) {
			// If the key schema isn't available, then simply set the dataother and exit out.
			// This generally means that the key isn't a registered column on this model
			// and that the developer is overloading the model with extra data.
			$this->_dataother[$k] = $v;
			return;
		}

		// Is there validation for this key?
		// That function will handle all of this logic, (including the exception throwing)
		$this->validate($k, $v, true);

		// Set the propagation FIRST, that way I have the old key in memory to lookup.
		$this->_setLinkKeyPropagation($k, $v);

		$column->setValueFromApp($v);
	}

	/**
	 * Set properties on this model from an associative array of key/value pairs.
	 *
	 * @param $array
	 *
	 * @throws ModelValidationException
	 */
	public function setFromArray(array $array) {
		foreach ($array as $k => $v) {
			$this->set($k, $v);
		}
	}

	/**
	 * Similar to set, but this is meant to be used with external data sources such as import operations.
	 *
	 * @param string $key
	 * @param mixed $value
	 *
	 * @throws ModelValidationException
	 */
	public function import(string $key, $value) {
		if(isset($this->_columns[$key])) {
			// It's a standard column object!
			// Try to utilize the underlying Column object for format handling.
			$col = $this->_columns[$key];

			if($col->importer) {
				$col->callImporter($value);
			}
			elseif($col->importer === null) {
				// No importer was registered, passthru.
				$this->set($key, $value);
			}
			else {
				// Importer was disabled
				throw new ModelValidationException('Field key ' . $key . ' does not support importing!');
			}
		}
		elseif(isset($this->_aliases[$key])) {
			// It's an alias of another key, re-call this method on that key instead.
			$this->import($this->_aliases[$key], $value);
		}
		elseif(isset($this->_metas[$key])) {
			// Metas are also standard column objects, mostly.
			// Try to utilize the underlying Column object for format handling.
			$col = $this->_metas[$key];

			if($col->importer){
				$col->callImporter($value);
			}
			elseif($col->importer === null) {
				// Meta fields generally require an importer of some sort, but just call the set
				// on the field keyname so at least something may happen.
				$this->set($key, $value);
			}
			else {
				// Importer was disabled
				throw new ModelValidationException('Field key ' . $key . ' does not support importing!');
			}
		}
	}

	// </editor-fold>


	// <editor-fold defaultstate="collapsed" desc="CHECK PUBLIC Methods">

	/**
	 * Get if this model exists in the datastore already.
	 *
	 * @return bool
	 */
	public function checkExists() : bool {
		return $this->_exists;
	}

	/**
	 * Lookup and see if this model instance has a draft saved for it.
	 *
	 * @return bool
	 */
	public function checkHasDraft() : bool {
		if(Core::IsComponentAvailable('model-audit')){
			/** @noinspection PhpUndefinedNamespaceInspection */
			/** @noinspection PhpUndefinedClassInspection */
			return ModelAudit\Helper::ModelHasDraft($this);
		}
		else{
			// If the underlying component is not available, drafts cannot be enabled!
			return false;
		}
	}

	/**
	 * Check if this Model has a given column
	 *
	 * @param string $column
	 * @return bool
	 */
	public function checkHasColumn(string $column) : bool {
		return isset($this->_columns[$column]) || isset($this->_aliases[$column]);
	}

	/**
	 * Get if this model is marked as deleted and/or deleted already.
	 *
	 * @return bool
	 */
	public function checkIsDeleted() : bool {
		foreach ($this->_columns as $c) {
			// Certain key types have certain functions.
			if($c->type === Model::ATT_TYPE_DELETED && $c->value) {
				// Is this record already set as deleted?
				// If value is > 0|NULL, it's marked as deleted!
				return true;
			}
		}

		// Otherwise, it's safe to check the value of exists.
		return !$this->_exists;
	}

	/**
	 * Get if this model is a new entity that doesn't exist in the datastore.
	 *
	 * @return bool
	 */
	public function checkIsNew() : bool {
		return !$this->_exists;
	}

	/**
	 * Get if this model has changes that are pending to be applied back to the datastore.
	 *
	 * @param string|null $key Optionally set a key name here to check only that one key.
	 *
	 * @return bool
	 */
	public function checkIsChanged(string $key = null) : bool {
		if($key === null){
			// Check all columns!
			foreach($this->_columns as $c){
				if($c->changed()){
					return true;
				}
			}
			return false;
		}

		try{
			// Try to use the Column's built-in changed() method for this check.
			// If doesn't exist, the catch will check the dataother array.
			return $this->getColumn($key)->changed();
		}
		catch (Exception $ex) {
			// Column doesn't exist or something else happened, check if there is content in the "other" array.
			return isset($this->_dataother[ $key ]);
		}
	}

	/**
	 * Check if the given User can access this particular object based on their access permissions.
	 *
	 * By default this is a simple 'return true' check, but can be extended for custom functionality.
	 *
	 * Please note, THIS DOES NOT PREVENT ANY ACTIONS, simply is a convenience function for controllers
	 * to utilize to check the object ACL.
	 *
	 * @param UserModel|null $user The user to check access.
	 *
	 * @return bool
	 */
	public function checkCanAccess(UserModel $user = null) : bool {
		return true;
	}

	// </editor-fold>

	// <editor-fold defaultstate="collapsed" desc="AS Transposing PUBLIC Methods">

	/**
	 * Get the properties of this record as a flat array
	 *
	 * @return array
	 */
	public function asArray() : array {
		$ret = [];

		foreach($this->_columns as $c){
			/** @var SchemaColumn $c */
			$ret[ $c->field ] = $c->valueTranslated;
		}

		return $ret;
	}

	/**
	 * Get this Model and all associated Columns as an array of FormElements
	 * that can be modified by an outside source if necessary.
	 *
	 * @return FormElement[]
	 */
	public function asFormArray() : array  {
		$ret = [];

		foreach($this->_columns as $c){
			/** @var SchemaColumn $c */
			$form = $c->getAsFormElement();
			if($form !== null){
				// Columns that are set as "disabled" do not register as form elements.
				$ret[ $c->field ] = $form;
			}
		}

		return $ret;
	}

	/**
	 * Get the properties of this record as a flat JSON-encoded string
	 *
	 * @return string
	 */
	public function asJSON() : string {
		return json_encode($this->asArray());
	}

	// </editor-fold>


	// <editor-fold defaultstate="collapsed" desc="Deprecated Methods (Generally renamed functionality)">

	/**
	 * Alias of asArray()
	 *
	 * @deprecated 201712
	 *
	 * @return array
	 */
	public function getAsArray() {
		return $this->asArray();
	}

	/**
	 * Alias of asJSON()
	 *
	 * @deprecated 201712
	 *
	 * @return string
	 */
	public function getAsJSON(){
		return $this->asJSON();
	}

	/**
	 * Alias of asFormArray()
	 *
	 * @deprecated 201712
	 *
	 * @return array
	 */
	public function getAsFormArray(){
		return $this->asFormArray();
	}

	/**
	 * Alias of checkHasDraft()
	 *
	 * @deprecated 201712
	 *
	 * @return bool
	 */
	public function hasDraft(){
		return $this->checkHasDraft();
	}

	/**
	 * Alias of checkExists()
	 *
	 * @deprecated 201712
	 *
	 * @return bool
	 */
	public function exists() {
		return $this->checkExists();
	}

	/**
	 * Alias of checkIsDeleted()
	 *
	 * @deprecated 201712
	 *
	 * @return bool
	 */
	public function isdeleted(){
		return $this->checkIsDeleted();
	}

	/**
	 * Alias of checkIsNew()
	 *
	 * @deprecated 201712
	 *
	 * @return bool
	 */
	public function isnew() {
		return $this->checkIsNew();
	}

	/**
	 * Alias of checkIsChanged()
	 *
	 * @deprecated 201712
	 *
	 * @return bool
	 */
	public function changed($key = null){
		return $this->checkIsChanged($key);
	}

	// </editor-fold>


	///////////////////////////////////////////////////////////////////////////
	////                      LINK SPECIFIC METHODS                        ////

	/**
	 * Get the model factory for a given link.
	 *
	 * Useful for manipulating the factory of the data.
	 *
	 * @param string $linkname
	 *
	 * @return ModelFactory
	 */
	public function getLinkFactory($linkname){
		$idx = $this->_getLinkIndex($linkname);
		if($idx === null){
			return null; // @todo Error Handling
		}

		$c = $this->_getLinkClassName($linkname);

		$f = new ModelFactory($c);
		switch($this->_linked[$idx]->type){
			case Model::LINK_HASONE:
			case Model::LINK_BELONGSTOONE:
				$f->limit(1);
				break;
		}

		$wheres = $this->_getLinkWhereArray($linkname);
		$f->where($wheres);

		// Yup, no get or anything, just build the factory.
		return $f;
	}

	/**
	 * Get linked models to this model based on a link name
	 *
	 * If the link type is a one-to-one or many-to-one, (HASONE), a single Model is returned.
	 * else this behaves as the Find function, where an array of models is returned.
	 *
	 * @param string      $linkname The linked model name (minus the Model part)
	 * @param null|string $order    Specify the order clause
	 *
	 * @return Model|array
	 */
	public function getLink($linkname, $order = null) {
		$idx = $this->_getLinkIndex($linkname);
		if($idx === null){
			return null; // @todo Error Handling
		}

		// Allow order to be set from the model itself.
		if($order === null){
			$order = $this->_linked[$idx]->order;
		}

		// Try to keep these in cache, so when they change I'll be able to save them on the parent's save function.
		if($this->_linked[$idx]->records === null) {
			$f       = $this->getLinkFactory($linkname);
			$c       = $this->_getLinkClassName($linkname);
			$wheres  = $this->_getLinkWhereArray($linkname);
			$isBlank = true;


			// Check and see if there is even a local key to attach anything onto!
			foreach($wheres as $val){
				if(trim($val) != ''){
					$isBlank = false;
					break;
				}
			}

			if($isBlank){
				// Skip the lookup, it'll be blank anyway!
				// This saves a few ms on bulk pages that check a lot of potentially empty records.
				$this->_linked[$idx]->records = ($f->getDataset()->_limit == 1) ? null : [];
			}
			else{
				if ($order){
					$f->order($order);
				}
				$this->_linked[$idx]->records = $f->get();
			}

			// Ensure that it's a valid record and not null.  If it's a LINK_ONE, the factory will return null if it doesn't exist.
			if ($this->_linked[$idx]->records === null) {
				$this->_linked[$idx]->records = new $c();
				foreach ($wheres as $k => $v) {
					$this->_linked[$idx]->records->set($k, $v);
				}
			}
		}

		return $this->_linked[$idx]->records;
	}

	/**
	 * In 1-to-1 mode, this returns either the single record matched or nothing at all.
	 * In 1-to-M mode, this returns an attached object with the requested search keys, either bound or new.
	 *
	 * @param string $linkname
	 * @param array  $searchkeys
	 * @return bool|\Model|null
	 */
	public function findLink($linkname, $searchkeys = []) {
		$l = $this->getLink($linkname);
		if ($l === null) return null;

		// aka 1-to-1 mode.
		if (!is_array($l)) {
			$f = true;
			foreach ($searchkeys as $k => $v) {
				if ($l->get($k) != $v) {
					$f = false;
					break;
				}
			}
			return ($f) ? $l : false;
		}
		else {
			foreach ($l as $model) {
				$f = true;
				foreach ($searchkeys as $k => $v) {
					if ($model->get($k) != $v) {
						$f = false;
						break;
					}
				}
				if ($f) return $model;
			}

			// Still here?  Guess it didn't find it.
			// Create the element, attach it and return that!
			$c = $this->_getLinkClassName($linkname);
			/** @var $model Model */
			$model = new $c();
			$model->setFromArray($this->_getLinkWhereArray($linkname));
			$model->setFromArray($searchkeys);
			//var_dump($model); die();
			$model->load();
			$this->_linked[$linkname]->records[] = $model;
			return $model;
		}
		// Search through each of these and find a matching
	}

	/**
	 * Add a model to the set of linked records, (or replace it in the case or HASONE).
	 *
	 * Administrative method used internally by some systems.  This allows a link to be overwritten externally.
	 *
	 * Particularly useful for BELONGSTOONE models being updated by their parent.
	 *
	 * @param       $linkname
	 * @param Model $model
	 *
	 * @return void
	 */
	public function setLink($linkname, Model $model) {
		$idx = $this->_getLinkIndex($linkname);
		if($idx === null){
			return null; // @todo Error Handling
		}

		// Update the cached model.
		switch($this->_linked[$idx]->type){
			case Model::LINK_HASONE:
			case Model::LINK_BELONGSTOONE:
				$this->_linked[$idx]->records = $model;
				break;
			case Model::LINK_HASMANY:
			case Model::LINK_BELONGSTOMANY:
				if($this->_linked[ $idx ]->records === null){
					$this->_linked[ $idx ]->records = [];
				}
				$this->_linked[ $idx ]->records[] = $model;
				break;
		}
	}

	/**
	 * Reset the linked models in this model.  Useful for deleting a child and not wanting them to come back as linked.
	 *
	 * @param $linkname
	 */
	public function resetLink($linkname) {
		$idx = $this->_getLinkIndex($linkname);
		if($idx === null){
			return null; // @todo Error Handling
		}

		$this->_linked[$idx]->records = null;
		$this->_linked[$idx]->purged = null;
	}

	/**
	 * Reset all linked models
	 *
	 * This is typically not required, but can be useful.
	 *
	 * Will clear the data cache of linked Models from this parent
	 * to allow them to be reimported, saved manually, etc.
	 *
	 */
	public function resetAllLinks() {
		foreach(array_keys($this->_linked) as $idx) {
			$this->_linked[$idx]->records = null;
			$this->_linked[$idx]->purged = null;
		}
	}

	/**
	 * Mark a linked model for deletion.
	 * Doesn't actually delete the linked model until this element is saved.
	 *
	 * @param Model $link
	 * @return bool
	 */
	public function deleteLink(Model $link){
		// Since I don't get the linkname like usual ones...
		foreach($this->_linked as $idx => $linkset){
			if($linkset->records === null){
				continue;
			}

			if(is_array($linkset->records)){
				foreach($linkset->records as $k => $rec){
					if($rec === $link){
						if($this->_linked[ $idx ]->purged === null){
							$this->_linked[ $idx ]->purged = [];
						}

						$this->_linked[ $idx ]->purged[] = $link;
						unset($this->_linked[$idx]->records[$k]);
						return true;
					}
				}
			}
			elseif($linkset->records === $link) {
				if($this->_linked[ $idx ]->purged === null){
					$this->_linked[ $idx ]->purged = [];
				}

				$this->_linked[ $idx ]->purged[] = $link;
				$this->_linked[ $idx ]->records = null;
				return true;
			}
		}

		return false;
	}

	/**
	 * Get if the given link by name has changed.
	 *
	 * This accounts for a newly created one, deleted one, or simply modified link.
	 *
	 * Also handles 1-M and 1-1 links
	 *
	 * @param string $linkname The link, (by name), to get if changed.
	 *
	 * @return bool
	 */
	public function changedLink($linkname){
		$idx = $this->_getLinkIndex($linkname);
		if($idx === null){
			return false; // @todo Error Handling
		}

		if($this->_linked[$idx]->records === null){
			// If this link is not even loaded, then it can't have been changed.
			return false;
		}

		if($this->_linked[ $idx ]->purged !== null){
			// If there is a deleted IDX and it's set to an array, then at least one link was deleted, CHANGE!
			return true;
		}

		if(is_array($this->_linked[$idx]->records)) {
			foreach($this->_linked[$idx]->records as $subm){
				if($subm->changed()){
					// There is at least one changed sub model!  CHANGE!
					return true;
				}
			}
		}
		elseif($this->_linked[$idx]->records instanceof Model){
			if($this->_linked[$idx]->records->changed()){
				// I think this can happen.... :/
				return true;
			}
		}

		// Shrugs!
		return false;
	}


	/**
	 * Render this Model object with the requested template type
	 *
	 * @param string $template The template base name to render this Model.
	 *
	 * @return string
	 */
	public function render(string $template = null) : string {
		if($template === null) {
			// If no format was requested, just return the Model's label
			return $this->getLabel();
		}

		// Convert it to a standard format and remove any navigatable characters.
		$template = str_replace(
			['-', '/'],
			['_', ''],
				strtolower($template)
		);

		try {
			// Render the requested template view.
			$tpl = new Template('models/' . $this->getClassKeyName() . '/' . $template);
		}
		catch(Exception $ex) {
			try {
				// Try the generic instead.
				$tpl = new Template('models/generic/' . $template);
			}
			catch(Exception $ex) {
				return '<p class="message-error">No template available for ' . $this->getClassKeyName() . '!</p>';
			}
		}

		try {
			$tpl->assign('model', $this);
			$html = $tpl->fetch();
			return $html;
		}
		catch(Exception $ex) {
			// Template rendering failed, just grab the label instead.
			log_exception($ex);
			return $this->getLabel();
		}
	}



	/**
	 * Set properties on this model from a form object, optionally with a specific prefix.
	 *
	 * @param Form        $form   Form object to pull data from
	 * @param string|null $prefix Prefix that all keys should be matched to, (optional)
	 */
	public function setFromForm(\Core\Forms\Form $form, $prefix = null){

		// Get every "prefix[...]" element, as they key up 1-to-1.
		$els = $form->getElements(true, false);
		foreach ($els as $e) {
			// If a specific prefix was requested and this element does not match up, skip it!
			if ($prefix){
				if(!preg_match('/^' . $prefix . '\[(.*?)\].*/', $e->get('name'), $matches)) continue;
				$key = $matches[1];
			}
			else{
				$key = $e->get('name');
			}

			$this->set($key, $e->get('value'));
		}
	}

	/**
	 * Converse to setFromForm, this method is called on each form element created when calling addModel or BuildFromModel.
	 *
	 * Any special instructions for your model's elements can go here, simply extend this method and add logic as necessary.
	 *
	 * @param             $key
	 * @param \Core\Forms\FormElement $element
	 */
	public function setToFormElement($key, \Core\Forms\FormElement $element){
		// This method left intentionally blank.
		// If custom logic is required here, extend this method in your model and do so there.
	}

	/**
	 * Method that is called on the model after "addModel" is called on a form.
	 *
	 * Any special logic such as adding custom elements from the model can be done here, simply extend this method and add logic as necessary.
	 *
	 * @param Form   $form
	 * @param string $prefix
	 */
	public function addToFormPost(\Core\Forms\Form $form, $prefix){
		// This method left intentionally blank.
		// If custom logic is required here, extend this method in your model and do so there.
	}

	/**
	 * Function to call to decrypt data from this model.
	 *
	 * As of 5.1.0, this is called automatically and therefores this does nothing.
	 */
	public function decryptData(){
		return;
	}

	/**
	 * Get the table name for this class
	 *
	 * @return null|string
	 */
	public function _getTableName(){
		return self::GetTableName();
	}

	/**
	 * Get the primary key value(s) of this model as a string
	 *
	 * @return string
	 */
	public function getPrimaryKeyString(){
		$bits = [];
		$i = self::GetIndexes();

		if(isset($i['primary'])){
			// It should be an array, but doesn't have to be.
			$pri = $i['primary'];
			if(!is_array($pri)) $pri = [$pri];

			foreach ($pri as $k) {
				$val = $this->get($k);
				if ($val === null) $val = 'null';
				elseif ($val === false) $val = 'false';
				$bits[] = $val;
			}
		}

		return implode('-', $bits);
	}

	// <editor-fold defaultstate="collapsed" desc="Magic PUBLIC Methods">

	/**
	 * Get this model as a string
	 *
	 * @return string
	 */
	public function __toString(){
		return $this->getLabel();
	}

	/**
	 * Magic method to allow accessing this Model's properties directly
	 *
	 * @param string $key
	 * @return mixed
	 */
	public function __get($key) {
		return $this->get($key, null);
	}

	/**
	 * Magic method to allow manipulating this Model as a property
	 *
	 * @param string $key
	 * @param mixed $value
	 */
	public function __set($key, $value) {
		return $this->set($key, $value);
	}

	/**
	 * Whether an offset exists
	 *
	 * @link http://php.net/manual/en/arrayaccess.offsetexists.php
	 *
	 * @param mixed $offset An offset to check for.
	 *
	 * @return boolean Returns true on success or false on failure.
	 */
	public function offsetExists($offset) {
		return (array_key_exists($offset, $this->_columns));
	}

	/**
	 * Offset to retrieve
	 *
	 * Alias of Model::get()
	 *
	 * @link http://php.net/manual/en/arrayaccess.offsetget.php
	 *
	 * @param mixed $offset The offset to retrieve.
	 *
	 * @return mixed Can return all value types.
	 */
	public function offsetGet($offset) {
		return $this->get($offset);
	}

	/**
	 * Offset to set
	 *
	 * Alias of Model::set()
	 *
	 * @link http://php.net/manual/en/arrayaccess.offsetset.php
	 *
	 * @param mixed $offset The offset to assign the value to.
	 * @param mixed $value The value to set.
	 *
	 * @return void
	 */
	public function offsetSet($offset, $value) {
		$this->set($offset, $value);
	}

	/**
	 * Offset to unset
	 *
	 * This just sets the value to null.
	 *
	 * @link http://php.net/manual/en/arrayaccess.offsetunset.php
	 *
	 * @param mixed $offset The offset to unset.
	 *
	 * @return void
	 */
	public function offsetUnset($offset) {
		$this->set($offset, null);
	}

	// </editor-fold>



	/**
	 * Go through any linked tables and update them if the linking key has been changed.
	 *
	 * This needs to actually go through the database and update the saved keys if necessary.
	 *
	 * @param string $key
	 * @param mixed $newval
	 */
	protected function _setLinkKeyPropagation($key, $newval) {
		// If this model does not exist yet, there will be no information linked in the database.
		$exists = $this->exists();

		foreach ($this->_linked as $lk => $link) {

			// I can't change my parent data.
			// NO CHANGING YOUR PARENTS!
			if(
				$link->type === Model::LINK_BELONGSTOMANY ||
				$link->type === Model::LINK_BELONGSTOONE
			){
				continue;
			}

			$dolink = false;
			// I can't use the getLinkWhereArray function, because that will resolve the key.
			// I need the key itself.
			foreach ($link->on as $k => $v) {
				if(
					(is_numeric($k) && $v === $key) ||
					(!is_numeric($k) && $k === $key)
				){
					$dolink = true;
				}
			}

			// $dolink should now be true/false.  If true I need to load that linked table and set the new key.
			if (!$dolink){
				continue;
			}

			if($exists){
				// Get the data and update it!
				$links = $this->getLink($lk);
				if (!is_array($links)) $links = [$links];

				foreach ($links as $model) {
					$model->set($key, $newval);
				}
			}
			else{
				if(!$link->records){
					// Only update the cached data, as nothing has been saved, so the database doesn't have anything.
					continue;
				}

				if(is_array($link->records)) {
					// Standard 1-to-M structure
					foreach($link->records as $model){
						$model->set($key, $newval);
					}
				}
				else{
					// Only a 1-to-1 structure
					$link->records->set($key, $newval);
				}
			}
		}
	}

	/**
	 * Get the fully resolved ClassName of the requested link name.
	 *
	 * @param string $linkname Name of one of the linked models.
	 *
	 * @return null|string
	 */
	protected function _getLinkClassName($linkname) {

		$idx = $this->_getLinkIndex($linkname);
		if($idx === null){
			return null; // @todo Error Handling
		}

		// Determine the class.
		$c = $this->_linked[$idx]->model;

		if (!is_subclass_of($c, '\\Core\\Model')){
			return null; // @todo Error Handling
		}

		return $c;
	}

	/**
	 * Called internally by the save() method for new records.
	 */
	protected function _saveNew($defer = false) {
		$i = self::GetIndexes();
		$s = self::GetSchema();
		$n = $this->_getTableName();

		if (!isset($i['primary'])) $i['primary'] = []; // No primary schema defined... just don't make the in_array bail out.

		if($defer){
			$inserts = []; // key => value map for this model
			if(!isset(self::$_DeferInserts[$n])){
				$dat = new Dataset();
				$dat->table($n);
				$dat->_mode = Dataset::MODE_BULK_INSERT;
				self::$_DeferInserts[$n] = [
					'dataset' => $dat,
					'interface' => $this->_interface,
				];
			}
			else{
				$dat = self::$_DeferInserts[$n]['dataset'];
			}
		}
		else{
			$dat = new Dataset();
			$dat->table($n);
		}

		$idcol = false;

		foreach($this->_columns as $c){
			/** @var \Core\Datamodel\Columns\SchemaColumn $c */
			if($c->type == Model::ATT_TYPE_UUID){
				if($c->value && $c->valueDB){
					// Yay, a UUID is already set, no need to really do much.
					// It's already set and this will most likely be ignored, but may not be for UPDATE statements...
					// although there shouldn't be any update statements here.... but ya never know
					if($defer){
						$inserts[$c->field] = $c->getInsertValue();
					}
					else{
						$dat->setID($c->field, $c->getInsertValue());
					}
				}
				else{
					// a UUID is already set, but it doesn't exist yet still.
					// This means that the UUID was set externally even though the record is new.
					// THIS IS ALLOWED!
					// Insert it as typical key.
					// Addtionally if this is a new key, the column will automatically generate a UUID as necessary.
					if($defer){
						$inserts[$c->field] = $c->getInsertValue();
					}
					else{
						$dat->insert($c->field, $c->getInsertValue());
						$dat->setID($c->field, $c->getInsertValue());
					}
				}
				// Remember this for after the save.
				$idcol = $c->field;
			}
			elseif($c->type == Model::ATT_TYPE_ID){
				if($c->value){
					// An ID is already set on this key, even though it's an auto-increment ID.
					// Allow this as you may be syncing a model from another system and want their IDs to match up.
					// NOTE, this can be a dangerous operation.
					if($defer){
						$inserts[$c->field] = $c->getInsertValue();
					}
					else{
						$dat->insert($c->field, $c->getInsertValue());
					}
				}
				if(!$defer){
					$dat->setID($c->field, $c->getInsertValue());
					// Remember this for after the save.
					$idcol = $c->field;
				}
			}
			elseif($c->type == Model::ATT_TYPE_META){
				// Skip these entirely.
				continue;
			}
			else{
				if($defer){
					$inserts[$c->field] = $c->getInsertValue();
				}
				else{
					$dat->insert($c->field, $c->getInsertValue());
				}
			}
		}
//var_dump($dat); die();
		if($defer) {
			$dat->bulkInsert($inserts);
		}
		else{
			$dat->execute($this->_interface);
			if ($idcol){
				$this->_columns[$idcol]->setValueFromDB($dat->getID());
			}
		}
	}

	/**
	 * Save an existing Model object into the database.
	 * Will create, set and execute a dataset object as appropriately internally.
	 *
	 * @param boolean $useset Set to true to have this model use an INSERT_UPDATE statement instead of just UPDATE.
	 * @return bool
	 */
	protected function _saveExisting($useset = false) {

		// If this model doesn't have any changes, don't save it!
		if(!$this->checkIsChanged()) {
			return false;
		}

		$i = self::GetIndexes();
		$s = self::GetSchema();
		$n = $this->_getTableName();
		$updateColumn = null;

		// No primary schema defined or it's a string, (single value)... just don't make the in_array bail out.
		$pri = isset($i['primary']) ? $i['primary'] : [];

		if($pri && !is_array($pri)) {
			$pri = [$pri];
		}


		// This is the dataset object that will be integral in this function.
		$dat = new Dataset();
		$dat->table($n);

		foreach($this->_columns as $c){
			/** @var \Core\Datamodel\Columns\SchemaColumn $c */

			if($c->type === Model::ATT_TYPE_META){
				// Skip these entirely.
				continue;
			}

			if($c->type === Model::ATT_TYPE_ID || $c->type === Model::ATT_TYPE_UUID) {
				// IDs and UUIDs get set as the ID, this is a special property in the datamodel.
				$dat->setID($c->field, $c->value);
			}
			elseif (in_array($c->field, $pri)) {
				// Any field keyed as a primary field sets the WHERE clause.
				// This needs to use the valueDB to ensure it matches exactly what the database
				// state was at the time of reading.
				$dat->whereEquals($c->field, $c->valueDB);
			}
			elseif($c->type === Model::ATT_TYPE_UPDATED) {
				// Remember this column for after; we'll only update the value if something else was changed on this model
				$updateColumn = $c;
				continue;
			}

			if($c->changed()){
				// For any column changed, the database will get the new updated value.
				$dat->update($c->field, $c->getUpdateValue());
			}
		}


		if(!count($dat->_updates)) {
			// No data.. nothing to change I guess!
			// This is done to increase database performance by attempting to issue a query that would in fact change nothing.
			return false;
		}

		if($updateColumn !== null) {
			// If it's still here and there's an update column request, ensure that's updated!
			// This is kept seperate from the above logic to not allow the updated column to trigger the update clause,
			// (eg, if nothing changed, don't save it, as the updated flag always changes).
			$dat->update($updateColumn->field, $updateColumn->getUpdateValue());
		}

		$dat->execute($this->_interface);

		return true;
	}

	/**
	 * Get the where array of criteria for a given link.
	 * Useful for manually tweaking the clause.
	 *
	 * @param string $linkname
	 *
	 * @return array|null
	 */
	protected function _getLinkWhereArray($linkname) {
		$idx = $this->_getLinkIndex($linkname);
		if($idx === null){
			return null; // @todo Error Handling
		}

		// Build a standard where criteria that can be used throughout this function.
		$wheres = [];

		foreach ($this->_linked[$idx]->on as $k => $v) {
			$wheres[ $k ] = $this->get($v);
		}


		// Pages have a special extra here.  If it's enterprise/multisite mode, enforce that relationship.
		if($linkname === 'Page' && Core::IsComponentAvailable('multisite') && MultiSiteHelper::IsEnabled()){
			// See if there's a site column on this schema.  If there is, enforce that binding too!
			$schema = self::GetSchema();
			if(isset($schema['site']) && $schema['site']['type'] == Model::ATT_TYPE_SITE){
				$wheres['site'] = $this->get('site');
			}
		}


		return $wheres;
	}

	/**
	 * Translate a link name, (be it full Model name, partial model name, or linked key name), to the index in _linked.
	 *
	 * @param string $name
	 *
	 * @return int|null
	 */
	protected function _getLinkIndex($name){
		if(isset($this->_linkIndexCache[$name])){
			return $this->_linkIndexCache[ $name ];
		}

		foreach($this->_linked as $idx => $link){
			if($idx === $name){
				// May happen if the index is passed in, (could happen internally).
				return $idx;
			}
			elseif(
				$link->key === $name ||
				$link->model === $name ||
				$link->model === $name . 'Model'
			){
				$this->_linkIndexCache[$name] = $idx;
				return $idx;
			}
		}

		// No matches?
		$this->_linkIndexCache[$name] = null;
		return null;
	}

	protected function _getCacheKey() {
		if (!$this->_cacheable) return false;
		$i = self::GetIndexes();

		if (!(isset($i['primary']) && sizeof($i['primary']))) return false;

		$keys = $this->getPrimaryKeyString();

		return 'DATA:' . self::GetTableName() . ':' . $keys;
	}


	/*************************************************************************
	 ****                    FACTORY-RELATED STATIC METHODS               ****
	 *************************************************************************/

	/**
	 * Constructor alternative that utilizes caching to save on database lookups.
	 *
	 * Since this caches the model in memory, it is ill-advised to use this for very large numbers of records.
	 * Around 50k records stored in memory, it'll consume about 256MB of RAM.
	 *
	 * @static
	 *
	 * @param $keys
	 * @return static
	 */
	public static function Construct(...$keys){
		$class = get_called_class();

		// New ones do not get cached.
		if($keys === null){
			return new $class();
		}

		$cache = '';
		foreach(func_get_args() as $a){
			$cache .= $a . '-';
		}
		$cache = substr($cache, 0, -1);

		if(!isset(self::$_ModelCache[$class])){
			self::$_ModelCache[$class] = [];
		}
		if(!isset(self::$_ModelCache[$class][$cache])){
			$reflection = new ReflectionClass($class);
			/** @var $obj Model */
			$obj = $reflection->newInstanceArgs(func_get_args());

			self::$_ModelCache[$class][$cache] = $obj;
		}

		//var_dump($cache);
		return self::$_ModelCache[$class][$cache];
	}

	/**
	 * Shortcut method to find instances of this Model that match a given where clause.
	 *
	 * If limit is set to 1, either a Model or null is returned.
	 * Else, an array of Models is returned, be it populated or empty.
	 *
	 * The WHERE clause expects everything to be a string,
	 * so DO NOT USE THIS for complex datatypes or anything requiring the use of NULL, ints, etc!
	 *
	 * @static
	 * @param array|string    $where Where clause
	 * @param int|string|null $limit Limit clause
	 * @param string|null     $order Order clause
	 *
	 * @return array|null|static
	 */
	public static function Find($where = [], $limit = null, $order = null) {

		if($where === null) {
			// $fac->where expects this to be an array, even if it's empty.
			$where = [];
		}

		$classname = get_called_class();

		if(
			(is_array($where) && !count($where) || $where === null) &&
			$limit === null &&
			$order === null
		){
			// Try to cache them :p

			if(!isset(self::$_ModelFindCache[$classname])){
				$fac = new ModelFactory($classname);
				self::$_ModelFindCache[$classname] = $fac->get();
			}

			return self::$_ModelFindCache[$classname];
		}

		$fac = new ModelFactory($classname);
		$fac->where($where);
		$fac->limit($limit);
		$fac->order($order);
		//var_dump($fac);
		return $fac->get();
	}

	/**
	 * Get all records of this Model type as a set of options that can be used with a select box.
	 *
	 * This can be extended in the specific Model if additional functionality is required;
	 * this is simply a default scaffolding that may not work on all instances.
	 *
	 * @static
	 *
	 * @return array
	 */
	public static function GetAllAsOptions() : array {

		$classname = get_called_class();
		$ref = new ReflectionClass($classname);

		// Execute Find using default parameters.
		// (That method will cache results by default).
		$results = $ref->getMethod('Find')->invoke(null);
		$idx = $ref->getMethod('GetIndexes')->invoke(null);

		if(!isset($idx['primary'])){
			return ['' => 'Unable to automatically get ' . $classname . ' as options because no primary key defined!'];
		}

		if(sizeof($idx['primary']) > 1){
			return ['' => 'Unable to automatically get ' . $classname . ' as options because primary key defined as multiple columns!'];
		}

		$id = $idx['primary'][0];

		$options = [];

		foreach($results as $res){
			/** @var Model $res */
			$options[ $res->get($id) ] = $res->getLabel();
		}

		return $options;
	}

	/**
	 * Factory shortcut function to do a search for the specific records and return them as a raw array.
	 *
	 * @static
	 *
	 * @param array $where
	 * @param null  $limit
	 * @param null  $order
	 *
	 * @return array
	 */
	public static function FindRaw($where = [], $limit = null, $order = null) {
		$fac = new ModelFactory(get_called_class());
		$fac->where($where);
		$fac->limit($limit);
		$fac->order($order);
		//var_dump($fac);
		return $fac->getRaw();
	}

	/**
	 * Get a count of records that match a given where criteria
	 *
	 * @static
	 *
	 * @param array $where
	 *
	 * @return int
	 */
	public static function Count($where = []) {
		$fac = new ModelFactory(get_called_class());
		$fac->where($where);
		return $fac->count();
	}

	/**
	 * Perform a model search on the records of this Model.
	 *
	 * @param string $query The base query to search
	 * @param array $where  Any additional where parameters to add onto the factory
	 *
	 * @return Model[]
	 */
	public static function Search(string $query, array $where = []) : array {
		$ret = [];

		// If this object does not support searching, simply return an empty array.
		$ref = new ReflectionClass(get_called_class());

		if(!$ref->getProperty('HasSearch')->getValue()){
			return $ret;
		}

		$fac = new ModelFactory(get_called_class());

		if(sizeof($where)){
			$fac->where($where);
		}

		if($ref->getProperty('HasDeleted')->getValue()){
			$fac->where('deleted = 0');
		}

		$fac->search($query);

		$ret = $fac->get();

		return $ret;
	}

	/*************************************************************************
	 ****                    OTHER STATIC METHODS                         ****
	 *************************************************************************/

	/**
	 * Method to encrypt a specific key for storage.
	 *
	 * Called internally by the set function.
	 * Will return the encrypted data.
	 *
	 * @param mixed $value The plain-text value to encrypt
	 * @return string Encrypted data
	 */
	public static function EncryptValue($value){
		$cipher = 'AES-256-CBC';
		$passes = 10;
		$size = openssl_cipher_iv_length($cipher);
        if(function_exists('mcrypt_create_iv')) {
			$iv = mcrypt_create_iv($size, MCRYPT_RAND);
		}
		else {
			$iv = random_bytes($size);
		}

		if($value === '') return '';
		elseif($value === null) return null;

		$enc = $value;
		for($i=0; $i<$passes; $i++){
			$enc = openssl_encrypt($enc, $cipher, SECRET_ENCRYPTION_PASSPHRASE, true, $iv);
		}

		$payload = '$' . $cipher . '$' . str_pad($passes, 2, '0', STR_PAD_LEFT) . '$' . $enc . $iv;
		return $payload;
	}

	/**
	 * Decrypt a given value, can be called internally or externally.
	 *
	 * @param $payload
	 *
	 * @return null|string
	 */
	public static function DecryptValue($payload) {
		if($payload === null || $payload === '' || $payload === false){
			return null;
		}

		preg_match('/^\$([^$]*)\$([0-9]*)\$(.*)$/m', $payload, $matches);

		$cipher = $matches[1];
		$passes = $matches[2];
		$size = openssl_cipher_iv_length($cipher);
		// Now I can trim off the beginning crap from the encrypted string.
		$dec = substr($payload, strlen($cipher) + 5, 0-$size);
		$iv = substr($payload, 0-$size);

		for($i=0; $i<$passes; $i++){
			$dec = openssl_decrypt($dec, $cipher, SECRET_ENCRYPTION_PASSPHRASE, true, $iv);
		}

		return $dec;
	}

	/**
	 * Get the table name for a given Model object
	 *
	 * @static
	 *
	 * @throws Exception
	 *
	 * @return string
	 */
	public static function GetTableName() : string {
		// Just a lookup table for rendered table names.
		// This is useful so the regex functions don't have to run more than once.
		static $_tablenames = [];
		$m = get_called_class();

		// Generic models cannot have tables.
		if ($m === 'Model'){
			throw new Exception('Invalid class name requested');
		}

		if(strpos($m, '\\') !== false){
			throw new Exception('Invalid class name requested');
		}

		if (!isset($_tablenames[$m])) {
			// Calculate the table class.

			// It's based on the main class's name.
			$tbl = $m;

			// If it ends in Model... trim that bit off.  It's assumed that the prefix is what we want.
			if(strrpos($tbl, 'Model') === strlen($tbl) - 5){
				$tbl = substr($tbl, 0, -5);
			}

			// Replace any capitalized letters with a _[letter].
			$tbl = preg_replace('/([A-Z])/', '_$1', $tbl);

			// Of course this would produce something similar to _Foo_Mep_Blah.. don't need the beginning _.
			if ($tbl[0] == '_'){
				$tbl = substr($tbl, 1);
			}

			// And lowercase.
			$tbl = strtolower($tbl);

			// Prepend the DB_PREFIX and save!
			$_tablenames[$m] = (defined('DB_PREFIX') ? DB_PREFIX : '') . $tbl;
		}

		return $_tablenames[$m];
	}

	public static function GetSchema($cloned = false) : ModelSchema {
		/** @var string $classname The class name of the extending class. */
		$classname = get_called_class();

		if(!isset(self::$_ModelSchemaCache[$classname])){
			self::$_ModelSchemaCache[$classname] = new ModelSchema($classname);
		}

		return $cloned ?
			clone self::$_ModelSchemaCache[$classname] :
			self::$_ModelSchemaCache[$classname];
	}

	/**
	 * Get the resolved schema for this Model type.
	 *
	 * This is called by several other methods, including getKeySchemas and getKeySchema.
	 *
	 * @throws Exception
	 *
	 * @return array
	 */
	public static function PrevGetSchema() : array {
		/** @var string $classname The class name of the extending class. */
		$classname = strtolower(get_called_class());

		if(!isset(self::$_ModelSchemaCache[$classname])){
			self::$_ModelSchemaCache[$classname] = new ModelSchema($classname);
		}

		return self::$_ModelSchemaCache[$classname]->definitions;
	}

	/**
	 * Internally used method to add a supplemental model to the base model.
	 *
	 * Used to allow components to append the database of another component!
	 *
	 * @param string $original
	 * @param string $supplemental
	 */
	public static function AddSupplemental($original, $supplemental){
		if(!isset(self::$ModelSupplementals[$original])){
			self::$ModelSupplementals[$original] = [];
		}

		self::$ModelSupplementals[$original][] = $supplemental;

		// Clear the cache so that the next check for this model contains the supplemental data.
		if(isset(self::$_ModelSchemaCache[$original])){
			self::$_ModelSchemaCache[$original] = null;
		}
	}

	public static function GetIndexes() {
		/** @var string $classname The class name of the extending class. */
		$classname = get_called_class();

		// First I need the parent's class, if it's not Model.
		$parent = get_parent_class($classname);
		if($parent != 'Model'){
			// Because the "Model" class doesn't have a schema... that's up to classes that extend it.
			$parentref = new ReflectionClass($parent);
			$ref = new ReflectionClass($classname);
			return array_merge(
				$parentref->getProperty('Indexes')->getValue(),
				$ref->getProperty('Indexes')->getValue()
			);
		}
		else{
			// Because the "Model" class doesn't have a schema... that's up to classes that extend it.
			$ref = new ReflectionClass($classname);
			return $ref->getProperty('Indexes')->getValue();
		}
	}

	public static function CommitSaves(){
		/** @var string $classname The class name of the extending class. */
		$classname = get_called_class();

		$tableName = self::GetTableName();

		if(!isset(self::$_DeferInserts[$tableName])){
			return;
		}

		/** @var Dataset $dat */
		$dat = self::$_DeferInserts[$tableName]['dataset'];
		$interface = self::$_DeferInserts[$tableName]['interface'];
		$dat->execute($interface);

		unset(self::$_DeferInserts[$tableName]);
	}

	/**
	 * @param array $schema
	 *
	 * @return array
	 */
	private static function _StandardizeSchemaDefinition($schema){
		// These are all defaults for schemas.
		// Setting them to the default if they're not set will ensure that
		// 'undefined index' notices are not incurred.
		if (!isset($schema['type']))               $schema['type']      = Model::ATT_TYPE_TEXT; // Default if not present.
		if (!isset($schema['maxlength']))          $schema['maxlength'] = false;
		if (!isset($schema['null']))               $schema['null']      = false;
		if (!isset($schema['comment']))            $schema['comment']   = false;
		if (!array_key_exists('default', $schema)) $schema['default']   = false;
		if (!isset($schema['encrypted']))          $schema['encrypted'] = false;
		if (!isset($schema['required']))           $schema['required']  = false;
		if (!isset($schema['encoding']))           $schema['encoding']  = false;
		if (!isset($schema['formatter']))          $schema['formatter'] = false;

		if($schema['default'] === false && $schema['null'] === true){
			// Easiest case!  NULL is allowed on this column.
			$schema['default'] = null;
		}

		if($schema['type'] == Model::ATT_TYPE_ENUM){
			// Enums have an options array!
			$schema['options'] = isset($schema['options']) ? $schema['options'] : [];
		}
		else{
			// Other fields don't.
			$schema['options'] = false;
		}

		return $schema;
	}
}
