<?php


namespace Core;

use Core\Filestore\Backends\FileRemote;
use Core\Filestore\Factory as FileFactory;
use Core\GPG\GPG;
use Exception;

/**
 * Meta data for a given Package based on a remote update site
 *
 * Used by the updater for retrieving new packages, before the actual
 * package has been downloaded.
 *
 * @package Core
 */
class PackageMeta {
	/** @var string */
	public $type;
	/** @var string */
	public $key;
	/** @var string */
	public $name;
	/** @var string */
	public $version;
	/** @var string */
	public $gpgKey;
	/** @var string */
	public $logo;
	/** @var string */
	public $url;
	/** @var string */
	public $download;
	/** @var string */
	public $description;
	/** @var string */
	public $packager;
	/** @var string */
	public $email;
	/** @var int */
	public $released;
	/** @var string|null */
	public $compatibility;
	/**
	 * Contains arrays of type, name, version, and operation
	 * @var ComponentRequirement[]
	 */
	public $requires = [];
	/**
	 * Contains arrays of type, name, and version
	 * @var ComponentProvide[]
	 */
	public $provides = [];
	/**
	 * Flat string of supported upgrade paths
	 * @var string[]
	 */
	public $upgrades = [];

	private $_downloadedFile;

	/**
	 * Get the local version of the component provided in this package
	 * or NULL if not available
	 *
	 * @return Component|null
	 */
	public function getLocalComponent() : ?Component {
		if($this->type === 'core') {
			return Core::GetComponent('core');
		}
		elseif($this->type === 'component') {
			return Core::GetComponent($this->key);
		}
		elseif($this->type === 'theme') {
			// Themes are difficult because it doesn't quite follow the standard rules.
			$allThemes = Theme::GetAllThemes();
			foreach($allThemes as $theme) {
				if($theme->getKeyName() === $this->key) {
					return $theme;
				}
			}
		}

		// Something needs to be returned.
		return null;
	}

	/**
	 * Get the base directory for this package
	 *
	 * @return string
	 */
	public function getBaseDirectory() : string {
		switch($this->type) {
			case 'core':
				return ROOT_PDIR;
			case 'component':
				return ROOT_PDIR . 'components/' . $this->key . '/';
			case 'theme':
				return ROOT_PDIR . 'themes/' . $this->key . '/';
			default:
				return '';
		}
	}

	/**
	 * Download the actual package to a local temp file
	 *
	 * @return Package
	 */
	public function downloadPackage() : Package {
		if(!$this->_downloadedFile) {
			$file = new FileRemote($this->download);

			// Set the license information if available.
			if(defined('SERVER_ID') && strlen(SERVER_ID) == 32) {
				$file->setRequestHeader('X-Core-Server-ID', SERVER_ID);
			}

			$this->_downloadedFile = $file->getLocalFile();
		}

		return Package::FromLocalFile($this->_downloadedFile);
	}

	/**
	 * Check if this package is installed locally
	 *
	 * @return bool
	 */
	public function checkIsInstalledLocally() : bool {
		return $this->getLocalComponent() !== null;
	}

	/**
	 * Check if there's an updated version available
	 *
	 * @return bool
	 */
	public function checkIsUpdateAvailable() : bool {
		$component = $this->getLocalComponent();

		if($component === null) {
			// Not locally installed, no update.
			return false;
		}

		return version_compare($component->getVersion(), $this->version, 'lt');
	}

	/**
	 * Simple check if the GPG key is currently available in the local system
	 *
	 * @return bool
	 */
	public function checkIsGPGKeyAvailable() : bool {
		try {
			$gpg = new GPG();
			return $gpg->getKey($this->gpgKey) !== null;
		}
		catch(Exception $ex) {
			\Core\log_exception($ex);
			return false;
		}
	}

	public static function FromJSON(array $source) : PackageMeta {
		$meta = new self();

		$meta->type = $source['type'];
		$meta->key = $source['key'];
		$meta->name = $source['name'];
		$meta->version = $source['version'];
		$meta->gpgKey = $source['gpg_key'];
		$meta->logo = $source['logo'];
		$meta->url = $source['url'];
		$meta->download = $source['download'];
		$meta->description = $source['description'];
		$meta->packager = $source['packager'];
		$meta->email = $source['email'];
		$meta->released = (int)$source['released'];
		$meta->compatibility = $source['compatibility'];
		$meta->upgrades = $source['upgrades'];

		foreach($source['requires'] as $require) {
			$meta->requires[] = new ComponentRequirement(
				$require['type'],
				$require['name'],
				$require['version'],
				$require['operation']
			);
		}

		foreach($source['provides'] as $provide) {
			$meta->provides[] = new ComponentProvide(
				$provide['type'],
				$provide['name'],
				$provide['version']
			);
		}

		return $meta;
	}
}