<?php


namespace Core;


use Exception;
use UpdateSiteModel;

class Updater {

	/** @var PackageMeta[] */
	private $_packages;

	public function getPackage(string $type, string $keyName) : ?PackageMeta {
		if($this->_packages === null) {
			// Ensure the cache is ready.
			$this->getAllPackages();
		}

		return $this->_packages[ $type . '-' . $keyName ] ?? null;
	}

	/**
	 * Get the queue of installation packages required for a given type and name
	 *
	 * This will auto resolve dependencies in the correct order.
	 *
	 * @param string $type
	 * @param string $keyName
	 *
	 * @throws Exception
	 *
	 * @return PackageMeta[]
	 */
	public function getInstallQueue(string $type, string $keyName) : array {
		$queue = [];
		$queueKeys = [];
		$package = $this->getPackage($type, $keyName);
		if(!$package) {
			// Original package not located, nothing to check.
			return [];
		}

		$pendingQueue = [
			$type . '-' . $keyName => $package
		];

		do {
			// Check the queue requirements while it's present and has changed since the last run.
			$pendingQueueSize = count($pendingQueue);
			foreach($pendingQueue as $key => $package) {
				$passed = true;

				foreach($package->requires as $requirement) {
					try {
						$requirement->check();
					}
					catch(Exception $ex) {
						// If the failure is from another component, we can try to resolve that.
						$supplied = $this->_resolveDependency($requirement);

						if(!$supplied) {
							// No package located to satisfy this requirement, unable to proceed.
							throw new Exception($requirement->getLabel());
						}

						$queueKey = $supplied->type . '-' . $supplied->key;
						if(!in_array($queueKey, $queueKeys)) {
							// Not already in queue, add this package for the next run and recheck with it.
							$pendingQueue[ $queueKey ] = $supplied;
							$passed = false;
						}
					}
				}

				if($passed) {
					$queue[] = $package;
					$queueKeys[] = $key;
					unset($pendingQueue[ $key ]);
				}
			}
		}
		while($pendingQueueSize != count($pendingQueue));

		return $queue;
	}

	/**
	 * Get all packages from remote update sites
	 *
	 * @return PackageMeta[]
	 */
	public function getAllPackages() : array {
		if($this->_packages === null) {
			$this->_packages = [];

			$updateSites = UpdateSiteModel::Find();
			foreach($updateSites as $site) {
				/** @var UpdateSiteModel $site */

				if(!$site->checkIsValid()) {
					// @todo should there be error reporting for invalid sites?
					continue;
				}

				// Retrieve the contents of this remote feed and parse it for data.
				$updates = json_decode($site->getFile()->getContents(), true);
				if(!$updates) {
					// @todo error reporting for invalid response
					continue;
				}

				if(!isset($updates['packages'])) {
					// @todo error reporting for invalid response
					continue;
				}

				foreach($updates['packages'] as $raw) {
					$package = PackageMeta::FromJSON($raw);
					$key = $package->type . '-' . $package->key;

					if(
						isset($ret[$key]) &&
						version_compare($this->_packages[$key]->version, $package->version, 'ge')
					) {
						// This package is provided in more than one repo,
						// only include the most updated version.
						continue;
					}

					$this->_packages[$key] = $package;
				}
			}
		}

		return array_values($this->_packages);
	}

	private function _resolveDependency(ComponentRequirement $requirement) : ?PackageMeta {
		if($requirement->getType() === 'component') {
			$package = $this->getPackage('component', $requirement->getName());

			if($package && $requirement->getVersion()) {
				// Check the version if it's set too.
				$op = $requirement->getOperation() ?? 'ge';
				if(!version_compare($package->version, $requirement->getVersion(), $op)) {
					return null;
				}
			}

			return $package;
		}
		elseif($requirement->getType() === 'library') {
			// This one is more difficult, as I need to sift through the packages and see which one offers this library.
			foreach($this->_packages as $package) {
				foreach($package->provides as $provide) {
					if($provide->name === $requirement->getName()) {
						if($requirement->getVersion()) {
							$op = $requirement->getOperation() ?? 'ge';
							if(!version_compare($provide->version, $requirement->getVersion(), $op)) {
								return null;
							}
						}

						return $package;
					}
				}
			}
		}
		else {
			return null;
		}
	}
}