<?php
/**
 * Systems for the control widgets used for page-level operations and inline operations.
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @since 2.1.2
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;

use ArrayAccess;
use Iterator;
use Exception;

/**
 * The main controller for a set of controls, can be instantiated with either page level or inline operations,
 * it doesn't care which.
 */
class ViewControls implements Iterator, ArrayAccess {

	/** @var string[] Set this to a class name (or series of classes), to provide on the top-level object */
	public $classNames = [];

	/** @var string Proxy label */
	public $proxyText = 't:STRING_CONTROLS';

	/** @var string Proxy icon */
	public $proxyIcon = 'ellipsis-v';

	/** @var string Rendering method, overridable by fetch parameter */
	public $renderAs = 'tray';

	/** @var string Hover-over text of the proxy */
	public $proxyTitle = 't:STRING_CONTROLS';

	public $buttonSize = 'default';

	/** @var string|null HTML ID of the menu (completely optional) */
	public $id;

	/** @var bool Set to true to include active/current links */
	public $includeActive = false;

	/** @var ViewControl[] */
	private $_links = [];

	private $_pos = 0;

	private $_data = [];

	/**
	 * Contains custom group icons, should any be set.
	 *
	 * @var array
	 */
	private $_groupIcons = [];

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Return the current element
	 * @link http://php.net/manual/en/iterator.current.php
	 * @return mixed Can return any type.
	 */
	public function current() {
		return $this->_links[$this->_pos];
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Move forward to next element
	 * @link http://php.net/manual/en/iterator.next.php
	 * @return void Any returned value is ignored.
	 */
	public function next() {
		++$this->_pos;
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Return the key of the current element
	 * @link http://php.net/manual/en/iterator.key.php
	 * @return mixed scalar on success, or null on failure.
	 */
	public function key() {
		return $this->_pos;
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Checks if current position is valid
	 * @link http://php.net/manual/en/iterator.valid.php
	 * @return boolean The return value will be casted to boolean and then evaluated.
	 * Returns true on success or false on failure.
	 */
	public function valid() {
		return isset($this->_links[$this->_pos]);
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Rewind the Iterator to the first element
	 * @link http://php.net/manual/en/iterator.rewind.php
	 * @return void Any returned value is ignored.
	 */
	public function rewind() {
		$this->_pos = 0;
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Whether a offset exists
	 * @link http://php.net/manual/en/arrayaccess.offsetexists.php
	 *
	 * @param mixed $offset <p>
	 * An offset to check for.
	 * </p>
	 *
	 * @return boolean true on success or false on failure.
	 * </p>
	 * <p>
	 * The return value will be casted to boolean if non-boolean was returned.
	 */
	public function offsetExists($offset) {
		return array_key_exists($offset, $this->_links);
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Offset to retrieve
	 * @link http://php.net/manual/en/arrayaccess.offsetget.php
	 * @param mixed $offset <p>
	 * The offset to retrieve.
	 * </p>
	 * @return mixed Can return all value types.
	 */
	public function offsetGet($offset) {
		return $this->_links[$offset];
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Offset to set
	 * @link http://php.net/manual/en/arrayaccess.offsetset.php
	 * @param mixed $offset <p>
	 * The offset to assign the value to.
	 * </p>
	 * @param mixed $value <p>
	 * The value to set.
	 * </p>
	 * @return void
	 *
	 * @throws Exception
	 */
	public function offsetSet($offset, $value) {
		if($offset === null){
			// The user just wants the next available one.

			if($this->valid()){
				$this->next();

			}
			$offset = $this->key();
		}

		if($value instanceof ViewControl){
			$this->_links[$offset] = $value;
		}
		elseif(is_array($value)){
			$control = new ViewControl();

			// Completely associative-array based version!
			foreach($value as $k => $v){
				$control->set($k, $v);
			}

			// Some legacy updates for the icon.
			if(!$control->icon){
				switch($control->class){
					case 'add':
					case 'edit':
					case 'directory':
						$control->icon = $control->class;
						break;
					case 'delete':
						$control->icon = 'remove';
						break;
					case 'view':
						$control->icon = 'eye-open';
						break;
				}
			}

			$this->_links[] = $control;
		}
		else{
			throw new Exception('Invalid offset type for ViewControls::offsetSet, please only set a ViewControl or an associative array');
		}
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Offset to unset
	 * @link http://php.net/manual/en/arrayaccess.offsetunset.php
	 * @param mixed $offset <p>
	 * The offset to unset.
	 * </p>
	 * @return void
	 */
	public function offsetUnset($offset) {
		unset($this->_links[$offset]);
	}
	
	/**
	 * Add an array of links to this control set.
	 *
	 * @param ViewControl[]|array[] $links
	 */
	public function addLinks(array $links){
		foreach($links as $l){
			if($l instanceof ViewControl) {
				$this->addLink($l);
			}
			elseif(is_array($l)) {
				$this->addLinkFromArray($l);
			}
		}
	}

	/**
	 * Add a single link to this control set.
	 *
	 * Useful for the times when you cannot access the ViewControls as an array.
	 *
	 * @param ViewControl $link
	 * 
	 * @return ViewControl
	 */
	public function addLink(ViewControl $link = null) : ViewControl {
		if($link === null) {
			$link = new ViewControl();
		}
		
		$this[] = $link;
		
		// Allow chaining
		return $link;
	}
	
	/**
	 * Add a single link based on an associative array
	 * 
	 * @param array $linkData
	 * 
	 * @return ViewControl
	 */
	public function addLinkFromArray(array $linkData) : ViewControl {
		$link = new ViewControl();
		
		foreach($linkData as $key => $val) {
			// Allow setting the group icon from an individual viewcontrol.
			// This allows all the required data to be contained within individual controls.
			if($key === 'groupicon' && isset($linkData['group'])) {
				$this->_groupIcons[ $linkData['group'] ] = $val;
			}
			else {
				$link->set($key, $val);
			}
		}
		
		$this[] = $link;
		
		// Allow chaining
		return $link;
	}

	/**
	 * Get this control set as HTML
	 *
	 * @return string HTML
	 */
	public function fetch(string $renderMethod = null, array $options = []) : string {
		if(!$this->hasLinks()){
			// If there are no links, simply skip all rendering.
			return '';
		}

		// If there are options provided, set those for this rendering,
		// but reset back to default once done.
		$resets = [];
		if(count($options)) {
			foreach($options as $optionKey => $optionValue) {
				switch($optionKey) {
					case 'proxy_text':
						$resets['proxy_text'] = $this->proxyText;
						$this->proxyText = $optionValue;
						break;
					case 'proxy_icon':
						$resets['proxy_icon'] = $this->proxyIcon;
						$this->proxyIcon = $optionValue;
						break;
					case 'title':
					case 'proxy_title':
						$resets['proxy_title'] = $this->proxyTitle;
						$this->proxyTitle = $optionValue;
						break;
					case 'button_size':
						$resets['button_size'] = $this->buttonSize;
						$this->buttonSize = $optionValue;
						break;
					case 'class':
						$resets['class'] = $this->classNames;
						$this->classNames = $optionValue;
						break;
					case 'id':
						$resets['id'] = $this->id;
						$this->id = $optionValue;
						break;
					case 'include_active':
						$resets['include_active'] = $this->includeActive;
						$this->includeActive = $optionValue;
						break;
				}
			}
		}

		$renderAs = $renderMethod === null ? $this->renderAs : $renderMethod;

		if(!is_array($this->classNames)) {
			$this->classNames = [ $this->classNames ];
		}

		switch($renderAs) {
			case 'tray':
				$html = $this->_renderTray();
				break;
			case 'subtrays':
				$html = $this->_renderSubTrays();
				break;
			case 'list':
				$html = $this->_renderList();
				break;
			case 'buttons':
				$html = $this->_renderButtons();
				break;
			case 'buttongroup':
				$html = $this->_renderButtonGroup();
				break;
			case 'buttongroups':
				$html = $this->_renderButtonGroups();
				break;
			default:
				$html = 'Unsupported render method for controls: ' . $renderAs;
		}


		if(count($resets)) {
			foreach($resets as $resetKey => $resetValue) {
				switch($resetKey) {
					case 'proxy_text':
						$this->proxyText = $resetValue;
						break;
					case 'proxy_icon':
						$this->proxyIcon = $resetValue;
						break;
					case 'title':
					case 'proxy_title':
						$this->proxyTitle = $resetValue;
						break;
					case 'button_size':
						$this->buttonSize = $resetValue;
						break;
					case 'class':
						$this->classNames = $resetValue;
						break;
					case 'id':
						$this->id = $resetValue;
						break;
					case 'include_active':
						$this->includeActive = $resetValue;
						break;
				}
			}
		}

		return $html;
	}

	/**
	 * Check if this control set has any links in it.
	 * Useful for templates where you don't want to render the container if there is nothing to render inside.
	 *
	 * @return bool
	 */
	public function hasLinks(){
		return (sizeof($this->_links) > 0);
	}

	/**
	 * Get all links sorted and grouped by their grouping.
	 *
	 * These will exclude the current page's link, if set.
	 *
	 * @return array[ViewControl[]]
	 */
	public function getGroupedLinks() : array {
		$links = $this->getSortedLinks();
		$groups = [];

		foreach($links as $link) {
			if(!isset($groups[ $link->group ])) {
				$groups[ $link->group ] = [];
			}

			$groups[ $link->group ][] = $link;
		}

		return $groups;
	}

	/**
	 * Get all the links
	 *
	 * Will return a flat array of links.
	 *
	 * @return ViewControl[]
	 */
	public function getSortedLinks() : array {
		$currentURL = \Core\resolve_link(Router::$DefaultBaseURL->getBaseURL());

		$links = [];
		foreach($this->_links as $link) {
			if($this->includeActive || $link->link !== $currentURL) {
				// Include the link if it's not a current page OR if includeActive is enabled
				$links[] = $link;
			}
		}
		usort($links, function($a, $b) { return $a->weight - $b->weight; });

		return $links;
	}

	public function setProxyText(string $text) {
		$this->proxyText = $text;
	}

	public function setProxyForce($force) {
		$this->_data['proxy-force'] = $force ? '1' : '0';
	}

	public function setGroupIcon(string $group, string $icon) {
		$this->_groupIcons[ $group ] = $icon;
	}

	private function _renderProxyLabel() : string {
		$proxyText = $this->proxyText;
		$proxyTitle = $this->proxyTitle;

		if(strpos($proxyText, 't:') === 0) {
			// Automatic I18N for proxy text when rendering uses UL.
			$proxyText = t(substr($proxyText, 2));
		}

		if(strpos($proxyTitle, 't:') === 0) {
			// Automatic I18N for proxy text when rendering uses UL.
			$proxyTitle = t(substr($proxyTitle, 2));
		}

		$proxyLabelParameters = [
			'class' => ['tray-toggle'],
			'title' => $proxyTitle,
		];

		// Build the proxy label
		$html = '<label ' . \Core\array_to_html_attributes_string($proxyLabelParameters) . '>';
		if($this->proxyIcon) {
			$html .= '<i class="icon icon-' . $this->proxyIcon . '"></i>';
		}
		if($proxyText) {
			$html .= '<span>' . $proxyText . '</span>';
		}
		$html .= '</label>';

		return $html;
	}

	private function _renderTray() : string {
		// Start building the actual links
		$html = '';
		$linkCount = 0;
		$groups = $this->getGroupedLinks();

		if(!count($groups)) {
			// Safety catch if there was only 1 link listed and that is the current page.
			return '';
		}

		// Render any parent-level links first
		if(isset($groups[''])) {
			foreach($groups[''] as $link) {
				/** @var $link ViewControl */
				$html .= $link->fetch();
				$linkCount++;
			}
			unset($groups['']);
		}

		// There is more than one group!
		foreach($groups as $groupName => $sublinks) {
			// Links with a group get rendered in that group
			$groupIcon = $this->_getGroupIconHTML($groupName);

			if(strpos($groupName, 't:') === 0) {
				$groupName = t(substr($groupName, 2));
			}

			$html .= '<li>' .
				'<label>' .
				$groupIcon .
				'<span>' . $groupName . '</span>' .
				'</label>' .
				'<ul class="control-subgroup">';
			foreach($sublinks as $link) {
				/** @var $link ViewControl */
				$html .= $link->fetch();
				$linkCount++;
			}
			$html .= '</ul>' .
				'</li>';
		}

		// At this stage, $html contains the inner group/link markup of these controls,
		// so now we'll need to wrap that with the containers.
		// This is done to detect some information about the contents for the container.
		$menuParameters = [
			'class' => array_merge($this->classNames, ['controls', 'tray']),
			'role' => 'menu'
		];

		if(count($groups)) {
			$menuParameters['class'][] = 'has-groups';
		}

		if($linkCount >= 14) {
			$menuParameters['class'][] = 'has-many-many-links';
		}
		elseif($linkCount >= 6) {
			$menuParameters['class'][] = 'has-many-links';
		}
		elseif($linkCount >= 1) {
			$menuParameters['class'][] = 'has-links';
		}

		// Put everything together
		return '<div ' . \Core\array_to_html_attributes_string($menuParameters) . '>' .
			$this->_renderProxyLabel() .
			'<ul>' .
			$html .
			'</ul>' .
			'</div>';
	}

	private function _renderSubTrays() : string {
		$menuParameters = [
			'class' => array_merge($this->classNames, ['controls', 'subtrays']),
			'role' => 'menu'
		];

		if($this->id) {
			$menuParameters['id'] = $this->id;
		}

		// Start the menu container
		$html = '<div ' . \Core\array_to_html_attributes_string($menuParameters) . '>';

		// Build the proxy label
		$html .= $this->_renderProxyLabel();

		// Start the primary containing UL
		$html .= '<ul>';

		// Start building the actual links
		$groups = $this->getGroupedLinks();

		if(!count($groups)) {
			// Safety catch if there was only 1 link listed and that is the current page.
			return '';
		}

		// Render any parent-level links first
		if(isset($groups[''])) {
			foreach($groups[''] as $link) {
				/** @var $link ViewControl */
				$html .= $link->fetch();
			}
			unset($groups['']);
		}

		// There is more than one group!
		foreach($groups as $groupName => $sublinks) {
			// Links with a group get rendered in that group
			$groupIcon = $this->_getGroupIconHTML($groupName);

			if(strpos($groupName, 't:') === 0) {
				$groupName = t(substr($groupName, 2));
			}

			$html .= '<li class="tray">' .
				'<label class="tray-toggle">' .
				$groupIcon .
				'<span>' . $groupName . '</span>' .
				'</label>' .
				'<ul class="control-subgroup">';
			foreach($sublinks as $link) {
				/** @var $link ViewControl */
				$html .= $link->fetch();
			}
			$html .= '</ul>' .
				'</li>';
		}

		$html .= '</ul>';

		$html .= '</div>';

		return $html;
	}

	private function _renderList() : string {
		$menuParameters = [
			'class' => array_merge($this->classNames, ['controls']),
			'role' => 'menu'
		];

		if($this->id) {
			$menuParameters['id'] = $this->id;
		}

		// Start the menu container
		$html = '<div ' . \Core\array_to_html_attributes_string($menuParameters) . '>';

		// Start the primary containing UL
		$html .= '<ul>';

		// Start building the actual links
		$links = $this->getSortedLinks();

		if(!count($links)) {
			// No links, nothing to render
			return '';
		}

		foreach($links as $link) {
			$html .= $link->fetch();
		}

		$html .= '</ul>';

		$html .= '</div>';

		return $html;
	}

	/**
	 * Render this control list as a flat list of buttons
	 *
	 * @return string
	 */
	private function _renderButtons() : string {
		$menuParameters = [
			'class' => array_merge($this->classNames, ['controls']),
			'role' => 'menu'
		];

		if($this->id) {
			$menuParameters['id'] = $this->id;
		}

		// Start the menu container
		$html = '<div ' . \Core\array_to_html_attributes_string($menuParameters) . '>';

		// Start building the actual links
		$links = $this->getSortedLinks();

		if(!count($links)) {
			// No links, nothing to render
			return '';
		}

		foreach($links as $link) {
			// Store so I can reset this link to the default state
			$class = $link->class;
			$link->addClass('button');

			$html .= $link->fetchSingle();

			// And reset classes
			$link->classname($class);
		}

		$html .= '</div>';

		return $html;
	}

	/**
	 * Render this ViewControls as a single button group
	 *
	 * @return string
	 */
	private function _renderButtonGroup() : string {
		$menuParameters = [
			'class' => array_merge($this->classNames, ['controls', 'button-group']),
			'role' => 'menu'
		];

		if($this->id) {
			$menuParameters['id'] = $this->id;
		}

		// Start the menu container
		$html = '<div ' . \Core\array_to_html_attributes_string($menuParameters) . '>';

		// Start building the actual links
		$links = $this->getSortedLinks();
		$buttonSize = $this->_getButtonSize();

		if(!count($links)) {
			// No links, nothing to render
			return '';
		}

		foreach($links as $link) {
			// Store so I can reset this link to the default state
			$class = $link->class;
			$link->addClass('button');

			if($buttonSize) {
				$link->addClass($buttonSize);
			}

			$html .= $link->fetchSingle();

			// And reset classes
			$link->classname($class);
		}

		$html .= '</div>';

		return $html;
	}

	/**
	 * Render this ViewControls as a group of discrete button groups, grouped by their applicable grouping
	 *
	 * @return string
	 */
	private function _renderButtonGroups() : string {
		$menuParameters = [
			'class' => array_merge($this->classNames, ['controls', 'button-groups']),
			'role' => 'menu'
		];

		if($this->id) {
			$menuParameters['id'] = $this->id;
		}

		// Start the menu container
		$html = '<div ' . \Core\array_to_html_attributes_string($menuParameters) . '>';

		// Start building the actual links
		$groups = $this->getGroupedLinks();
		$buttonSize = $this->_getButtonSize();

		if(!count($groups)) {
			// Safety catch if there was only 1 link listed and that is the current page.
			return '';
		}

		// Render any parent-level links first
		if(isset($groups[''])) {
			$html .= '<div class="button-group">';

			foreach($groups[''] as $link) {
				/** @var $link ViewControl */
				// Store so I can reset this link to the default state
				$class = $link->class;
				$link->addClass('button');

				if($buttonSize) {
					$link->addClass($buttonSize);
				}

				$html .= $link->fetchSingle();

				// And reset classes
				$link->classname($class);
			}

			$html .= '</div>';
			unset($groups['']);
		}

		// There is more than one group!
		foreach($groups as $groupName => $sublinks) {

			$groupClass = 'group-' . \Core\str_to_url($groupName);

			$html .= '<div class="button-group ' . $groupClass . '">';

			foreach($sublinks as $link) {
				/** @var $link ViewControl */
				// Store so I can reset this link to the default state
				$class = $link->class;
				$link->addClass('button');

				$html .= $link->fetchSingle();

				// And reset classes
				$link->classname($class);
			}

			$html .= '</div>';
		}

		$html .= '</div>';

		return $html;
	}

	private function _getGroupIconHTML(string $group) : string {
		if(isset($this->_groupIcons[ $group ])) {
			$icon = $this->_groupIcons[ $group ];
		}
		else {
			switch($group){
				case ViewControl::GROUP_ADMIN:
					$icon = 'cogs';
					break;
				case ViewControl::GROUP_MANAGE:
					$icon = 'wrench';
					break;
				case ViewControl::GROUP_VIEW:
					$icon = 'eye';
					break;
				case ViewControl::GROUP_CREATE:
					$icon = 'add';
					break;
				default:
					$icon = '';
			}
		}


		if($icon) {
			return '<i class="icon icon-' . $icon . '"></i> ';
		}
		else {
			return '<i class="icon">&nbsp;</i>';
		}
	}

	/**
	 * Get the CSS class for the button size
	 *
	 * If "default" or an otherwise invalid size, an empty string is returned.
	 *
	 * Otherwise, button-size-small, button-size-large, button-size-xlarge
	 * @return string
	 */
	private function _getButtonSize() : string {
		switch($this->buttonSize) {
			case 'large':
			case 'l':
				return 'button-size-large';
			case 'xl':
			case 'xlarge':
				return 'button-size-xlarge';
			case 'small':
			case 'sm':
			case 's':
				return 'button-size-small';
			default:
				return '';
		}
	}



	/**
	 * Shortcut function to dispatch the /core/controllinks hook to request functions for a given subject.
	 *
	 * @param string $baseurl The baseurl, (excluding /core/controllinks), of the request
	 * @param mixed  $subject The subject matter of this hook, (if any)
	 *
	 * @return ViewControls
	 */
	public static function Dispatch($baseurl, $subject){
		$links = HookHandler::DispatchHook('/core/controllinks' . $baseurl, $subject);

		$controls = new ViewControls();
		$controls->addLinks($links);

		return $controls;
	}

	/**
	 * Shortcut function to dispatch the /core/controllinks hook to request functions for a given subject.
	 *
	 * @param Model $model The subject matter of this hook, (if any)
	 *
	 * @return ViewControls
	 */
	public static function DispatchModel(Model $model){
		$links = $model->getControlLinks();
		$controls = new ViewControls();
		$controls->addLinks($links);

		return $controls;
	}

	/**
	 * Shortcut function to dispatch the /core/controllinks hook to request functions for a given subject.
	 *
	 * @param string $baseurl The baseurl, (excluding /core/controllinks), of the request
	 * @param mixed  $subject The subject matter of this hook, (if any)
	 *
	 * @return string HTML of the <ul/> tag.
	 */
	public static function DispatchAndFetch($baseurl, $subject){
		$links = HookHandler::DispatchHook('/core/controllinks' . $baseurl, $subject);

		$controls = new ViewControls();
		$controls->addLinks($links);

		return $controls->fetch();
	}
}
