<?php
/**
 * File: Memcache
 *
 * @package Core\Cache
 *
 */

namespace Core\Cache;

/**
 * Class: Memcache
 */
class Memcache implements CacheInterface {
	
	private $_gzip;
	private $_conn;
	private $_isMemcached;
	
	/**
	 * Construct a new cache backend, capable of storing and retrieving values
	 */
	public function __construct() {
		// Enable compression, if available
		$this->_gzip = (extension_loaded('zlib'));

		// Check and see if the connection has been initialized.
		// Only do this once to save time.
		// Prefer Memcached over Memcache.
		if (class_exists('Memcached')) {
			$this->_conn        = new \Memcached();
			$this->_isMemcached = true;
			
			if($this->_gzip){
				$this->_conn->setOption(\Memcached::OPT_COMPRESSION, true);
			}
		}
		elseif (class_exists('Memcache')) {
			$this->_conn        = new \Memcache();
			$this->_isMemcached = false;
		}
		
		// Process Memcached servers.
		// @todo Make this configurable!
		$this->_conn->addServer('localhost', 11211);
	}

	/**
	 * Writes the requested key and data optionally with an expiration date.
	 * 
	 * @param string $key  Key to save
	 * @param mixed $data  The data to cache
	 * @param int $expires Expiration in number of seconds from now
	 * 
	 * @return boolean Whether the operation was successful.
	 */
	public function write(string $key, $data, int $expires = null) : bool {
		return $this->_isMemcached ?
			$this->_conn->set($key, $data, $expires) :
			$this->_conn->set($key, $data, $this->_gzip, $expires);
	}

	/**
	 * Read the given key from the cache backend and return the contents
	 * 
	 * @param string $key Key to read
	 * 
	 * @return mixed Either the content of the cache object, or _boolean_ false.
	 */
	public function read(string $key) {
		$val = $this->_isMemcached ?
			$this->_conn->get($key) :
			$this->_conn->get($key, $this->_gzip);
		
		// memcache likes to return false if the value is not present.
		return $val === false ? null : $val;
	}

	/**
	 * Delete the cache entry matching the key
	 * 
	 * @param string $key Key to delete
	 * 
	 * @return boolean Whether the operation was successful.
	 */
	public function delete(string $key) : bool {
		return $this->_conn->delete($key);
	}
	
	/**
	 * Invalidate all items in the cache
	 * 
	 * @since 2011.07.28
	 * @return boolean Whether the operation was successful.
	 */
	public function flush() : bool {
		return $this->_conn->flush();
	}
	
	/**
	 * Return a list of all keys on this backend, not supported by all processors!
	 * 
	 * @since 2017.04
	 * @return array
	 */
	public function ls() : array {
		return $this->_isMemcached ?
			$this->_conn->getAllKeys() :
			[]; // memcache does not support ls!
	}
	
	/**
	 * Check if this interface is ready to store/receive values.
	 * 
	 * @return bool
	 */
	public function isReady() : bool {
		return $this->_conn !== null;
	}
}
