<?php
/**
 * Provides the main interface system for the DMI subsystem.
 * 
 * @package Core\Datamodel
 * @since 0.1
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Datamodel;

use Core\ConfigHandler;


/**
 * A top level interface class for the Data Model Interface.
 * Provides abstraction for different backends.
 *
 */
class DMI {
	
	/** @var BackendInterface The backend currently in use for this DMI object. */
	protected $_backend = null;
	
	/**
	 * This points to the system/global DMI object.
	 * 
	 * @var DMI
	 */
	static protected $_Interface = null;
	
	public function __construct($backend = null, $host = null, $user = null, $pass = null, $database = null){
		// Provide shortcut to set the backend directly in the constructor.
		if($backend) $this->setBackend($backend);
		
		// Provide shortcut to set connection information directly in the constructor.
		if($host) $this->connect($host, $user, $pass, $database);
	}
	
	public function setBackend($backend){
		if($this->_backend){
			throw new DMIException('Backend already set');
		}

		// All backends are lowercase.
		$backend     = strtolower($backend);
		$class       = 'Core\\Datamodel\\Drivers\\' . $backend . '\\' . $backend . '_backend';
		$this->_backend = new $class();
	}

	/**
	 * @param $host
	 * @param $user
	 * @param $pass
	 * @param $database
	 *
	 * @throws DMIException
	 * @throws DMIAuthenticationFailedException
	 *
	 * @return \Core\Datamodel\BackendInterface|null
	 */
	public function connect($host, $user, $pass, $database){
		$this->_backend->connect($host, $user, $pass, $database);
		
		return $this->_backend;
	}

	/**
	 * @return \Core\Datamodel\BackendInterface
	 */
	public function connection(){
		return $this->_backend;
	}
	
	
	/**
	 * Get the current system DMI based on configuration values.
	 *
	 * @throws DMIException
	 * @throws DMIAuthenticationFailedException
	 *
	 * @return DMI
	 */
	public static function GetSystemDMI() : DMI {
		if(self::$_Interface === null){
			// Try to establish a new connection!
			$settings = ConfigHandler::LoadConfigFile('configuration');
			
			$connection = new DMI();
			
			if(isset($settings['database_type'])){
				$connection->setBackend($settings['database_type']);
			}
			else{
				throw new DMIException('No database type defined.');
			}
			
			$connection->connect(
				$settings['database_server'],
				$settings['database_user'],
				$settings['database_pass'],
				$settings['database_name']
			);
			
			// If everything was successful thus far, we can store this connection!
			self::$_Interface = $connection;
		}
		
		return self::$_Interface;
	}
}

