<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core\Datamodel\Drivers\sqlite3;

use Core\Datamodel\DatasetFrom;
use SQLite3;
use Core\Datamodel\BackendInterface;
use Core\Datamodel\Dataset;
use Core\Datamodel\DatasetWhere;
use Core\Datamodel\DatasetWhereClause;
use Core\Datamodel\DMIException;
use Core\Datamodel\DMIQueryException;
use Core\Utilities\Profiler\DatamodelProfiler;

/**
 * Description of sqlite3
 *
 * @author charlie
 */
class sqlite3_backend implements BackendInterface {
	
	private $_filename;
	private $_isReadonly = false;

	/** @var SQLite3 */
	private $_conn;
	
	private $_autoKey = 0;
	
	private static $_ReservedKeywords = [
		'ABORT',
		'ACTION',
		'ADD',
		'AFTER',
		'ALL',
		'ALTER',
		'ANALYZE',
		'AND',
		'AS',
		'ASC',
		'ATTACH',
		'AUTOINCREMENT',
		'BEFORE',
		'BEGIN',
		'BETWEEN',
		'BY',
		'CASCADE',
		'CASE',
		'CAST',
		'CHECK',
		'COLLATE',
		'COLUMN',
		'COMMIT',
		'CONFLICT',
		'CONSTRAINT',
		'CREATE',
		'CROSS',
		'CURRENT_DATE',
		'CURRENT_TIME',
		'CURRENT_TIMESTAMP',
		'DATABASE',
		'DEFAULT',
		'DEFERRABLE',
		'DEFERRED',
		'DELETE',
		'DESC',
		'DETACH',
		'DISTINCT',
		'DROP',
		'EACH',
		'ELSE',
		'END',
		'ESCAPE',
		'EXCEPT',
		'EXCLUSIVE',
		'EXISTS',
		'EXPLAIN',
		'FAIL',
		'FOR',
		'FOREIGN',
		'FROM',
		'FULL',
		'GLOB',
		'GROUP',
		'HAVING',
		'IF',
		'IGNORE',
		'IMMEDIATE',
		'IN',
		'INDEX',
		'INDEXED',
		'INITIALLY',
		'INNER',
		'INSERT',
		'INSTEAD',
		'INTERSECT',
		'INTO',
		'IS',
		'ISNULL',
		'JOIN',
		'KEY',
		'LEFT',
		'LIKE',
		'LIMIT',
		'MATCH',
		'NATURAL',
		'NO',
		'NOT',
		'NOTNULL',
		'NULL',
		'OF',
		'OFFSET',
		'ON',
		'OR',
		'ORDER',
		'OUTER',
		'PLAN',
		'PRAGMA',
		'PRIMARY',
		'QUERY',
		'RAISE',
		'RECURSIVE',
		'REFERENCES',
		'REGEXP',
		'REINDEX',
		'RELEASE',
		'RENAME',
		'REPLACE',
		'RESTRICT',
		'RIGHT',
		'ROLLBACK',
		'ROW',
		'SAVEPOINT',
		'SELECT',
		'SET',
		'TABLE',
		'TEMP',
		'TEMPORARY',
		'THEN',
		'TO',
		'TRANSACTION',
		'TRIGGER',
		'UNION',
		'UNIQUE',
		'UPDATE',
		'USING',
		'VACUUM',
		'VALUES',
		'VIEW',
		'VIRTUAL',
		'WHEN',
		'WHERE',
		'WITH',
		'WITHOUT',
	];
	
	public function setFilename(string $filename){
		if($this->_conn !== null){
			throw new DMIException('Unable to set filename on an already-opened database!');
		}
		
		$this->_filename = $filename;
	}
	
	public function setReadonly(){
		if($this->_conn !== null){
			throw new DMIException('Unable to set readonly on an already-opened database!');
		}
		
		$this->_isReadonly = true;
	}
	
	public function connect(){
		$flag = $this->_isReadonly ? SQLITE3_OPEN_READONLY : SQLITE3_OPEN_READWRITE;
		
		$this->_conn = new SQLite3($this->_filename, $flag);
	}
	
	public function createTable($table, \Core\Datamodel\Schema $schema): bool {
		
	}

	public function describeTable($table): \Core\Datamodel\Schema {
		
	}

	public function dropTable($table): bool {
		
	}
	
	public function executePrepared(string $type, string $query, array $params){
		$stmt = $this->_conn->prepare($query);
		
		foreach($params as $k => $v){
			$stmt->bindValue($k, $v);
		}
		
		$profiler = DatamodelProfiler::GetDefaultProfiler();
		$profiler->start($type, $query);

		$res = $stmt->execute();

		if(DEVELOPMENT_MODE && is_object($res) && property_exists($res, 'num_rows')){
			// Tack on how many rows were selected or affected for debugging purposes.
			$rows = $res->num_rows;
		}
		else{
			$rows = null;
		}

		if($this->_conn->errno){
			$profiler->stopError($this->_conn->errno, $this->_conn->error);

			$e = new DMIQueryException($this->_conn->error, $this->_conn->errno);
			$e->query = $query;
			throw $e;
		}
		else{
			$profiler->stopSuccess($rows);

			return $res;
		}
	}

	public function execute(Dataset $dataset) {
		$mode = $dataset->getMode();

		switch($mode){
			case Dataset::MODE_GET:
				$this->_executeGet($dataset);
				break;
			default:
				throw new DMIException('Invalid dataset mode [' . $mode . ']');
				break;
		}
	}

	public function modifyTable($table, \Core\Datamodel\Schema $schema) {
		
	}

	public function showTables(): array {
		
	}

	public function tableExists($tablename): bool {
		
	}
	
	/**
	 * Escape a keyword such as a column name, table name, or alias.
	 * 
	 * If the string matches a given keyword, it's wrapped with backticks.
	 * 
	 * @param string $keyword
	 * @return string
	 */
	private function _escapeKeyword(string $keyword) : string {
		if(in_array(strtoupper($keyword), self::$_ReservedKeywords)){
			return '`' . $keyword . '`';
		}
		else{
			return $keyword;
		}
	}

	private function _executeGet(Dataset $dataset){
		// Generate a query to run.
		$bindings = [];
		$q = 'SELECT';
		
		if($dataset->uniquerecords) {
			$q .= ' DISTINCT';
		}
		
		$ss = [];
		foreach($dataset->_selects as $select){
			// Check the escaping for this column.
			$s = '';
			
			if($select->table){
				$s .= $this->_escapeKeyword($select->table) . '.';
			}
			
			$s .= $this->_escapeKeyword($select->column);
			
			if($select->alias){
				$s .= ' ' . $this->_escapeKeyword($select->alias);
			}
			
			$ss[] = $s;
		}
		
		// Nothing to select?  The user probably wanted '*'.
		if(!sizeof($ss)){
			$ss[] = '*';
		}
		
		$q .= ' ' . implode(', ', $ss);

		if(count($dataset->_tables)) {
			$q .= ' FROM ' . $this->_parseFrom($dataset->_tables);
		}
		
		if($dataset->_joins !== null){
			foreach($dataset->_joins as $join){
				$q .= ' ' . $join->type . ' JOIN ' . $this->_escapeKeyword($join->table);
				if($join->alias){
					$q .= ' ' . $this->_escapeKeyword($join->alias);
				}

				if($join->on) {
					$whereData = $this->_parseWhere($join->on);
					$bindings = array_merge($bindings, $whereData[1]);
					$q .= ' ON ' . $whereData[0];
				}
			}
		}
		
		if($dataset->_where !== null){
			$whereData = $this->_parseWhere($dataset->_where);
			
			if(trim($whereData[0])){
				$bindings = array_merge($bindings, $whereData[1]);
				$q .= ' WHERE ' . $whereData[0];
			}
		}
		

		if($dataset->_order){
			// Support keys as complex as "key DESC, value ASC"
			// and as simple as "sortnum"
			$orders = explode(',', $dataset->_order);
			$os = [];
			foreach($orders as $o){
				$o = trim($o);

				// Allow for mycolumn DESC or order ASC
				if(strpos($o, ' ') !== false) $os[] = '`' . substr($o, 0, strpos($o, ' ')) . '` ' . substr($o, strpos($o, ' ') + 1);
				// Allow for RAND() or other functions.
				elseif(strpos($o, '()') !== false) $os[] = $o;
				// Everything else just gets escaped normally.
				else $os[] = '`' . $o . '`';
			}
			$q .= ' ORDER BY ' . implode(', ', $os);
		}

		if($dataset->_limit){
			$q .= ' LIMIT ' . $dataset->_limit;
		}
		
		
		// Lastly, execute this query while supporting the profiler.
		$this->_exec('read', $q, $dataset, $bindings);
	}

	/**
	 * @param DatasetFrom[] $froms
	 *
	 * @return string
	 *
	 * @throws DMIQueryException
	 */
	private function _parseFrom(array $froms) : string {
		$fs = [];

		foreach($froms as $from) {
			if($from->table) {
				$f = $this->_escapeKeyword($from->table);
			}
			else {
				throw new DMIQueryException('Invalid or incomplete FROM statement');
			}

			if($from->alias) {
				$f .= ' AS ' . $this->_escapeKeyword($from->alias);
			}

			$fs[] = $f;
		}

		return implode(', ', $fs);
	}
	
	private function _exec(string $type, string $query, Dataset $dataset, array $bindings = null){
		
		$profiler = DatamodelProfiler::GetDefaultProfiler();
		$profiler->start($type, $query);

		if(is_array($bindings) && count($bindings)){
			// Use a statement!
			$stmt = $this->_conn->prepare($query);
			foreach($bindings as $k => $v){
				$stmt->bindValue($k, $v);
			}
			$res = $stmt->execute();
		}
		else{
			// No prepared statement necessary; no values provided.
			$res = $this->_conn->query($query);
		}
		
		$dataset->_data = [];
		while(($row = $res->fetchArray(SQLITE3_ASSOC)) !== false){
			$dataset->_data[] = $row;
		}
		$dataset->num_rows = count($dataset->_data);
		
		$profiler->stopSuccess($dataset->num_rows);
	}
	
	/**
	 * The recursive function that will return the actual SQL string from a group.
	 *
	 * @param DatasetWhereClause $group
	 * @return array
	 */
	private function _parseWhere(DatasetWhereClause $group) : array {
		$statements = $group->getStatements();
		$parts = [];
		$bindings = [];
		
		foreach($statements as $w){
			if($w instanceof DatasetWhereClause){
				// Recursively recurring recursion, RECURSE!
				$sub = $this->_parseWhere($w);
				if($sub[0]){
					$parts[] = '( ' . $sub[0] . ' )';
					$bindings += $sub[1];
				}
			}
			elseif($w instanceof DatasetWhere){
				// No field, what can I do?
				if(!$w->field){
					continue;
				}

				$op = $w->op;

				if($op === 'NULL') {
					$op = 'IS';
					$key = 'NULL';
				}
				elseif($op === 'NOT NULL') {
					$op = 'IS NOT';
					$key = 'NULL';
				}
				elseif($w->value !== null) {
					$key = ':key' . $this->_autoKey++;
					$bindings[ $key ] = $w->value;
				}
				elseif($w->valueKey !== null) {
					if($w->valueTable) {
						$key = $this->_escapeKeyword($w->valueTable) . '.' . $this->_escapeKeyword($w->valueKey);
					}
					else {
						$key = $this->_escapeKeyword($w->valueKey);
					}
				}

				
				$q = '';
				if($w->table){
					$q .= $this->_escapeKeyword($w->table) . '.';
				}
				$q .= $this->_escapeKeyword($w->field) . ' ' . $op . ' ' . $key;
				
				$parts[] = $q;
			}
		}

		return [
			implode(' ' . $group->getSeparator() . ' ', $parts),
			$bindings
		];
	}

	/**
	 * Parse a Dataset and convert it to a native format for this backend
	 *
	 * @param Dataset $dataset
	 *
	 * @return string
	 *
	 * @throws DMIQueryException
	 */
	public function parseDataset(Dataset $dataset): string {
		// TODO: Implement parseDataset() method.
	}
}
