<?php
/**
 * File for the common file functions useful throughout Core.
 * 
 * @package Core\Filestore
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20130530.1735
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Filestore;

use Exception;
use Core\Filestore\CDN;
use Core\Filestore\FTP\FTPConnection;
use Core\I18N\I18NLoader;
use Core\ConfigHandler;

/**
 * Utility function to translate a filesize in bytes into a human-readable version.
 *
 * @param int $filesize Filesize in bytes
 * @param int $round Precision to round to
 *
 * @return string
 */
function format_size($filesize, $round = 2) {
	\Core\log_deprecated('Filestore\\format_size', debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS, 2)[1]);
	return FileHelper::FormatSize($filesize, $round);
}


/**
 * Resolve the correct contents object type for the requested file.
 * This mainly operates off the mimetype.
 *
 * @param File $file
 *
 * @return Contents
 *
 * @throws Exception
 */
function resolve_contents_object(File $file){
	\Core\log_deprecated('Filestore\\resolve_contents_object', debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS, 2)[1]);
	return Factory::ContentsFromFile($file);
}


/**
 * Resolve a name for a public to an actual file.
 *
 * @param $filename
 *
 * @return \Core\Filestore\File
 *
 * @throws \Exception
 */
function resolve_public_file($filename){
	return _resolve_type_file($filename, get_public_path(), 'public/', 'file');
}

/**
 * Resolve a name for a private to an actual file.
 *
 * @param $filename
 *
 * @return \Core\Filestore\File
 *
 * @throws \Exception
 */
function resolve_private_file($filename){
	return _resolve_type_file($filename, get_private_path(), 'private/', 'file');
}

/**
 * Resolve a name for a public to an actual file.
 *
 * @param $filename
 *
 * @return \Core\Filestore\Directory
 *
 * @throws \Exception
 */
function resolve_public_directory($filename){
	return _resolve_type_file($filename, get_public_path(), 'public/', 'directory');
}

/**
 * Resolve a name for a private to an actual file.
 *
 * @param $filename
 *
 * @return \Core\Filestore\Directory
 *
 * @throws \Exception
 */
function resolve_private_directory($filename){
	return _resolve_type_file($filename, get_private_path(), 'private/', 'directory');
}

/**
 * Resolve a name for a public/private to an actual file.
 *
 * @param $filename
 *
 * @return \Core\Filestore\File
 *
 * @throws \Exception
 */
function _resolve_type_file($filename, $resolved, $prefix, $type){

	if (strpos($filename, $prefix) === 0) {
		// Allow "assets/blah" to be passed in
		$filename = substr($filename, strlen($prefix));
	}
	elseif(strpos($filename, $resolved) === 0){
		// Allow the fully resolved name to be passed in
		$filename = substr($filename, strlen($resolved));
	}

	switch(CDN_TYPE){
		case 'local':
			if(\Core\ftp()){
				// FTP has its own sub-type.
				return 
					$type === 'file' ? 
					new Backends\FileFTP($resolved . $filename) : 
					new Backends\DirectoryFTP($resolved . $filename);
			}
			else{
				return
					$type === 'file' ? 
					new Backends\FileLocal($resolved . $filename) :
					new Backends\DirectoryLocal($resolved . $filename);
			}
			break;

		case 'ftp':
			return 
				$type === 'file' ? 
				new Backends\FileFTP($resolved  . $filename, cdn_ftp()) :
				new Backends\DirectoryFTP($resolved . $filename, cdn_ftp());
			break;

		default:
			throw new \Exception('Unsupported CDN type: ' . CDN_TYPE);
			break;
	}
}

/**
 * Resolve a name for an asset to an actual file.
 *
 * @param $filename
 *
 * @return \Core\Filestore\Directory
 *
 * @throws \Exception
 */
function resolve_asset_directory($filename){
	$resolved = get_asset_path();

	if (strpos($filename, 'assets/') === 0) {
		// Allow "assets/blah" to be passed in
		$filename = substr($filename, 7);
	}
	elseif(strpos($filename, 'asset/') === 0){
		// Allow "asset/blah" to be passed in.
		$filename = substr($filename, 6);
	}
	elseif(strpos($filename, $resolved) === 0){
		// Allow the fully resolved name to be passed in
		$filename = substr($filename, strlen($resolved));
	}
	
	
	// I need to check the custom, current theme, and finally default locations for the file.
	switch(CDN_TYPE){
		case 'local':
			if(\Core\ftp()){
				return new Backends\DirectoryFTP($resolved . $filename);
			}
			else{
				return new Backends\DirectoryLocal($resolved . $filename);
			}

			break;
		default:
			throw new \Exception('Unsupported CDN type: ' . CDN_TYPE);
			break;
	}
}




/**
 * Function to translate a PHP upload error into a human-readable string.
 *
 * If no error is seen, a blank string is returned.
 *
 * @param int
 * @return string
 */
function translate_upload_error($errno){
	switch($errno){
		case UPLOAD_ERR_OK:
			return '';
		case UPLOAD_ERR_INI_SIZE:
			if(DEVELOPMENT_MODE){
				return 'The uploaded file exceeds the upload_max_filesize directive in php.ini [' . ini_get('upload_max_filesize') . ']';
			}
			else{
				return 'The uploaded file is too large, maximum size is ' . ini_get('upload_max_filesize');
			}
		case UPLOAD_ERR_FORM_SIZE:
			if(DEVELOPMENT_MODE){
				return 'The uploaded file exceeds the MAX_FILE_SIZE directive that was specified in the HTML form. ';
			}
			else{
				return 'The uploaded file is too large.';
			}
		default:
			return 'An error occurred while trying to upload the file.';
	}
}

/**
 * Check a file mimetype against a base.
 *
 * Useful for upload scripts.
 *
 * @param $acceptlist string List of accept types to accept.
 *                    Can be wildcard types too, for example image/*, or simply *.
 * @param $mimetype string File Mimetype to check against
 * @param $extension string Extension of the file to check against.
 *                  (not required, but provides better true positives, [and potentially more false positives] ).
 *
 * @return string Empty string if passed, else the error.
 */
function check_file_mimetype($acceptlist, $mimetype, $extension = null){
	$acceptgood = false;
	$accepts = array_map(
		'trim',
		explode(
			',',
			strtolower($acceptlist)
		)
	);

	// Also lowercase the incoming extension.
	$extension = strtolower($extension);

	foreach($accepts as $accepttype){
		// '*' is the wildcard to accept any filetype....
		// why would this even be set?!?
		if($accepttype == '*'){
			$acceptgood = true;
			break;
		}
		// accepts that are standard full mimetypes are also pretty easy.
		elseif(preg_match('#^[a-z\-\+]+/[0-9a-z\-\+\.]+#', $accepttype)){
			if($accepttype == $mimetype){
				$acceptgood = true;
				break;
			}
		}
		// wildcard mimetypes are allowed too.
		elseif(preg_match('#^[a-z\-\+]+/\*#', $accepttype)){
			if(strpos($mimetype, substr($accepttype, 0, -1)) === 0){
				$acceptgood = true;
				break;
			}
		}
		// extensions are allowed as well, (if provided)
		elseif($extension && preg_match('#^\.*#', $accepttype)){
			if(substr($accepttype, 1) == $extension){
				$acceptgood = true;
				break;
			}
		}
		// Umm....
		else{
			return 'Unsupported accept option, ' . $accepttype;
		}
	}

	// Now that all the mimetypes have run through, I can see if one matched.
	if(!$acceptgood){
		if(sizeof($accepts) > 1){
			$err = 'matches one of [ ' . implode(', ', $accepts) . ' ]';
		}
		else{
			$err = 'is a ' . $accepts[0] . ' file';
		}
		return 'Invalid file uploaded, please ensure it ' . $err;
	}
	else{
		return '';
	}
}


/**
 * Simple function to try to guess the mimetype for a given file extension.
 * This is not guaranteed to be the actual mimetype of the file, but just the most common for a given extension.
 *
 * @param string $ext
 *
 * @return string
 */
function extension_to_mimetype($ext){
	\Core\log_deprecated('Filestore\\extension_to_mimetype', debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS, 2)[1]);
	return FileHelper::GuessMimeFromExt($ext);
}

/**
 * Convert a common mimetype to its extension.
 *
 * @param $mimetype
 *
 * @return string
 */
function mimetype_to_extension($mimetype){
	\Core\log_deprecated('Filestore\\mimetype_to_extension', debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS, 2)[1]);
	return FileHelper::GuessExtFromMime($mimetype);
}

/**
 * Get the resource for FTP based CDN connections.
 *
 * Returns the FTP resource or false on failure.
 *
* @return FTPConnection | false
 */
function cdn_ftp(){
	static $ftp = null;

	if($ftp === null){

		$ftp = new FTPConnection();
		$ftp->host = CDN_FTP_HOST;
		$ftp->username = CDN_FTP_USERNAME;
		$ftp->password = CDN_FTP_PASSWORD;
		$ftp->root = CDN_FTP_PATH;
		$ftp->url = CDN_FTP_URL;
/*
		try{
			$ftp->connect();

			$ftp->reset();
		}
		catch(\Exception $e){
			\Core\ErrorManagement\exception_handler($e);
			$ftp = false;
			return false;
		}
		*/
	}
/*
	if($ftp && $ftp instanceof FTPConnection){
		try{
			$ftp->reset();
		}
		catch(\Exception $e){
			\Core\ErrorManagement\exception_handler($e);
			$ftp = false;
			return false;
		}
	}
*/

	return $ftp;
}


/**
 * Get an array of the various resize components from a given dimension set.
 * These include: width, height, mode, key.
 *
 * @param string|int $dimensions
 * @param File $file
 *
 * @return array
 */
function get_resized_key_components($dimensions, $file){
	// The legacy support for simply a number.
	if (is_numeric($dimensions)) {
		$width  = $dimensions;
		$height = $dimensions;
		$mode = '';
	}
	elseif ($dimensions === null) {
		$width  = 300;
		$height = 300;
		$mode = '';
	}
	elseif($dimensions === false){
		$width = false;
		$height = false;
		$mode = '';
	}
	else {
		// Allow some special modifiers.
		if(strpos($dimensions, '^') !== false){
			// Fit the smallest dimension instead of the largest, (useful for overflow tricks)
			$mode = '^';
			$dimensions = str_replace('^', '', $dimensions);
		}
		elseif(strpos($dimensions, '!') !== false){
			// Absolutely resize, regardless of aspect ratio
			$mode = '!';
			$dimensions = str_replace('!', '', $dimensions);
		}
		elseif(strpos($dimensions, '>') !== false){
			// Only increase images.
			$mode = '>';
			$dimensions = str_replace('>', '', $dimensions);
		}
		elseif(strpos($dimensions, '<') !== false){
			// Only decrease images.
			$mode = '<';
			$dimensions = str_replace('<', '', $dimensions);
		}
		else{
			// Default mode
			$mode = '';
		}
		// New method. Split on the "x" and that should give me the width/height.
		$vals   = explode('x', strtolower($dimensions));
		$width  = (int)$vals[0];
		$height = (int)$vals[1];
	}

	$ext = $file->getExtension();
	// Ensure that an extension is used if none present, (may happen with temporary files).
	if(!$ext){
		$ext = mimetype_to_extension($file->getMimetype());
	}
	$fileBase = \Core\str_to_url($file->getBasename(true));
	
	// If this file already has a dimension hash attached, drop that off the end.
	if(preg_match('/-[0-9]*x[0-9]*$/', $fileBase)){
		$fileBase = preg_replace('/-[0-9]*x[0-9]*$/', '', $fileBase);
	}
	
	// Ensure this isn't too long, just in case.
	if(strlen($fileBase) > 42){
		$fileBase = substr($fileBase, 0, 42);
	}
	
	if(strpos($file->getMimetype(), 'image/svg') === 0){
		$sizeKey = '';
	}
	else{
		$sizeKey = '-' . $width . 'x' . $height . $mode;
	}

	// The basename is for SEO purposes, that way even resized images still contain the filename.
	// The hash is just to ensure that no two files conflict, ie: /public/a/file1.png and /public/b/file1.png
	//  might conflict without this hash.
	// Finally, the width and height dimensions are there just because as well; it gives more of a human
	//  touch to the file. :p
	// Also, keep the original file extension, this way PNGs remain PNGs, GIFs remain GIFs, JPEGs remain JPEGs.
	// This is critical particularly when it comes to animated GIFs.
	$key = $fileBase . '-' . $file->getHash() . $sizeKey . '.' . $ext;

	// The directory can be used with the new File backend to create this file in a correctly nested subdirectory.
	$dir = dirname($file->getFilename(false)) . '/';

	if(substr($dir, 0, 7) == 'public/'){
		// Replace the necessary prefix with a more useful one.
		// Anything within public/ needs to be remapped to public/tmp
		$dir = 'public/tmp/' . substr($dir, 7);
	}
	else{
		// Everything else gets prepended to public/tmp/
		// so if the original file is in themes/blah/imgs/blah.jpg,
		// it will be copied to public/tmp/blah.jpg
		$dir = 'public/tmp/';
	}

	return array(
		'width'  => $width,
		'height' => $height,
		'mode'   => $mode,
		'key'    => $key,
		'ext'    => $ext,
		'dir'    => $dir,
	);
}