<?php


namespace Core\Forms;


use Core\CoreStyles;
use Core\Model;

class ButtonCreator extends FormElement {
	private $_defaultColor;
	private $_defaultBackgroundColor;
	private $_defaultColorHover;
	private $_defaultBackgroundColorHover;
	private $_defaultBorderColor;
	private $_defaultBorderColorHover;

	public function  __construct($atts = null) {
		parent::__construct($atts);

		// Some defaults
		$this->_attributes['class'] = 'formelement formbuttoncreator';
		//$this->_attributes['multiple'] = true;
		$this->_validattributes     = array('accesskey', 'dir', 'disabled', 'lang', 'name', 'required', 'tabindex', 'style');

		$this->_defaultColor = strtolower(CoreStyles::GetStyle('button-color'));
		$this->_defaultBackgroundColor = strtolower(CoreStyles::GetStyle('button-background-color'));
		$this->_defaultColorHover = strtolower(CoreStyles::GetStyle('button-color-hover'));
		$this->_defaultBackgroundColorHover = strtolower(CoreStyles::GetStyle('button-background-color-hover'));
		$this->_defaultBorderColor = strtolower(CoreStyles::GetStyle('button-border-color'));
		$this->_defaultBorderColorHover = strtolower(CoreStyles::GetStyle('button-border-color-hover'));
	}

	public function getEmptyValues() : array {
		return [
			'text' => '',
			'href' => '',
			'title' => '',
			'size' => '',
			'color' => null,
			'background_color' => null,
			'color_hover' => null,
			'background_color_hover' => null,
			'border_color' => null,
			'border_color_hover' => null,
		];
	}

	public function getDefaultValues() : array {
		return [
			'text' => '',
			'href' => '',
			'title' => '',
			'size' => '',
			'color' => $this->_defaultColor,
			'background_color' => $this->_defaultBackgroundColor,
			'color_hover' => $this->_defaultColorHover,
			'background_color_hover' => $this->_defaultBackgroundColorHover,
			'border_color' => $this->_defaultBorderColor,
			'border_color_hover' => $this->_defaultBorderColorHover,
		];
	}

	public function render() {

		// W3C states that color inputs must have *something* for the value!
		if(!isset($this->_attributes['value'])) {
			$this->_attributes['value'] = [];
		}

		$defaults = $this->getDefaultValues();

		foreach($defaults as $key => $value) {
			if(
				!isset($this->_attributes['value'][ $key ]) ||
				$this->_attributes['value'][ $key ] === null ||
				$this->_attributes['value'][ $key ] === ''
			) {
				$this->_attributes['value'][ $key ] = $value;
			}
		}

		// Resolve the href to the fully resolved href.
		if($this->_attributes['value']['href']) {
			$this->_attributes['value']['href'] = \Core\resolve_link($this->_attributes['value']['href']);
		}

		return parent::render();
	}

	/**
	 * This set explicitly handles the value, and has the extended logic required
	 *  for error checking and validation.
	 *
	 * @param mixed $value The value to set
	 * @return boolean
	 */
	public function setValue($value) {
		$defaultValues = $this->getDefaultValues();
		$emptyValues = $this->getEmptyValues();
		$newValues = [];

		if(is_array($value)) {
			foreach($defaultValues as $key => $defaultValue) {
				if(!isset($value[ $key ])) {
					$newValues[ $key ] = $emptyValues[ $key ];
				}
				elseif($value[ $key ] === $defaultValue || $value[ $key ] === $emptyValues[ $key ]) {
					$newValues[ $key ] = $emptyValues[ $key ];
				}
				else {
					$valueCheck = trim($value[ $key ]);
					switch($key) {
						case 'text':
							if(!($valid = $this->validateText($valueCheck))) {
								$this->_error = 'Button text failed validation - ' . $valid;
								return false;
							}
							break;
						case 'title':
							if(!($valid = $this->validateText($valueCheck))) {
								$this->_error = 'Button title failed validation - ' . $valid;
								return false;
							}
							break;
						case 'href':
							if(!($valid = $this->validateLink($valueCheck))) {
								$this->_error = 'Button link failed validation - ' . $valid;
								return false;
							}
							$valueCheck = \Core\reverse_resolve_link($valueCheck);
							break;
						case 'size':
							if(!($valid = $this->validateSize($valueCheck))) {
								$this->_error = 'Button size failed validation - ' . $valid;
								return false;
							}
							break;
						case 'color':
							if(!($valid = $this->validateColor($valueCheck))) {
								$this->_error = 'Color failed validation - ' . $valid;
								return false;
							}
							break;
						case 'background_color':
							if(!($valid = $this->validateColor($valueCheck))) {
								$this->_error = 'Background color failed validation - ' . $valid;
								return false;
							}
							break;
						case 'color_hover':
							if(!($valid = $this->validateColor($valueCheck))) {
								$this->_error = 'Hover color failed validation - ' . $valid;
								return false;
							}
							break;
						case 'background_color_hover':
							if(!($valid = $this->validateColor($valueCheck))) {
								$this->_error = 'Background hover color failed validation - ' . $valid;
								return false;
							}
							break;
						case 'border_color':
							if(!($valid = $this->validateColor($valueCheck))) {
								$this->_error = 'Border color failed validation - ' . $valid;
								return false;
							}
							break;
						case 'border_color_hover':
							if(!($valid = $this->validateColor($valueCheck))) {
								$this->_error = 'Hover border color failed validation - ' . $valid;
								return false;
							}
							break;
					}
					$newValues[ $key ] = $valueCheck;
				}
			}
		}
		else {
			$value = $emptyValues;
		}

		if($this->get('required')) {
			if($value['text'] === $emptyValues['text'] && $value['link'] === $emptyValues['link']) {
				$this->_error = $this->get('title') . ' is required!';
				return false;
			}
		}

		$this->_attributes['value'] = $newValues;
		return true;
	}

	public function validateText(string $text) {
		if(strlen($text) > 128) {
			return 'text must be 128 characters or less';
		}

		return true;
	}

	public function validateLink(string $link) {
		if(strlen($link) > 128) {
			return 'URL must be 128 characters or less';
		}


		/*if(!preg_match(Model::VALIDATION_URL, $link)) {
			return 'URL does not appear to be a valid link';
		}*/


		return true;
	}

	public function validateSize(string $size) {
		$valid = ['', 'small', 'large', 'xlarge'];
		if(!in_array($size, $valid)) {
			return 'Invalid size requested';
		}
		return true;
	}

	public function validateColor(string $color) {
		if(!preg_match('/^#[a-fA-F0-9]{6}$/', $color)) {
			return 'Invalid color format, please ensure it is a full hex code, starting with #';
		}
		return true;
	}
}