<?php
/**
 * Class file for the several FormPage* classes that exist.
 *
 * @package Core\Forms
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Forms;


use Core\Theme;

/**
 * Class FormPageRewriteURLInput provides a text input with custom javascript to pull from the title and urlify it.
 *
 * @package Core\Forms
 */
class PageRewriteURLInput extends TextInput{

}

/**
 * Input type for the page selection input.  Only really useful from within the PageModel.
 *
 * @package Core\Forms
 */
class PagePageSelectInput extends SelectInput{
	public function __construct($atts = null){

		parent::__construct($atts);
	}

	public function render(){

		if(!$this->get('templatename')){
			throw new \Exception('Unable to render pageselectinput element without templatename set!');
		}
		// Figure out the template directory for custom pages, (if it exists)
		// In order to get the types, I need to sift through all the potential template directories and look for a directory
		// with the matching name.
		$tmpname = $this->get('templatename') . '/';

		$matches = array();

		foreach(\Core\Templates\Template::GetPaths() as $d){
			if(is_dir($d . $tmpname)){
				// Yay, sift through that and get the files!
				$dir = \Core\Filestore\Factory::Directory($d . $tmpname);
				foreach($dir->ls(['tpl', 'phtml']) as $file){
					// Skip directories
					if($file instanceof \Core\Filestore\Directory) continue;

					/** @var $file \Core\Filestore\File */
					//$fullpath = $tmpname . $file->getBaseFilename();
					$fullpath = $file->getBaseFilename(true);
					$name = $file->getBaseFilename();
					// Do some template updates and make it a little more friendlier to read.
					$name = ucwords(str_replace(['-', '_'], ' ', substr($name, 0, -4)));
					$matches[ $fullpath ] = $name;
				}
			}
		}

		// Are there any matches?  If not just return a blank string.
		if(!sizeof($matches)){
			return '';
		}
		
		// Sort!
		asort($matches);

		$options = array_merge(array('' => '-- Generic Page (No Taxonomy) --'), $matches);
		$this->set('options', $options);

		return parent::render();
	}
}

/**
 * Class FormPageParentSelectInput provides a select input with the list of selectable parent URLs in the system.
 *
 * @package Core\Forms
 */
class PageParentSelectInput extends SelectInput{

	public function setValue($value){
		// This strtoloweris required here because base baseurls are case insensitive, but the form system is case sensitive.
		return parent::setValue(strtolower($value));
	}

	public function render(){
		$f = new \Core\ModelFactory('PageModel');
		if ($this->get('baseurl')) $f->where('baseurl != ' . $this->get('baseurl'));
		$opts = \PageModel::GetPagesAsOptions($f, '-- No Parent Page --');

		$this->set('options', $opts);

		// And continue!
		return parent::render();
	}
}


/**
 * Provides inputs for editing:
 * rewriteurl, parenturl, theme template and page template
 * along with page insertables.
 *
 * @deprecated 2013.07.11 cpowell - Candidate for immediate removal.
 */
class PageMeta extends FormGroup {

	public function  __construct($atts = null) {
		error_log(__CLASS__ . ' is candidate for immediate removal, please change this code!', E_USER_DEPRECATED);

		// Defaults
		$this->_attributes['name']    = 'page';

		if ($atts instanceof \PageModel) {
			parent::__construct(array('name' => 'page'));

			$page = $atts;
		}
		else {
			if(isset($atts['model']) && $atts['model'] instanceof \PageModel){
				// Everything is based off the page.
				$page = $atts['model'];
				unset($atts['model']);

				parent::__construct($atts);
			}
			else{
				parent::__construct($atts);

				// BaseURL needs to be set for this to work.
				//if(!$this->get('baseurl')) return null;

				// Everything is based off the page.
				$page = new \PageModel($this->get('baseurl'));
			}
		}

		$this->_attributes['baseurl'] = $page->get('baseurl');
		$name = $this->_attributes['name'];

		// I need to get a list of pages to offer as a dropdown for selecting the "parent" page.
		$f = new \Core\ModelFactory('PageModel');
		if ($this->get('baseurl')) $f->where('baseurl != ' . $this->get('baseurl'));
		$opts = \PageModel::GetPagesAsOptions($f, '-- No Parent Page --');

		$this->addElement(
			'pageparentselect',
			array(
				'name'    => $name . "[parenturl]",
				'title'   => 'Parent Page',
				'value'   => strtolower($page->get('parenturl')),
				'options' => $opts
			)
		);

		// Title
		$this->addElement(
			'text', array(
				      'name'        => $name . "[title]",
				      'title'       => 'Title',
				      'value'       => $page->get('title'),
				      'description' => 'Every page needs a title to accompany it, this should be short but meaningful.',
				      'required'    => true
			      )
		);

		// Rewrite url.
		$this->addElement(
			'pagerewriteurl', array(
				                'name'        => $name . "[rewriteurl]",
				                'title'       => 'Page URL',
				                'value'       => $page->get('rewriteurl'),
				                'description' => 'Starts with a "/", omit ' . ROOT_URL,
				                'required'    => true
			                )
		);

		$this->addElement(
			'access', array(
				        'name'  => $name . "[access]",
				        'title' => 'Access Permissions',
				        'value' => $page->get('access')
			        )
		);

		$this->addElement(
			'pagemetas',
			array(
				'name' => $name . '_meta',
				'model' => $page,
			)
		);

		// Give me all the skins available on the current theme.
		$skins = array('' => '-- Site Default Skin --');
		foreach(Theme::GetTheme(null)->getSkins() as $s) {
			$n = ($s['title']) ? $s['title'] : $s['file'];
			if($s['default']) $n .= ' (default)';
			$skins[$s['file']] = $n;
		}
		if(sizeof($skins) > 2){
			$this->addElement(
				'select', array(
					        'name'    => $name . "[theme_template]",
					        'title'   => 'Theme Skin',
					        'value'   => $page->get('theme_template'),
					        'options' => $skins
				        )
			);
		}
	}

	/**
	 * Save the elements back to the database for the bound base_url.
	 */
	public function save() {

		$page = $this->getModel();

		//if($this->get('cachedaccess')) $page->set('access', $this->get('cachedaccess'));
		//if($this->get('cachedtitle')) $page->set('title', $this->get('cachedtitle'));
		return $page->save();

		// Ensure the children have the right baseurl if that changed.
		$els = $this->getElements();
		foreach ($els as $e) {
			if (!preg_match('/^insertable\[(.*?)\].*/', $e->get('name'), $matches)) continue;
			$e->set('baseurl', $this->get('baseurl'));
		}

		// And all the insertables.
		$i = $this->getElementByName('insertables');
		$i->save();

		return true;
	}

	/**
	 * Get the model for the page subform.
	 * This is on the group because a page object can be set embedded in another form.
	 *
	 * @param null $page
	 *
	 * @return null|\PageModel
	 */
	public function getModel($page = null) {
		// Allow linked models.
		if (!$page) $page = new \PageModel($this->get('baseurl'));

		// Because name can be changed.
		$name = $this->_attributes['name'];

		$els = $this->getElements(true, false);
		foreach ($els as $e) {
			/** @var $e FormElement */

			if (!preg_match('/^[a-z_]*\[(.*?)\].*/', $e->get('name'), $matches)) continue;

			$key = $matches[1];
			$val = $e->get('value');

			// Meta attributes
			if(strpos($e->get('name'), $name . '_meta') === 0){
				$val = $e->get('value');
				$page->setMeta($key, $val);
			}
			elseif(strpos($e->get('name'), $name) === 0){
				$page->set($key, $val);
			}
			else{
				continue;
			}
		}

		return $page;

		// Add in any insertables too, if they're attached.
		// DISABLING 2012.05 cpowell
		/*
		if(($i = $this->getElementByName('insertables'))){
			$els = $i->getElements();
			foreach($els as $e){
				if(!preg_match('#^insertable\[(.*?)\].*#', $e->get('name'), $matches)) continue;

				$submodel = $page->findLink('Insertable', array('name' => $matches[1]));
				$submodel->set('value', $e->get('value'));
			}
		}

		return $page;
		*/
	}

	/**
	 * This group does not have a template of its own, should be rendered directly to the form.
	 * @return null
	 */
	public function getTemplateName() {
		return null;
	}

//

} // class FormPageInsertables