<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core\Installer;
use Exception;
use Core\BaconIpsumGenerator;
use Core\XMLLoader;

/**
 * Description of Configuration
 *
 * @author charlie
 */
class Configuration {
	private $_values;
	private $_xmlLoader;
	
	public function __construct(){
		if(!file_exists(ROOT_PDIR . 'core/resources/configuration.xml')){
			throw new Exception('No example configuration found!');
		}
		
		$this->_xmlLoader = new XMLLoader();
		$this->_xmlLoader->setRootName('configuration');
		$this->_xmlLoader->loadFromFilename(ROOT_PDIR . 'core/resources/configuration.xml');
	}
	
	public function getValues() : array {
		if($this->_values === null){
			$this->_generateValues();
		}
		return $this->_values;
	}
	
	public function setValues($newValues){
		$this->_values = $newValues;
	}
	
	public function getFormElements() : array {
		// Retrieve the current set values, (or generate new ones)
		$values = $this->getValues();
		
		$elements = $this->_xmlLoader->getElements('return|define');
		$formelements = [];
				
		foreach($elements as $el){
			$node        = $el->nodeName;
			$name        = $el->getAttribute('name');
			$type        = $el->getAttribute('type');
			$formtype    = $el->getAttribute('formtype');
			$advanced    = $el->getAttribute('advanced');
			$children    = $el->childNodes;
			$value       = $values[ $name ];
			$description = null;
			$options     = [];
			
			if($advanced === null || $advanced === ''){
				$advanced = '1';
			}

			foreach($children as $c){
				switch($c->nodeName){
					case 'description':
						$description = trim($c->nodeValue);
						break;
					case 'option':
						$options[] = trim($c->nodeValue);
						break;
				}
			}

			// Throw this element onto the array for the template to render out.
			$formelements[] = [
				'name'        => $name,
				// Make the title more appealing than machine names...
				'title'       => ucwords(strtolower(str_replace('_', ' ', $name))),
				// Remap "formtype" to "type", since this will be used in a form afterall!
				'type'        => $formtype,
				'value'       => $value,
				'description' => $description,
				'options'     => $options,
				'advanced'    => $advanced,
			];
		}
		
		return $formelements;
	}
	
	public function getContents() : string{
		$values = $this->getValues();
		
		$elements = $this->_xmlLoader->getElements('return|define');
		foreach($elements as $el){
			$name        = $el->getAttribute('name');
			$children    = $el->childNodes;

			foreach($children as $c){
				// Iterate through each child node of this return or define looking for the 'value' node.
				if($c->nodeName == 'value' && isset($values[$name])){
					// An override from the user is available for this value, use that instead!
					$c->nodeValue = $values[$name];
				}
			}
		}

		return $this->_xmlLoader->asPrettyXML();
	}
	
	private function _generateValues(){
		$this->_values = [];

		$pw = new \Core\Utilities\PasswordGenerator();
		$pw->ambiguousLetters = true;
		$pw->symbols = 0;
		
		$elements = $this->_xmlLoader->getElements('return|define');

		$pw->minLength = 20;
		$encryptionPassphrase = $pw->generate();

		$pw->minLength = 15;
		$databasePassphrase = $pw->generate();
		
		// Similar treatment for the database user; random info.
		$scriptOwner = get_current_user();
		$base = str_replace('-', '', $scriptOwner) . '_';
		$databaseUser = $base . strtolower(\Core\random_hex(rand(4,6)));
		$databaseName = $base . strtolower(\Core\random_hex(rand(4,6)));

		// Since we're pulling from the ant version, set some nice defaults for the user.
		$defaults = [
			'database_server'              => 'localhost',
			'database_port'                => '3306',
			'database_type'                => 'mysqli',
			'database_name'                => $databaseName,
			'database_user'                => $databaseUser,
			'database_pass'                => $databasePassphrase,
			'SERVER_ID'                    => \Core\random_hex(32),
			'DEVELOPMENT_MODE'             => 'false',
			'tmp_dir_web'                  => '/tmp/' . HOST . '-web/',
			'tmp_dir_cli'                  => '/tmp/' . HOST . '-cli/',
			'SECRET_ENCRYPTION_PASSPHRASE' => $encryptionPassphrase,
		];
		
		foreach($elements as $el){
			$key = $el->getAttribute('name');
			if(isset( $defaults[ $key ])){
				// Set from a dynamic default value
				$this->_values[ $key ] = $defaults[ $key ];
			}
			else{
				// Pulled from the existing value in the XML
				$children    = $el->childNodes;
				foreach($children as $c){
					if($c->nodeName === 'value'){
						$this->_values[ $key ] = trim($c->nodeValue);
					}
				}
			}
		}
	}
	
	private function _generatePassphrase($wordLength){
		// Use Bacon ipsum to generate a secret passphrase for the user.
		$bacon = new BaconIpsumGenerator();
		
		$baconWords = str_replace(' ', '', $bacon->getWord($wordLength));
		
		// Manipulate the string a bit to add some complexity.
		for($i = 0; $i < strlen($baconWords); $i++){
			$change = rand(0, 50);
			if($change < 30){
				continue;
			}
			elseif($change < 45){
				$baconWords[$i] = strtoupper($baconWords[$i]);
			}
			elseif($change < 47){
				$set = ['!', '@', '#', '$', '%', '^', '*', '(', ')', '{', '}', '[', ']', '?', '~'];
				$baconWords[$i] = $set[ rand(0, sizeof($set) - 1) ];
			}
			else{
				$set = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'];
				$baconWords[$i] = $set[ rand(0, sizeof($set) - 1) ];
			}
		}
		
		return $baconWords;
	}
}
