<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core\Installer\Checks;
use Exception;
use Core\Installer\InstallHelper;

/**
 * Description of DirectoryProtection
 *
 * @author charlie
 */
class DirectoryProtectionCheck implements CheckInterface {
	private $_dest;
	private $_src;
	private $_templateValues = [
		'@{rewritebase}@' => ROOT_WDIR,
	];
	
	public function __construct(string $dir, string $type){
		switch($type){
			case 'assets':
			case 'private':
			case 'public':
			case 'root':
				break;
			default:
				throw new Exception('Invalid protection type provided!  Please use only [assets,private,public,root]');
		}

		// Ensure directory ends in a trailing slash.
	   if(substr($dir, -1) !== '/'){
		   $dir .= '/';
	   }
	   
		$server = false;

		if(function_exists('apache_get_version')){
			$server = apache_get_version();
		}

		if($server === false || $server === 'Apache') {
			// Default to Apache 2.4.
			// This is the case for both running in FCGI mode and with Prod server tokens.
			$server = 'Apache/2.4';
		}
		
		// The server version determines which version of the directory protection is required.
		if(strpos($server, 'Apache/2.4') === 0){
			$this->_src = ROOT_PDIR . 'core/resources/apache-2-4/htaccess.' . $type;
			$this->_dest = $dir . '.htaccess';
		}
		elseif(strpos($server, 'Apache/2.2') === 0){
			$this->_src = ROOT_PDIR . 'core/resources/apache-2-2/htaccess.' . $type;
			$this->_dest = $dir . '.htaccess';
		}
		else{
			throw new Exception('Unsupported server, ' . $server . '.  Unable to detect what directory protection is required.');
		}
		// @todo Support Lighttpd and nginx.

		if(!file_exists($this->_src)){
			throw new Exception('Source protection file for ' . $server . ' not located!');
		}
	}
	
	public function checkContents() : bool {
		if(!file_exists($this->_dest)){
			return false;
		}
		
		// Spaces, newlines, and tabs seem to trip up this check, but are otherwise irrelivent to the file.
		$target = str_replace([' ', "\n", "\t"], '', file_get_contents($this->_dest));
		$source = str_replace([' ', "\n", "\t"], '', $this->getContents());
		
		return ($target == $source);
	}
	
	public function checkPermissions() : bool {
		$perms = substr(sprintf('%o', fileperms($this->_dest)), -4);
		
		return $perms === '0444';
	}
	
	public function getFix() : string {
		$fixes = [];

		if(!is_dir(dirname($this->_dest))){
			$fixes[] = 'mkdir -p ' . escapeshellarg(dirname($this->_dest));
		}

		if(file_exists($this->_dest) && !is_readable($this->_dest)) {
			$fixes[] = 'chmod a+r ' . escapeshellarg($this->_dest);
		}
		elseif(!$this->checkContents()){
			// Here we simplify the amount of code needed to be copy/pasted to reduce errors.
			$fix = 'cat "' . $this->_src . '" ';
			foreach($this->_templateValues as $k => $v) {
				$fix .= '| sed \'s#' . $k . '#' . $v . '#g\' ';
			}
			$fix .= '> "' . $this->_dest . '"';

			$fixes[] = $fix;
		}
		
		/*if(!$this->checkPermissions()){
			$fix .= 'chmod 0444 ' . escapeshellarg($this->_dest) . NL;
		}*/
		
		return implode(NL, $fixes);
	}

	/**
	 * Get additional details about this check, but not necessarily about the fix.
	 * @return string
	 */
	public function getInformation() : string {
		return '';
	}
	
	public function getContents() : string {
		$fdata = file_get_contents($this->_src);
		$fdata = str_replace(array_keys($this->_templateValues), array_values($this->_templateValues), $fdata);
		
		return $fdata;
	}
	
	public function getTitle() : string {
		$name = (dirname($this->_dest) . '/' === ROOT_PDIR) ?
			'application root' :
			dirname(substr($this->_dest, strlen(ROOT_PDIR)));
		
		return 'Directory protection for ' . $name;
	}
	
	public function getDetails() : string {
		$test = $this->test();

		if($test) {
			return 'The directory ' . dirname($this->_dest) . ' is protected by the access file ' . basename($this->_dest);
		}
		else {
			if(InstallHelper::CheckIsCPanel()) {
				$header = '<ol>' .
					'<li>Browse to CPanel >> Files >> File Manager</li>' .
					'<li>Click "Settings" in the top-right and ensure "Show Hidden Files" is selected</li>' .
					'<li>Browse to <strong><code>' . dirname($this->_dest) . '</code></strong> and create the folder if necessary</li>' .
					'<li>Create a new file if necessary called <strong><code>' . basename($this->_dest) . '</code></strong> and paste the following contents inside</li>' .
					'</ol>';
			}
			else {
				$header = 'Create a new located in ' . dirname($this->_dest) . ' with the name "' . basename($this->_dest) . '" and paste the following contents';
			}

			return $header . '<br/>' .
				'<pre class="code-htaccess"><code>' . htmlentities($this->getContents()) . '</code></pre>';
		}
	}
	
	public function setTemplateValue($key, $overrideValue){
		$this->_templateValues[ $key ] = $overrideValue;
	}
	
	public function test() : bool {
		// This test is safe to perform for most destinations.
		if(
			(dirname($this->_dest) . '/' === ROOT_PDIR) && 
			file_exists($this->_dest)
		){
			// No operations required; this allows the owner to tweak the root .htaccess file as necessary.
			return true;
		}
		
		if($this->checkContents()){
			// It passed anyway!
			return true;
		}
		
		// Try to auto-create!
		if(
			!is_dir(dirname($this->_dest)) &&
			is_dir(dirname($this->_dest, 2)) && 
			is_writable(dirname($this->_dest, 2))
		){
			mkdir(dirname($this->_dest), 0755);
			if(EXEC_MODE === 'CLI'){
				chown(dirname($this->_dest), InstallHelper::GetWebServerUser());
				chgrp(dirname($this->_dest), InstallHelper::GetWebServerGroup());
			}
		}
		
		if(
			(!file_exists($this->_dest) && is_writable(dirname($this->_dest))) ||
			is_writable($this->_dest) 
		){
			// File itself exists and is writeable
			// or doesn't but can be created.
			if(file_put_contents($this->_dest, $this->getContents()) === false) {
				return false;
			}
			
			if(EXEC_MODE === 'CLI'){
				chown($this->_dest, InstallHelper::GetWebServerUser());
				chgrp($this->_dest, InstallHelper::GetWebServerGroup());
			}
			chmod($this->_dest, 0644);
			
			return true;
		}
		
		return false;
	}

	public function testExtended(): string {
		return $this->test() ? self::STATUS_PASSED : self::STATUS_FAILED;
	}
}
