<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core\Installer\Checks;

use Core\Installer\InstallHelper;

/**
 * Description of PHPExtensionCheck
 *
 * @author charlie
 */
class PHPExtensionCheck implements CheckInterface {
	
	private $_extension;
	
	public function __construct(string $extension) {
		$this->_extension = $extension;
	}
	
	/**
	 * Get the title of this check
	 * @return string
	 */
	public function getTitle() : string {
		return 'PHP Extension ' . $this->_extension;
	}
	
	/**
	 * Run the test for this check
	 * @return bool True if passed, False otherwise
	 */
	public function test() : bool {
		$loaded = extension_loaded($this->_extension);
		
		if($loaded){
			return true;
		}
		elseif(EXEC_MODE === 'CLI' && ($fix = $this->getFix()) !== ''){
			// AUTO-INSTALL!
			$ret = null;
			$out = [];
			exec($fix, $out, $ret);
			return ($ret === 0);
		}
		else{
			return false;
		}
	}

	public function testExtended(): string {
		return $this->test() ? self::STATUS_PASSED : self::STATUS_FAILED;
	}
	
	/**
	 * Get additional details that may help the user fix the problem
	 * @return string
	 */
	public function getDetails() : string {
		if($this->test()) {
			return $this->_extension . ' ' . phpversion($this->_extension) . ' detected';
		}
		else {
			if(InstallHelper::CheckIsCPanel()) {
				return $this->_getDetailsCPanel();
			}
			else {
				return $this->_getDetailsManual();
			}
		}
	}

	private function _getDetailsCPanel() : string {
		$version = new \Core\VersionString(phpversion());
		$cpanelVersion = 'ea-php' . $version->major . $version->minor;

		if($this->_extension === 'hash') {
			return 'Cryptographic hash capability requires PHP to be compiled without the "--disable-hash" flag!';
		}
		elseif($this->_extension === 'zlib') {
			return 'ZLib compression and decompresion requires PHP to be compiled with the "--with-zlib" flag!';
		}
		elseif($this->_extension === 'sodium') {
			return 'If you have access to the root WHM:<br/>' .
				'<ol>' .
				'<li>Browse to WHM >> Server Configuration >> Terminal</li>' .
				'<li>Click "I understand and want to proceed" if prompted</li>' .
				'<li>Type the following commands (note, copy/paste should <strong>not</strong> be used here)' .
				'<pre class="code-shell"><code>' .
				'<span class="cli-root-line">yum install -y epel-release</span>' . NL .
				'<span class="cli-root-line">yum update -y</span>' . NL .
				'<span class="cli-root-line">yum install -y libsodium-devel</span>' . NL .
				'</code></pre>' .
				'</li>' .
				'<li><strong>THEN</strong>, browse to WHM >> Software >> Module Installers</li>' .
				'<li>Select "Manage" for "PHP PECL"</li>' .
				'<li>Select the <strong><code>' . $cpanelVersion . '</code></strong> PHP version to manage and press "Apply"</li>' .
				'<li>Search for <strong><code>' . $this->_extension . '</code></strong> in the "Find a PHP PECL" section</li>' .
				'<li>Install the cooresponding module</li>' .
				'</ol>';
		}
		elseif($this->_extension === 'yaml') {
			return 'If you have access to the root WHM:<br/>' .
				'<ol>' .
				'<li>Browse to WHM >> Server Configuration >> Terminal</li>' .
				'<li>Click "I understand and want to proceed" if prompted</li>' .
				'<li>Type the following commands (note, copy/paste should <strong>not</strong> be used here)' .
				'<pre class="code-shell"><code>' .
				'<span class="cli-root-line">yum install -y libyaml-devel</span>' . NL .
				'</code></pre>' .
				'</li>' .
				'<li><strong>THEN</strong>, browse to WHM >> Software >> Module Installers</li>' .
				'<li>Select "Manage" for "PHP PECL"</li>' .
				'<li>Select the <strong><code>' . $cpanelVersion . '</code></strong> PHP version to manage and press "Apply"</li>' .
				'<li>Search for <strong><code>' . $this->_extension . '</code></strong> in the "Find a PHP PECL" section</li>' .
				'<li>Install the cooresponding module</li>' .
				'</ol>';
		}
		elseif(
			$this->_extension === 'bcmath' ||
			$this->_extension === 'curl' ||
			$this->_extension === 'gd' ||
			$this->_extension === 'mbstring' ||
			$this->_extension === 'xml' ||
			$this->_extension === 'zip'
		) {
			return 'If you have access to the root WHM:<br/>' .
				'<ol>' .
				'<li>Browse to WHM >> Software >> EasyApache 4</li>' .
				'<li>Click "Customize" for "Current Installed Packages"</li>' .
				'<li>Click on "PHP Extensions"</li>' .
				'<li>Search for <strong><code>' . $this->_extension . '</code></strong> in the search field</li>' .
				'<li>Tick to install the modules for all cooresponding PHP versions</li>' .
				'<li>Click on "Review"</li>' .
				'<li>Review the changes to be performed and click "Provision" at the bottom to apply the changes</li>' .
				'</ol>';
		}
		else{
			return 'Install the ' . $this->_extension . ' extension for PHP';
		}
	}

	private function _getDetailsManual() : string {
		if($this->_extension === 'hash') {
			return 'Cryptographic hash capability requires PHP to be compiled without the "--disable-hash" flag!';
		}
		elseif($this->_extension === 'zlib') {
			return 'ZLib compression and decompresion requires PHP to be compiled with the "--with-zlib" flag!';
		}
		else{
			return 'Install the ' . $this->_extension . ' extension for PHP';
		}
	}
	
	/**
	 * Get the scripted fix for this problem that provides the user how to fix it
	 * @return string
	 */
	public function getFix(): string {
		if(InstallHelper::CheckIsCPanel()) {
			return $this->_getFixCPanel();
		}
		else {
			return $this->_getFixManual();
		}
	}

	private function _getFixCPanel() : string {
		$version = new \Core\VersionString(phpversion());

		if($this->_extension === 'hash' || $this->_extension === 'zlib') {
			// These extensions do not support easy fixes.
			return '';
		}
		elseif($this->_extension === 'sodium') {
			$fixes = [];

			$fixes[] = InstallHelper::GetPackageInstallCommand() . ' epel-release';
			$fixes[] = InstallHelper::GetPackageInstallCommand() . ' libsodium-devel';
			$fixes[] = 'sudo /opt/cpanel/ea-php' . $version->major . $version->minor . '/root/usr/bin/pecl install libsodium';
			$fixes[] = 'sudo systemctl restart httpd';

			return implode(' && ' . NL, $fixes);
		}
		elseif($this->_extension === 'yaml') {
			$fixes = [];
			$fixes[] = InstallHelper::GetPackageInstallCommand() . ' libyaml-devel';
			$fixes[] = 'sudo /opt/cpanel/ea-php' . $version->major . $version->minor . '/root/usr/bin/pecl install yaml';
			$fixes[] = 'sudo systemctl restart httpd';

			return implode(' && ' . NL, $fixes);
		}
		elseif(
			$this->_extension === 'bcmath' ||
			$this->_extension === 'curl' ||
			$this->_extension === 'gd' ||
			$this->_extension === 'mbstring' ||
			$this->_extension === 'xml' ||
			$this->_extension === 'zip'
		) {
			$fixes = [];
			$fixes[] = InstallHelper::GetPackageInstallCommand() . ' ea-php' . $version->major . $version->minor . '-php-' . $this->_extension;
			$fixes[] = 'sudo systemctl restart httpd';
			return implode(' && ' . NL, $fixes);
		}
		elseif($this->_extension === 'mysql') {
			// mysql has an additional requirement, but is otherwise auto-executable.
			if(trim(exec('which mysql')) === '') {
				return InstallHelper::GetPackageInstallCommand() . ' mariadb-client' .
					' && ' . NL .
					InstallHelper::GetPackageInstallCommand() . ' php' . $version->major . '.' . $version->minor . '-' . $this->_extension;
			}
			else {
				return InstallHelper::GetPackageInstallCommand() . ' php' . $version->major . '.' . $version->minor . '-' . $this->_extension;
			}
		}
		else {
			return InstallHelper::GetPackageInstallCommand() . ' php' . $version->major . '.' . $version->minor . '-' . $this->_extension;
		}
	}

	private function _getFixManual() : string {
		$version = new \Core\VersionString(phpversion());

		if($this->_extension === 'hash' || $this->_extension === 'zlib') {
			// These extensions do not support easy fixes.
			return '';
		}
		elseif($this->_extension === 'mysql') {
			// mysql has an additional requirement, but is otherwise auto-executable.
			if(trim(exec('which mysql')) === '') {
				return InstallHelper::GetPackageInstallCommand() . ' mariadb-client' .
					' && ' . NL .
					InstallHelper::GetPackageInstallCommand() . ' php' . $version->major . '.' . $version->minor . '-' . $this->_extension;
			}
			else {
				return InstallHelper::GetPackageInstallCommand() . ' php' . $version->major . '.' . $version->minor . '-' . $this->_extension;
			}
		}
		else {
			return InstallHelper::GetPackageInstallCommand() . ' php' . $version->major . '.' . $version->minor . '-' . $this->_extension;
		}
	}
}
