<?php

namespace Core\ListingTable\Renderer;

use Core\Core;
use Core\ListingTable\Action;
use Core\ListingTable\Column;
use Core\ListingTable\Report;
use Core\ListingTable\Table;
use Core\ListingTable\TableInterface;
use Core\Model;
use Core\Templates\Template;
use Exception;

/**
 * Class TableRenderer
 *
 * Renders the available reports for this table and provides a listing for the user to select
 * which report to run.
 *
 * This DOES NOT power the actual report; that is handled by the action on the specific report.
 *
 * @package Core\ListingTable\Renderer
 */
class ReportRenderer implements RendererInterface {
	/** @var array|Report[] */
	private $_reports;
	/** @var null|Report */
	private $_selected;

	/**
	 * TableRenderer constructor.
	 */
	public function __construct(TableInterface $table) {
		$this->_reports = $table->getReports();
		$selected = $table->getReportSelected();
		if($selected) {
			try {
				$this->_selected = $table->getReport($selected);
			}
			catch(Exception $ex) {
				$this->_selected = null;
			}
		}
	}

	/**
	 * Render the contents of this interface with all the requested data
	 *
	 * @return array
	 *         +pager   bool   REQUIRED true/false if the pager is enabled
	 *         +options string REQUIRED HTML output of options for this interface
	 *         +body    string REQUIRED HTML output of the body
	 *         +page    int    Current page number, required if pager is set to true
	 *         +total   int    Total number of results, required if pager is set to true
	 *         +pages   int    Total number of pages, required if pager is set to true
	 */
	public function render() : array {
		if($this->_selected === null) {
			return [
				'body'    => $this->_renderReportSelection(),
				'download' => null,
			];
		}
		else {
			return [
				'body'    => $this->_renderReport(),
				'download' => (Core::IsLibraryAvailable('HTMLtoPDF') ? 'pdf' : null),
			];
		}
	}

	private function _renderReportSelection() : string {
		$tpl = Template::Factory('includes/listingtable/report-selector');
		$tpl->assign('reports', $this->_reports);
		return $tpl->fetch();
	}

	private function _renderReport() : string {
		$file = $this->_selected->generate(false);
		$html = $file->getContents();
		$file->delete();

		return $html;
	}

	private function _renderOptions() : string {
		$tpl = Template::Factory('includes/listingtable/report-options');
		$tpl->assign('report', $this->_selected);
		return $tpl->fetch();
	}
}