<?php

namespace Core\Page;

use Core\ListingTable\Table;
use Core\FilterForm;
use Core\Core;
use Core\Component;
use Core\Datamodel\Dataset;
use PageMetaModel;
use PageModel;

class PageTable extends Table {
	
	public function __construct() {
		$this->setModelName('PageModel');

		parent::__construct();
	}

	protected function _setupFilters() {
		$components = Core::GetComponents();

		$componentopts = ['' => '-- ' . t('STRING_VIEW_ALL_COMPONENTS') . ' --'];
		foreach($components as $c){
			/** @var Component $c */
			$componentopts[$c->getKeyName()] = $c->getName();
		}
		// Sort them by name!
		asort($componentopts);

		// Load all the tags on the site from the various pages and load them as a filter.
		$pageMetas = Dataset::Init()
			->select('meta_value', 'meta_value_title')
			->table('page_meta')
			->where('meta_key = keyword')
			->unique(true)
			->order('meta_value_title ASC')
			->executeAndGet();
		$pageMetaOptions = ['' => '-- All Tags --'];
		foreach($pageMetas as $dat){
			$pageMetaOptions[$dat['meta_value']] = $dat['meta_value_title'];
		}

		// Load the parent URLs to provide some assistance for selecting child pages.
		$parentOptions = [];
		$parents = Dataset::Init()
			->select('parenturl')
			->table('page')
			->whereNotEquals('parenturl', '')
			->unique(true)
			->whereEquals('admin', '0')
			->whereEquals('selectable', '1')
			->executeAndGet();

		if(is_array($parents)) {
			foreach($parents as $url) {
				$parent = PageModel::GetByURI($url);
				$parentOptions[ $url ] = $parent->getFullTitle() . ' (' . $parent->rewriteurl . ')';
			}
		}
		asort($parentOptions);
		$parentOptions = array_merge(
			['' => '-- All Pages --', '_NONE_' => 'No Parent Page'],
			$parentOptions
		);
		
		// Pull the templates used on the site to allow the user to filter by specific taxonomies
		$templateOptions = [];
		$templates = Dataset::Init()
			->select('last_template')
			->table('page')
			->whereNotEquals('last_template', '')
			->unique(true)
			->whereEquals('admin', '0')
			->whereEquals('selectable', '1')
			->executeAndGet();
		if(is_array($templates)) {
			foreach($templates as $path) {
				// Trim pages/ off the beginning, this is redundant.
				if(strpos($path, 'pages/') === 0) {
					$pathName = substr($path, 6);
				}
				else {
					$pathName = $path;
				}

				$templateOptions[ $path ] = $pathName;
			}
		}
		asort($templateOptions);
		$templateOptions = array_merge(
			['' => '-- All Templates --'],
			$templateOptions
		);


		$this->addFilter(
			'text',
			[
				'name' => 'search',
				'title' => 'Search',
				'link' => FilterForm::LINK_TYPE_SEARCH,
			]
		);

		$this->addFilter(
			'text',
			[
				'name' => 'title',
				'title' => t('STRING_TITLE'),
				'link' => FilterForm::LINK_TYPE_CONTAINS,
			]
		);


		$this->addFilter(
			'select',
			[
				'name' => 'parenturl',
				'title' => t('STRING_PARENT_URL'),
				'options' => $parentOptions,
			]
		);

		$this->addFilter(
			'select',
			[
				'name' => 'last_template',
				'title' => t('STRING_MODEL_PAGEMODEL_PAGE_TEMPLATE'),
				'options' => $templateOptions,
				'link' => FilterForm::LINK_TYPE_STANDARD,
			]
		);

		$this->addFilter(
			'select',
			[
				'name' => 'component',
				'title' => t('STRING_COMPONENT'),
				'options' => $componentopts,
				'link' => FilterForm::LINK_TYPE_STANDARD,
			]
		);

		$this->addFilter(
			'select',
			[
				'name' => 'page_types',
				'title' => t('STRING_INCLUDE_ADMIN_PAGES'),
				'options' => ['all' => t('STRING_VIEW_ALL_PAGES'), 'no_admin' => t('STRING_EXCLUDE_ADMIN_PAGES')],
				'value' => 'no_admin',
			]
		);

		if(sizeof($pageMetaOptions) > 1){
			$this->addFilter(
				'select',
				[
					'name' => 'keyword',
					'title' => 'Page Keyword',
					'options' => $pageMetaOptions
				]
			);
		}
	}

	protected function _processFilters() {
		parent::_processFilters();

		$ds = $this->getModelFactory()->getDataset();

		if($this->getFilterValue('page_types') == 'no_admin'){
			$ds
				->whereEquals('admin', '0')
				->whereEquals('selectable', '1');
		}

		$parentFilter = $this->getFilterValue('parenturl');
		if($parentFilter === '_NONE_') {
			$ds->whereEquals('parenturl', '');
		}
		elseif($parentFilter !== '' && $parentFilter !== null) {
			$ds->whereEquals('parenturl', $parentFilter);
		}
		// No else required; else is simply "MATCH ALL"

		if($this->getFilterValue('keyword')){
			$pageMetas = PageMetaModel::FindRaw(
				['meta_value = ' . $this->getFilterValue('keyword'), 'meta_key = keyword']
			);
			$pageURLs = [];
			foreach($pageMetas as $row){
				$pageURLs[] = $row['baseurl'];
			}

			if(count($pageURLs)) {
				$ds->whereEquals('baseurl', $pageURLs);
			}
		}
	}

	protected function _setupColumns() {
		$this->addColumn('title_link')->disableSortable();
		$this->addColumn('rewriteurl')->hidden();
		$this->addColumn('baseurl')->hidden();
		$this->addColumn('pageviews')->hidden();
		$this->addColumn('popularity');
		$this->addColumn('expires')->hidden();
		$this->addColumn('created')->hidden();
		$this->addColumn('updated')->hidden();
		$this->addColumn('published_status')->hidden();
		$this->addColumn('status_details')->disableSortable();
		$this->addColumn('published')->hidden();
		$this->addColumn('published_expires')->hidden();
		$this->addColumn('teaser')->hidden();
		$this->addColumn('access')->hidden();
		$this->addColumn('component')->hidden();
		$this->addColumn('seo_details')->disableSortable();
	}
}
