<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core\Parser\JPEGExif;

use Core\Core;
use Core\Filestore\File;
use Core\Parser\ImageMetaTag;
use Core\Parser\ImageParser;
use Core\Parser\JPEGExif\Tags\GPSLatitudeTag;
use Core\Parser\JPEGExif\Tags\GPSLongitudeTag;

/**
 * Parser for JPEG Exif metadata
 * 
 * uses the following I18N strings: (required to extract out strings by the scanner)
 * 
 * t:STRING_CORE_PARSER_IMAGEMETA_DATETIME
 * t:STRING_CORE_PARSER_IMAGEMETA_MAKE
 * t:STRING_CORE_PARSER_IMAGEMETA_MODEL
 * t:STRING_CORE_PARSER_IMAGEMETA_SOFTWARE
 * t:STRING_CORE_PARSER_IMAGEMETA_ARTIST
 * t:STRING_CORE_PARSER_IMAGEMETA_COPYRIGHT
 * t:STRING_CORE_PARSER_IMAGEMETA_ORIENTATION
 * t:STRING_CORE_PARSER_IMAGEMETA_EXPOSURETIME
 * t:STRING_CORE_PARSER_IMAGEMETA_EXPOSUREPROGRAMDESC
 * t:STRING_CORE_PARSER_IMAGEMETA_ISOSPEEDRATINGS
 * t:STRING_CORE_PARSER_IMAGEMETA_FNUMBER
 * t:STRING_CORE_PARSER_IMAGEMETA_FOCALLENGTH
 * t:STRING_CORE_PARSER_IMAGEMETA_METERINGMODEDESC
 * t:STRING_CORE_PARSER_IMAGEMETA_FLASH
 * t:STRING_CORE_PARSER_IMAGEMETA_LIGHTSOURCEDESC
 * t:STRING_CORE_PARSER_IMAGEMETA_GPSLATITUDE
 * t:STRING_CORE_PARSER_IMAGEMETA_GPSLONGITUDE
 * t:STRING_CORE_PARSER_IMAGEMETA_GPSALTITUDE
 * t:STRING_CORE_PARSER_IMAGEMETA_GPSIMGDIRECTION
 * t:STRING_CORE_PARSER_IMAGEMETA_GEO
 *
 * @author charlie
 */
class JPEGExifParser extends ImageParser {
	/**
	 * @var ImageMetaTag[] Array of tags from this EXIF data
	 */
	public $fields = [];
	
	public function parseFile(File $file) {

		$this->fields = [];


		if($file->getMimetype() !== 'image/jpeg') {
			// Only JPEGs support this type of meta parser.
			return;
		}

		$exif = exif_read_data($file->getFilename());

		if(!$exif) {
			// This image doesn't have EXIF metadata, nothing to do.
			return;
		}
		
		// Geographic data is used by a few fields, so lookup the data first if it's available.
		if(
			Core::IsComponentAvailable('geographic-codes') &&
			!isset($exif['Geo']) &&
			isset($exif['GPSLatitude']) &&
			isset($exif['GPSLongitude'])
		){
			$lat = new GPSLatitudeTag($exif, 'GPSLatitude');
			$lng = new GPSLongitudeTag($exif, 'GPSLongitude');
			
			$exif['Geo'] = \GeoLookup\Lookup::ByLatLng($lat->formatted, $lng->formatted);
		}
		
		$mappings = [
			'DateTime'            => 'DateTimeFieldTag',
			'Geo'                 => 'GeoLocationTag',
			'Make'                => 'ExifFieldTag',
			'Model'               => 'ExifFieldTag',
			'Software'            => 'ExifFieldTag',
			'Artist'              => 'ExifFieldTag',
			'Copyright'           => 'ExifFieldTag',
			'Orientation'         => 'OrientationTag',
			'ExposureTime'        => 'ExposureTimeTag',
			'ExposureProgramDesc' => 'ExposureProgramTag',
			'ISOSpeedRatings'     => 'ISOSpeedTag',
			'FNumber'             => 'FNumberTag',
			'FocalLength'         => 'FocalLengthTag',
			'MeteringModeDesc'    => 'MeteringModeTag',
			'Flash'               => 'FlashTag',
			'LightSourceDesc'     => 'LightSourceTag',
			'GPSLatitude'         => 'GPSLatitudeTag',
			'GPSLongitude'        => 'GPSLongitudeTag',
			'GPSAltitude'         => 'GPSAltitudeTag',
			'GPSImgDirection'     => 'GPSDirectionTag',
		];
		
		
		foreach($mappings as $key => $classname) {
			if(
				isset($exif[ $key ]) && 
				(
					(is_scalar($exif[ $key ]) && trim($exif[ $key ]) !== '') ||
					(!is_scalar($exif[ $key ]) && isset($exif[ $key ]))
				)
			) {
				$classname = 'Core\\Parser\\JPEGExif\\Tags\\' . $classname;
				$field = new $classname($exif, $key);
				if($field->formatted !== ''){
					$this->fields[ $key ] = $field;
				}
			}
		}
	}
}
