<?php
/**
 * File for class ModelHelper definition in Core Plus
 * 
 * @package Core\Search
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20131126.1744
 * @copyright Copyright (C) 2009-2021  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Search;
use Core\Datamodel\Dataset;
use Core\Datamodel\DatasetSelect;
use Core\Datamodel\DatasetWhereClause;
use Core\Model;


/**
 * A short teaser of what ModelHelper does.
 *
 * More lengthy description of what ModelHelper does and why it's fantastic.
 *
 * <h3>Usage Examples</h3>
 *
 *
 * @todo Write documentation for ModelHelper
 * <h4>Example 1</h4>
 * <p>Description 1</p>
 * <code>
 * // Some code for example 1
 * $a = $b;
 * </code>
 *
 *
 * <h4>Example 2</h4>
 * <p>Description 2</p>
 * <code>
 * // Some code for example 2
 * $b = $a;
 * </code>
 *
 * 
 * @package Core\Search
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class Helper {
	/**
	 * Hook handler to save the index data for a given model record.
	 *
	 * @param Model $model
	 *
	 * @return bool
	 */
	public static function ModelPreSaveHandler(Model $model) {
		// Only process models that have a search index on them.
		if(!$model::$HasSearch) return true;

		// Retrieve the search index from the Model's integrated helper.
		// this includes additional information such as weights and ordering.
		$index = $model->getSearchIndex();

		// Update the Model and check if a new save is required.
		$model->search_index_str = $index->getString();
		if($model->checkIsChanged('search_index_str')) {
			$model->search_index_pri = $index->getPrimary();
			$model->search_index_sec = $index->getSecondary();
		}

		return true;
	}

	/**
	 * Add the necessary SELECT, WHERE, and ORDER clauses to a dataset to perform a search
	 *
	 * Can be used directly against Datasets used for Model retrieval,
	 * or manual lookups.
	 *
	 * @param string $query Textual query to perform
	 * @param Dataset $ds Dataset to manipulate
	 * @param float $relevancy Minimum relevancy for matching results
	 *
	 * @throws \Core\Datamodel\DMIException
	 * @throws \Core\Datamodel\DMIQueryException
	 */
	public static function AddSearchToDataset(string $query, Dataset $ds, float $relevancy = 0.5) {
		list($indexes, $primaries, $secondaries) = self::ConvertStringToKeywords($query);

		$ds->select(
			DatasetSelect::Init()
				->alias('__search_query')
				->value($query)
		);

		$subwhere = new DatasetWhereClause('search');
		$subwhere->setSeparator('or');

		$selectCount    = 0;
		$selectFormulas = [];
		$selectParts = [];

		$matchFormulas = [];

		// And add a where clause for each one.
		foreach($indexes as $word){
			$subwhere->addWhere('search_index_str LIKE %' . $word . '%');
			$selectFormulas[] = 'LEAST((LENGTH(search_index_str) - LENGTH(REPLACE(search_index_str, %s, ""))), 1) * 3';
			$matchFormulas[] = '(MATCH(search_index_str) AGAINST (%s IN BOOLEAN MODE)) * 3';
			$selectParts[] = $word;
			$selectCount += 3;
		}
		foreach($primaries as $word){
			if(strlen($word) > 2) {
				// We need at least 3 syllables for this lookup to really work.
				$subwhere->addWhere('search_index_pri LIKE %' . $word . '%');
				$selectFormulas[] = 'LEAST((LENGTH(search_index_pri) - LENGTH(REPLACE(search_index_pri, %s, ""))), 1) * 2';
				$matchFormulas[] = '(MATCH(search_index_pri) AGAINST (%s IN BOOLEAN MODE)) * 2';
				$selectParts[] = $word;
				$selectCount += 2;
			}
		}
		foreach($secondaries as $word){
			if(strlen($word) > 2) {
				// We need at least 3 syllables for this lookup to really work.
				$subwhere->addWhere('search_index_sec LIKE %' . $word . '%');
				$selectFormulas[] = 'LEAST((LENGTH(search_index_sec) - LENGTH(REPLACE(search_index_sec, %s, ""))), 1) * 1';
				$matchFormulas[] = '(MATCH(search_index_sec) AGAINST (%s IN BOOLEAN MODE)) * 1';
				$selectParts[] = $word;
				$selectCount += 1;
			}
		}

		if($selectCount > 0) {
			$searchRelevancySelect = new DatasetSelect();
			$searchRelevancySelect->formula = '( ' . implode(' + ', $selectFormulas) . ' ) / ' . $selectCount;
			$searchRelevancySelect->alias = '__search_relevancy';
			$searchRelevancySelect->formulaParams = $selectParts;

			$searchWeightSelect = new DatasetSelect();
			$searchWeightSelect->formula = '( ' . implode(' + ', $matchFormulas) . ' ) / ' . $selectCount;
			$searchWeightSelect->alias = '__search_weight';
			$searchWeightSelect->formulaParams = $selectParts;

			$ds->where($subwhere)
				->select($searchRelevancySelect)
				->select($searchWeightSelect)
				->having('__search_relevancy >= ' . $relevancy)
				->order('__search_weight DESC');
		}
		else {
			// No valid search parameters were found, but we need something to allow the query to continue, albeit with no results.
			$subwhere->addWhere('search_index_pri = -1');

			$searchRelevancySelect = new DatasetSelect();
			$searchRelevancySelect->value = 0;
			$searchRelevancySelect->alias = '__search_relevancy';

			$searchWeightSelect = new DatasetSelect();
			$searchWeightSelect->value = 0;
			$searchWeightSelect->alias = '__search_weight';

			$ds->where($subwhere)
				->select($searchRelevancySelect)
				->select($searchWeightSelect);
		}
	}

	/**
	 * Sanitize and convert a query string to the set of queries necessary for searching using double metaphone.
	 *
	 * Will return an array consisting of 3 keys:
	 * * 0 base value of the string
	 * * 1 primary double metaphonic string
	 * * 2 secondary double metaphonic string
	 *
	 * @param string $query
	 *
	 * @return array
	 */
	public static function ConvertStringToKeywords(string $query) : array {
		// Lowercase it to make operations easier
		$query       = strtolower($query);

		// Convert this string to latin.
		$query       = \Core\str_to_latin($query);

		// Skip punctuation.
		$query       = preg_replace('/[^a-z0-9 ]/', '', $query);

		// Split out words.
		$parts       = explode(' ', $query);

		// Retrieve a list of "stop" words; ie: words that have little search relevancy.
		$skips       = \Core\get_stop_words();

		$indexes     = [];
		$primaries   = [];
		$secondaries = [];

		foreach($parts as $word){
			if(!$word) {
				// Skip empty words
				continue;
			}

			if(in_array($word, $skips)) {
				// Skip short or common words
				continue;
			}

			$it = new DoubleMetaPhone($word);

			$indexes[] = $word;

			if($it->primary) {
				$primaries[] = $it->primary;
			}

			if($it->secondary) {
				$secondaries[] = $it->secondary;
			}
		}

		// Remove duplicates
		$indexes = array_unique($indexes);
		$primaries = array_unique($primaries);
		$secondaries = array_unique($secondaries);

		return [$indexes, $primaries, $secondaries];
	}
} 
