<?php


namespace Core\Updater;


use Core\ConfigHandler;
use Core\Date\DateTime;
use Core\Filestore\Directory;
use Core\Filestore\Factory as FileFactory;
use Core\Filestore\File;
use Exception;
use Ifsnop\Mysqldump\Mysqldump;
use mikehaertl\shellcommand\Command;
use ZipArchive;

class ArchiveExporter {

	private $_exportTables = [];

	private $_exportPublic = false;

	private $_exportCustom = false;

	/** @var Directory */
	private $_archiveDirectory;

	/** @var string */
	private $_archiveFilename;

	/**
	 * Add a table to be exported in the datafile
	 *
	 * @param string $tablename
	 */
	public function exportTable(string $tablename) {
		$this->_exportTables[] = $tablename;
	}

	public function exportPublicFiles() {
		$this->_exportPublic = true;
	}

	public function exportCustomFiles() {
		$this->_exportCustom = true;
	}

	/**
	 * Process the actual export operation with the requested assets
	 */
	public function process() : File {
		$this->_archiveDirectory = FileFactory::Directory('tmp/export-' . \Core\random_hex(8));
		$this->_archiveDirectory->mkdir();

		$this->_archiveFilename = $this->_archiveDirectory->getPath() . \Core\str_to_url(SITENAME) . '-archive-' . DateTime::Now() . '.tar';

		if(count($this->_exportTables)) {
			$this->_processTables();
		}

		if($this->_exportPublic) {
			$this->_processPublicFiles();
		}

		if($this->_exportCustom) {
			$this->_processCustomFiles();
		}

		$this->_processReadme();

		return FileFactory::File($this->_archiveFilename);
	}

	public function __destruct() {
		if($this->_archiveDirectory !== null) {
			$this->_archiveDirectory->delete();
		}
	}

	private function _processTables() {
		// If the user opted to export database tables, export those!
		$tableDest = $this->_archiveDirectory->getPath() . 'data.sql.gz';

		$settings = ConfigHandler::LoadConfigFile('configuration');

		switch($settings['database_type']) {
			case 'mysqli':
			case 'mysql':
				$dsn = 'mysql';
				break;
			default:
				throw new Exception('Unable to export database, unsupported database type');
		}

		// mysql:host=localhost;port=3307;dbname=testdb
		$dsn .= ':host=' . $settings['database_server'] . ';port=' . $settings['database_port'] . ';dbname=' . $settings['database_name'];

		$exporter = new Mysqldump(
			$dsn,
			$settings['database_user'],
			$settings['database_pass'],
			[
				'include-tables' => $this->_exportTables,
				'compress'       => Mysqldump::GZIP,
				'add-drop-table' => true,
			]
		);

		$exporter->start($tableDest);

		$cmd = new Command('tar');
		$cmd->addArgKey('--append');
		$cmd->addArgKeyValue('--file', $this->_archiveFilename);
		$cmd->addArgKeyValue('--directory', $this->_archiveDirectory->getPath());
		$cmd->addArgKey('data.sql.gz');

		$cmd->execute();
	}

	private function _processPublicFiles() {
		$dir = FileFactory::Directory('public/');

		$cmd = new Command('tar');
		$cmd->addArgKey('--append');
		$cmd->addArgKeyValue('--file', $this->_archiveFilename);
		$cmd->addArgKeyValue('--directory', $dir->getPath());
		$cmd->addArgKeyValue('--exclude', './tmp');
		$cmd->addArgKeyValue('--transform', 's:^./:public/:');
		$cmd->addArgKey('.');

		$cmd->execute();
	}

	private function _processCustomFiles() {
		$dir = FileFactory::Directory(ROOT_PDIR . 'themes/custom/');

		$cmd = new Command('tar');
		$cmd->addArgKey('--append');
		$cmd->addArgKeyValue('--file', $this->_archiveFilename);
		$cmd->addArgKeyValue('--directory', $dir->getPath());
		$cmd->addArgKeyValue('--transform', 's:^./:custom/:');
		$cmd->addArgKey('.');

		$cmd->execute();
	}

	private function _processReadme() {
		$destination = $this->_archiveDirectory->getPath() . 'README.md';

		$site = ROOT_URL;
		$date = \Core\Date\DateTime::Now('FDT');
		$email = \Core\user()->email;

		$text = <<<EOT
# Site Export Archive

* Original Site: {$site}
* Creation Date: {$date}
* Creator: {$email}

## Contents
EOT;

		if(count($this->_exportTables)) {
			$tables = $this->_exportTables;
			sort($tables);

			$text .= "\n\n### Database Export\n\nIncludes the following database tables, these will be overwritten completely when imported.\n";
			$text .= "\n* " . implode("\n* ", $tables) . "\n";
		}

		if($this->_exportPublic) {
			$text .= "\n\n### Public Files" .
				"\n\nThese include everything uploaded to the site and will be merged with the destination import, overwriting existing files, but keeping non-changed files\n";
		}

		if($this->_exportCustom) {
			$text .= "\n\n### Custom Files" .
				"\n\nThese include theme, CSS, Javascript, and template changes performed that get saved to themes/custom\n";
		}

		file_put_contents($destination, $text);

		$cmd = new Command('tar');
		$cmd->addArgKey('--append');
		$cmd->addArgKeyValue('--file', $this->_archiveFilename);
		$cmd->addArgKeyValue('--directory', $this->_archiveDirectory->getPath());
		$cmd->addArgKey('README.md');

		$cmd->execute();
	}
}