<?php


namespace Core\Updater;

use Core\ConfigHandler;
use Core\Date\DateTime;
use Core\Filestore\Directory;
use Core\Filestore\Factory as FileFactory;
use Core\Filestore\File;
use Exception;
use Ifsnop\Mysqldump\Mysqldump;
use mikehaertl\shellcommand\Command;
use ZipArchive;


class ArchiveImporter {

	/** @var string */
	private $_archive;

	/** @var Directory */
	private $_archiveDirectory;

	/**
	 * @param File $file
	 *
	 * @throws Exception
	 */
	public function process(File $file) {
		// Use the /private datastore here because /tmp will often get remapped due to systemd.
		$this->_archiveDirectory = FileFactory::Directory('private/tmp/import-' . \Core\random_hex(8));
		$this->_archiveDirectory->mkdir();

		$this->_archive = $file->getFilename();

		$cmd = new Command('tar');
		$cmd->addArgKey('--list');
		$cmd->addArgKeyValue('--file', $this->_archive);
		$cmd->addArg('data.sql.gz');
		$cmd->execute();
		if($cmd->getExitCode() === 0) {
			$this->_processDatafile();
		}

		$cmd = new Command('tar');
		$cmd->addArgKey('--list');
		$cmd->addArgKeyValue('--file', $this->_archive);
		$cmd->addArg('public/');
		$cmd->execute();
		if($cmd->getExitCode() === 0) {
			$this->_processPublicFiles();
		}

		$cmd = new Command('tar');
		$cmd->addArgKey('--list');
		$cmd->addArgKeyValue('--file', $this->_archive);
		$cmd->addArg('custom/');
		$cmd->execute();
		if($cmd->getExitCode() === 0) {
			$this->_processCustomFiles();
		}
	}

	public function __destruct() {
		if($this->_archiveDirectory !== null) {
			$this->_archiveDirectory->delete();
		}
	}

	private function _processDatafile() {
		$path = $this->_archiveDirectory->getPath();
		$tempfile = FileFactory::File($path . 'data.sql.gz');

		$cmd = new Command('tar');
		$cmd->addArgKey('--extract');
		$cmd->addArgKeyValue('--file', $this->_archive);
		$cmd->addArgKeyValue('--directory', $path);
		$cmd->addArg('data.sql.gz');
		$cmd->execute();

		if($cmd->getExitCode() !== 0) {
			throw new Exception($cmd->getError(true));
		}

		$settings = ConfigHandler::LoadConfigFile('configuration');

		switch($settings['database_type']) {
			case 'mysqli':
			case 'mysql':
				$dsn = 'mysql';
				break;
			default:
				throw new Exception('Unable to export database, unsupported database type');
		}

		$command = <<<EOC
gunzip -c "{$path}data.sql.gz" | \
mysql \
 --host={$settings['database_server']} \
 --port={$settings['database_port']} \
 --user={$settings['database_user']} \
 --password="{$settings['database_pass']}" \
 --database={$settings['database_name']}
EOC;

		shell_exec($command);

		// Auto cleanup to save diskspace when possible
		$tempfile->delete();
	}

	private function _processPublicFiles() {
		$dir = FileFactory::Directory('public/');
		$dir->mkdir();

		$cmd = new Command('tar');
		$cmd->addArgKey('--extract');
		$cmd->addArgKeyValue('--file', $this->_archive);
		$cmd->addArgKeyValue('--directory', $dir->getPath());
		$cmd->addArgKeyValue('--transform', 's:^public/:./:');
		$cmd->addArgKey('--wildcards');
		$cmd->addArg('public/*');
		$cmd->execute();

		if($cmd->getExitCode() !== 0) {
			throw new Exception($cmd->getError(true));
		}
	}

	private function _processCustomFiles() {
		$dir = FileFactory::Directory(ROOT_PDIR . 'themes/custom/');
		$dir->mkdir();

		$cmd = new Command('tar');
		$cmd->addArgKey('--extract');
		$cmd->addArgKeyValue('--file', $this->_archive);
		$cmd->addArgKeyValue('--directory', $dir->getPath());
		$cmd->addArgKeyValue('--transform', 's:^custom/:./:');
		$cmd->addArgKey('--wildcards');
		$cmd->addArg('custom/*');
		$cmd->execute();

		if($cmd->getExitCode() !== 0) {
			throw new Exception($cmd->getError(true));
		}
	}
}