<?php
/**
 * @package Core\Templates\Smarty
 * @since 1.9
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

/**
 * Primary method for a block of user-customizable content inside a template.
 *
 * Insertables are the core method of injecting blocks of user-customizable content into a template.
 *
 * An insertable must be on a template that has a registered page URL, as the baseurl is what is tracked as one of the main primary keys.
 * The other PK is the insertable's name, which must be unique on that one template.
 *
 * #### Smarty Parameters
 *
 *  * name
 *    * The key name of this input value, must be present and unique on this template.
 *  * assign
 *    * Assign the value instead of outputting to the screen.
 *  * title
 *    * When editing the insertable, the title displayed along side the input field.
 *  * type
 *
 * #### Example Usage
 *
 * <pre>
 * {insertable name="body" title="Body Content"}
 * <p>
 * This is some example content!
 * </p>
 * {/insertable}
 * </pre>
 *
 * <pre>
 * {insertable name="img1" title="Large Image" assign="img1"}
 * {img src="`$img1`" placeholder="generic" dimensions="800x400"}
 * {/insertable}
 * </pre>
 *
 * @param array       $params  Associative (and/or indexed) array of smarty parameters passed in from the template
 * @param string|null $content Null on opening pass, rendered source of the contents inside the block on closing pass
 * @param Smarty      $smarty  Parent Smarty template object
 * @param boolean     $repeat  True at the first call of the block-function (the opening tag) and
 * false on all subsequent calls to the block function (the block's closing tag).
 * Each time the function implementation returns with $repeat being TRUE,
 * the contents between {func}...{/func} are evaluated and the function implementation
 * is called again with the new block contents in the parameter $content.
 *
 * @return string
 */
function smarty_block_insertable($params, $content, $smarty, &$repeat){

	global $SMARTY_IN_INSERTABLEGROUP;

	if(isset($SMARTY_IN_INSERTABLEGROUP) && $SMARTY_IN_INSERTABLEGROUP) {
		// Skip all rendering for insertables inside a group.
		// The group itself will take care of that!
		return;
	}

	$assign = (isset($params['assign']))? $params['assign'] : false;

	// This only needs to be called once.
	// If a value is being assigned, then it's on the first pass so the value will be assigned by the time the content is hit.
	if($assign && !$repeat){
		return $content;
	}
	elseif(!$assign && $repeat){
		// No assign requested, run on the second only.
		return '';
	}

	// I need to use the parent to lookup the current base url.
	$page = Core\Router::$DefaultPageModel;

	if(!isset($params['name'])){
		// If no name was requested, there's nothing to link for this insertable.
		return '';
	}
	
	if(!$page){
		// This page doesn't exist, (or otherwise is a dynamic-only page)
		return '';
	}

	$i = InsertableModel::Construct($page->get('site'), $page->get('baseurl'), $params['name']);

	if($i->checkExists()) {
		$value = $i->get('value');
	}
	elseif(isset($params['value'])) {
		$value = $params['value'];
	}
	else{
		$value = $content;
	}


	// Here we'll use the Form system to determine the specific types
	if(isset($params['type'])) {
		$el = \Core\Forms\FormElement::Factory($params['type'], $params);

		if($el instanceof \Core\Forms\MarkdownInput) {
			$md = new \Core\Converter\MDtoHTML\MarkdownToHTMLConverter();
			if($value !== null) {
				$value = $md->convert($value);
			}
			else {
				$value = '';
			}
		}
		else {
			// Default converter.
			$value = \Core\parse_html($value);
		}
	}
	else {
		// Default converter.
		$value = \Core\parse_html($value);
	}

	if($assign){
		$smarty->assign($assign, $value);
	}
	else{
		return $value;
	}
}