<?php
/**
 * @package Core\Templates\Smarty
 * @since 1.9
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\Core;
use Core\Datamodel\DatasetWhereClause;
use Core\Router;
use Core\ModelFactory;
use Core\Templates\Template;
use Core\View;
use Core\ViewControls;
use function Core\resolve_link;


/**
 * @todo Finish documentation of smarty_function_widgetarea
 * @param array  $params  Associative (and/or indexed) array of smarty parameters passed in from the template
 * @param Smarty_Internal_Template $smarty  Parent Smarty template object
 *
 * @return string|void
 */
function smarty_function_widgetarea($params, $smarty) {
	// Get all widgets set to load in this area.

	$body           = '';
	$baseurl        = Router::$DefaultBaseURL->getBaseURL();
	$page           = Router::$DefaultPageModel;
	$template       = $smarty->template_resource;
	$tmpl           = $smarty->getTemplateVars('__core_template');
	$topview        = Router::$DefaultView;
	$isAdmin        = \Core\user()->checkAccess('p:/core/widgets/manage');
	$inlineControls = $isAdmin ? \Core\user()->getPreference('core-widgets-inline-controls', false) : false;
	
	$parameters  = [];
	$name        = null;
	$installable = null;
	$assign      = null;
	foreach($params as $k => $v){
		switch($k){
			case 'name':
				$name = $v;
				break;
			case 'installable':
				$installable = $v;
				break;
			case 'assign':
				$assign = $v;
				break;
			default:
				$parameters[$k] = $v;
				break;
		}
	}

	// I need to resolve the page template down to the base version in order for the lookup to work.
	foreach(Template::GetPaths() as $base){
		if(strpos($template, $base) === 0){
			$template = substr($template, strlen($base));
			break;
		}
	}

	// Given support for page-level widgets, this logic gets slightly more difficult...
	$factory = new ModelFactory('WidgetInstanceModel');
	$factory->order('weight');
	if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::IsEnabled()){
		$factory->whereGroup('or', ['site = -1', 'site = ' . MultiSiteHelper::GetCurrentSiteID()]);
	}

	$subwhere = new DatasetWhereClause();
	$subwhere->setSeparator('OR');

	// First, the skin-level where clause.
	$skinwhere = new DatasetWhereClause();
	$skinwhere->setSeparator('AND');
	$skinwhere->addWhere('template = ' . $template);
	$skinwhere->addWhere('widgetarea = ' . $name);
	$subwhere->addWhere($skinwhere);

	// And second, the page-level where clause.
	if($page && $page->baseurl){
		$pagewhere = new DatasetWhereClause();
		$pagewhere->setSeparator('AND');
		$pagewhere->addWhere('page_baseurl = ' . $page->baseurl);
		$pagewhere->addWhere('widgetarea = ' . $name);
		$subwhere->addWhere($pagewhere);
	}

	$factory->where($subwhere);

	try{
		$widgets = $factory->get();
	}
	catch(Exception $e){
		if(DEVELOPMENT_MODE){
			$body .= '<p class="message-error">Exception while trying to load widget area ' . $name . '!</p>';
			$body .= '<pre class="xdebug-var-dump">' . $e->getMessage() . '</pre>';
		}
		else{
			\Core\ErrorManagement\exception_handler($e, false);
		}
		$widgets = [];
	}


	foreach ($widgets as $wi) {
		/** @var $wi WidgetInstanceModel */
		// User cannot access this widget? Don't display it...
		if(!\Core\user()){
			continue;
		}
		if (!\Core\user()->checkAccess($wi->get('access'))){
			continue;
		}

		if($installable){
			$wi->set('installable', $installable);
		}
		$view = $wi->execute($parameters);

		// Some widgets may return simply a blank string.  Those should just be ignored.
		if ($view == ''){
			continue;
		}

		// If it's just a string, return that.
		if (is_string($view)) {
			$contents = $view;
		}
		elseif($view->status === View::STATUS_OK){
			// Ensure that the widget's View knows it's linked to a parent!
			$view->parent = $topview;

			$contents = $view->fetch();
		}
		else{
			$contents = 'Error displaying widget [' . $wi->get('baseurl') . '], returned error [' . $view->status . ']';
		}
		
		if($inlineControls) {
			// Does this widget have controls attached to it?
			$widget = $wi->getWidget();
			if($widget === null){
				continue;
			}
			
			// The model contains other useful information such as the standard control links
			// and user edit permissions.
			/** @var $model WidgetModel */
			$model = $wi->getLink('Widget');
			
			$controls = new ViewControls();
			$controls->setProxyForce(true);
			$controls->setProxyText($model->get('title'));
			//$controls->setProxyText(t('STRING_CORE_WIDGET_CONTROLS'));
			
			// Add any on-the-fly widget defined controls.
			// These reside in the custom Widget class.
			if($widget->controls instanceof ViewControls){
				foreach($widget->controls as $link){
					$controls->addLink($link);
				}
			}
			
			// Now I can add the standard WidgetInstance control options.
			$controls->addLinks($wi->getControlLinks());
			
			// Now I can add the standard Model control options.
			$controls->addLinks($model->getControlLinks());
			
			if($controls->hasLinks()){
				$contents = $controls->fetch() . $contents;
			}
		}
		

		$body .= $contents;
	}

	if($body || $inlineControls) {
		// Do some sanitizing for the css data
		$class = 'widgetarea-' . strtolower(str_replace(' ', '', $name));
		
		if($inlineControls) {
			if($template) {
				$skinLevel = (strpos($template, 'skins/') === 0);
				$icon = $skinLevel ? 'core-widgets-skins.svg' : 'core-widgets-pages.svg';
				$title = $skinLevel ? 'Skin Widgets' : 'Page Widgets';

				$type = '<img' .
					' src="' . \Core\resolve_asset('assets/images/icons/' . $icon) . '"' .
					' style="max-height:32px; vertical-align:middle;"' .
					' title="' . $title . '"/>';
				
				$editLink = '  <a href="' . \Core\resolve_link('/widget/admin?template=' . $template) . '"' .
					' title="Manage Widgets Here" class="button">' .
					'<i class="icon icon-cogs"></i> ' .
					'<span>Manage</span>' .
					'</a>';
			}
			else {
				$type = '';
				$editLink = '';
			}
			
			$class .= ' widgetarea-inline-editable';
			$body = '<div class="widgetarea-title">' . $type . $name . $editLink . '</div>' . $body;
		}

		$html = '<div class="widgetarea ' . $class . '" widgetarea="' . $name . '">' . $body . '</div>';
	}
	else{
		$html = '';
	}
	
	if($assign){
		$smarty->assign($assign, $html);
		return '';
	}
	else{
		return $html;
	}
}
