<?php
/**
 * File for class SystemLogModel definition in the coreplus project
 * 
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20131015.2105
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\ConfigHandler;
use Core\Datamodel\Dataset;
use Core\Date\DateTime;
use Core\Model;
use Core\UserAgent;
use geocode\IPLookup;

/**
 * Class SystemLogModel
 *
 * @property string $affected_user_id If this action potentially affects a user, list the ID here.
 * @property string $code             A short phrase or code for this log event, used by sorting
 * @property int    $datetime         Creation timestamp, populated automatically
 * @property string $details          Any details or backtrace that needs to accompany this log event
 * @property string $icon             An optional icon to display along with this log message.
 * @property string $id               Primary ID, as a UUID-formatted string
 * @property string $ip_addr
 * @property string $message          Any primary message that goes with this log event
 * @property string $session_id
 * @property string $source           An optional source to keep track of what type of message this comes from.
 * @property string $type
 * @property string $user_id
 * @property string $useragent
 */
class SystemLogModel extends Model {
	const TYPE_SECURITY = 'security';
	const TYPE_ERROR    = 'error';
	const TYPE_INFO     = 'info';
	const TYPE_DEBUG    = 'debug';

	/** @var null|UserAgent */
	public $_ua = null;

	public static $Schema = array(
		'id' => [
			'type' => Model::ATT_TYPE_UUID,
		],
		'datetime' => [
			'type' => Model::ATT_TYPE_CREATED,
			'formatter' => '\\Core\\Formatter\\GeneralFormatter::DateStringSDT',
		],
		'type' => [
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 32,
			//'options' => ['security', 'error', 'info', 'debug'],
			'default' => 'general',
		],
		'level' => [
			'type' => Model::ATT_TYPE_ENUM,
			'options' => ['critical', 'high', 'info', 'debug'],
			'default' => 'info',
		],
		'session_id' => [
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 160,
		],
		'user_id'    => [
			'type'    => Model::ATT_TYPE_UUID_FK,
			'default' => 0,
			'formatter' => '\\Core\\Formatter\\GeneralFormatter::User',
		],
		'ip_addr'    => [
			'type'      => Model::ATT_TYPE_STRING,
			'maxlength' => 39,
			'formatter' => '\\Core\\Formatter\\GeneralFormatter::IPAddress',
		],
		'useragent' => [
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 256,
			'formatter' => '\\Core\\Formatter\\GeneralFormatter::UserAgent',
		],
		'url' => [
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 1024,
		],
		'referrer' => [
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 1024,
		],
		'affected_user_id'    => [
			'type'    => Model::ATT_TYPE_UUID_FK,
			'default' => null,
			'null'    => true,
			'comment' => 'If this action potentially affects a user, list the ID here.',
			'formatter' => '\\Core\\Formatter\\GeneralFormatter::User',
		],
		'code' => [
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 64,
			'comment' => 'A short phrase or code for this log event, used by sorting'
		],
		'message' => [
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 512,
			'comment' => 'Any primary message that goes with this log event',
			'formatter' => ['this', 'getMessage'],
		],
		'details' => [
			'type' => Model::ATT_TYPE_TEXT,
			'comment' => 'Any details or backtrace that needs to accompany this log event',
		],
		'icon' => [
			'type' => Model::ATT_TYPE_STRING,
			'comment' => 'An optional icon to display along with this log message.',
		],
		'source' => [
			'type' => Model::ATT_TYPE_STRING,
			'comment' => 'An optional source to keep track of what type of message this comes from.',
		],
	);

	public static $Indexes = array(
		'primary'       => ['id'],
		'datetime'      => ['datetime'],
		'user'          => ['user_id'],
		'affected_user' => ['affected_user_id'],
		'code'          => ['code'],
		'source'        => ['source'],
		// Type is required for SELECT DISTINCT `type` FROM `system_log` ORDER BY `type`
		'type'          => ['type'],
	);

	/**
	 * Function that guesses if this user request was a bot.
	 *
	 * @return boolean
	 */
	public function isBot(){
		switch($this->getUserAgent()->type){
			case 'Robot':
			case 'Offline Browser':
			case 'Other':
				return true;
			default:
				return false;
		}
	}

	/**
	 * Get the matching user agent for this model.
	 *
	 * @return UserAgent
	 */
	public function getUserAgent(){
		if($this->_ua === null){
			$this->_ua = new UserAgent($this->get('useragent'));
		}
		return $this->_ua;
	}
	/*
	public function render($key){
		if($key == 'type'){
			switch($this->get('type')){
				case 'info':
					if($this->get('icon')){
						$i = '<i class="icon icon-' . $this->get('icon') . '"></i>';
					}
					else{
						$i = '';
					}
					break;
				case 'error':
					$i = '<i class="icon icon-exclamation" title="Error Entry"></i>';
					break;
				case 'security':
					$i = '<i class="icon icon-exclamation-triangle" title="Security Entry"></i>';
					break;
				default:
					$i = '[ ' . $this->get('type') . ' ]';
			}
			
			return $i . ' ' . $this->get('code');
		}
		elseif($key == 'datetime'){
			return DateTime::FormatString($this->get('datetime'), 'SDT');
		}
		elseif($key == 'ip_addr'){
			$ua = new UserAgent($this->get('useragent'));
			$ip = new IPLookup($this->get('ip_addr'));
			return $ua->getAsHTML() . ' ' . $ip->getAsHTML(true) . ' ' . $this->get('ip_addr');
		}
		else{
			return parent::render($key);
		}
	}
	*/
	
	public function getControlLinks() : array {
		$r = [];
		$id = $this->get('id');
		$ip = $this->get('ip_addr');
		
		if(\Core\user()->checkAccess('g:admin')){
			$r[] = [
				'title' => 'View Details',
				'link' => '/admin/log/details/' . $id,
				'data-target' => 'modal',
				'icon' => 'view',
			];
			
			$r[] = [
				'title' => 'Ban IP',
				'link' => '/security/blacklistip/add?ip_addr=' . $ip . '/32',
				'icon' => 'thumbs-down',
			];
		}
		
		if(\Core\user()->checkAccess('p:/user/activity/view')){
			$r[] = [
				'title' => 'View Activity by IP',
				'link' => '/useractivity/details?filter[ip_addr]=' . $ip,
				'icon' => 'list-alt',
			];
			
			if($this->get('user_id')){
				$r[] = [
					'title' => '/useractivity/details?filter[userid]=' . $this->get('user_id'),
					'title' => 'View Activity by User',
					'icon' => 'user',
				];
			}
		}
		
		$r = array_merge($r, parent::getControlLinks());
		
		return $r;
	}

	public function getMessage(string $format = null) : string {
		if($format === \Core\View::CTYPE_HTML) {
			if($this->icon) {
				$i = '<i class="icon icon-' . $this->icon . '"></i> ';
			}
			else {
				$i = '';
			}

			return $i . $this->message;
		}
		else {
			return $this->message;
		}
	}


	/**
	 * Shortcut method available to log a security-based event on the system
	 *
	 * @param string   $code          Any code that may be associated to this event.
	 * @param string   $message       The subject of the security event.
	 * @param string   $details       Any additional details of the security event beyond the subject line.
	 * @param null|int $affected_user If this security event affected a given user, the affected user's id.
	 */
	public static function LogSecurityEvent($code, $message = '', $details = '', $affected_user = null) {
		try{
			$log = self::Factory();
			$log->set('type', 'security');
			$log->set('code', $code);
			$log->set('message', $message);
			$log->set('details', $details);
			$log->set('affected_user_id', $affected_user);
			$log->save();
		}
		catch(Exception $e){
			// If the model isn't available for some reason, (Core isn't set up, etc),
			// just fall back to the traditional error log mechanism.
			error_log($code . ': ' . $message);
			error_log('ADDITIONALLY, ' . $e->getMessage());
		}
	}

	/**
	 * Shortcut method available to log a error-based event on the system
	 *
	 * @param string $code    Any code that may be associated to this event.
	 * @param string $message The subject of the error event.
	 * @param string $details Any additional details of the error event beyond the subject line.
	 */
	public static function LogErrorEvent($code, $message, $details = '') {
		try{
			$log = self::Factory();
			$log->set('type', 'error');
			$log->set('code', $code);
			$log->set('message', $message);
			$log->set('details', $details);
			$log->save();
		}
		catch(Exception $e){
			// If the model isn't available for some reason, (Core isn't set up, etc),
			// just fall back to the traditional error log mechanism.
			error_log($code . ': ' . $message);
			error_log('ADDITIONALLY, ' . $e->getMessage());
		}
	}

	/**
	 * Shortcut method available to log a informative-based event on the system
	 *
	 * @param string $code    Any code that may be associated to this event.
	 * @param string $message The subject of the info event.
	 * @param string $details Any additional details of the info event beyond the subject line.
	 */
	public static function LogInfoEvent($code, $message, $details = '') {
		try{
			$log = self::Factory();
			$log->set('type', 'info');
			$log->set('code', $code);
			$log->set('message', $message);
			$log->set('details', $details);
			$log->save();
		}
		catch(Exception $e){
			// If the model isn't available for some reason, (Core isn't set up, etc),
			// just fall back to the traditional error log mechanism.
			error_log($code . ': ' . $message);
			error_log('ADDITIONALLY, ' . $e->getMessage());
		}
	}

	/**
	 * Create a new log entry populated with the standard information.
	 *
	 * @return SystemLogModel
	 */
	public static function Factory() {
		$log = new self();
		// @todo These fields can probably be removed as they've been moved to DatabaseLogSink
		$log->setFromArray(
			array(
				'session_id' => session_id(),
				'user_id' => (\Core\user() ? \Core\user()->get('id') : 0),
				'ip_addr' => REMOTE_IP,
				'useragent' => isset($_SERVER['HTTP_USER_AGENT']) ? $_SERVER['HTTP_USER_AGENT'] : '',
			)
		);

		return $log;
	}

	/**
	 * Purge old system logs from the database, as per configuration options.
	 * 
	 * @return bool
	 */
	public static function PurgeHook(){
		$len = ConfigHandler::Get('/core/logs/db/keep');
		if(!$len){
			echo "Not purging any logs, as per configuration option.\n";
			return true;
		}
		
		// Otherwise, the length will be the number of days to keep.
		$d = new DateTime();
		$d->modify('-' . $len . ' days');
		
		echo "Deleting system logs older than " . $d->format(DateTime::FULLDATE) . "\n";
		
		$count = Dataset::Init()
			->count()
			->table('system_log')
			->where('datetime < ' . $d->format('U'))
			->executeAndGet();
		
		echo "Found " . $count . " log entries, deleting!\n";
		Dataset::Init()
			->table('system_log')
			->where('datetime < ' . $d->format('U'))
			->delete()
			->execute();
		
		return true;
	}
}