<?php

use Core\Date\DateTime;
use Core\Date\Timezone;
use Core\Logger\LogEntry;
use Core\Model;
use Core\Router;

class UserDeviceModel extends Model {

	public static $ExportGroup = 't:STRING_SITE_USERS';

	public static $Schema = [
		'id'                   => [
			'type'     => Model::ATT_TYPE_UUID,
			'required' => true,
			'null'     => false,
		],
		'user' => [
			'type' => Model::ATT_TYPE_UUID_FK,
			'comment' => 'Associated user ID',
			'required' => true,
			'null' => false,
			'link' => [
				'model' => 'User',
				'type' => Model::LINK_BELONGSTOONE,
				'on' => 'id',
			],
		],
		'hash' => [
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 40,
			'required' => true,
			'comment' => 'SHA1 hash of the token',
		],
		'useragent' => [
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 128,
			'formatter' => 'Core\\Formatter\\GeneralFormatter::UserAgent',
		],
		'ip_addr'    => [
			'type'      => Model::ATT_TYPE_STRING,
			'maxlength' => 39,
			'formatter' => '\\Core\\Formatter\\GeneralFormatter::IPAddress',
		],
	];

	public static $Indexes = [
		'primary' => ['id'],
		'user'    => ['user'],
	];

	public static $HasCreated = true;
	public static $HasUpdated = true;

	private function _setCookie(string $token) {
		if(!defined('COOKIE_USER_DEVICE_NAME')) {
			define('COOKIE_USER_DEVICE_NAME', 'userdevice');
		}

		$ttl = time() + SECONDS_ONE_MONTH * 3;
		$ssl = SSL_MODE === 'allowed' || SSL_MODE === 'required';

		if(defined('SESSION_COOKIE_DOMAIN') && SESSION_COOKIE_DOMAIN){
			// A valid session name is required for xsite sessions to work. (not sure why)
			$host = SESSION_COOKIE_DOMAIN;
		}
		else {
			$host = $_SERVER['SERVER_NAME'] ?? '';
		}

		setcookie(COOKIE_USER_DEVICE_NAME, $this->id . ':' . $token, $ttl, '/', $host, $ssl);
	}



	/**
	 * Mark this as a failed lookup for a given reason, log the attempt, and delete the device registration.
	 *
	 * @param string $message
	 * @param array $tags
	 */
	private function _failedLookup(string $message, array $tags = []) {
		$log = LogEntry::Init()
			->code('/user/login')
			->type('security')
			->level(LOG_LEVEL_HIGH)
			->icon('user-times')
			->message($message);

		foreach($tags as $k => $v) {
			$log->tag($k, $v);
		}

		if($this->checkExists()) {
			$log->user($this->getLink('UserModel'));
		}

		$log->send();

		self::UnsetCookie();

		if($this->checkExists()) {
			try {
				$this->delete();
			}
			catch(Exception $ex) {
				\Core\log_exception($ex);
			}
		}
	}

	/**
	 * Hook that performs the cookie device auth check.
	 *
	 * Generally called immediately after Core is available.
	 */
	public static function CookieAuthCheck() : bool {
		if(!defined('COOKIE_USER_DEVICE_NAME')) {
			define('COOKIE_USER_DEVICE_NAME', 'userdevice');
		}

		// If there's already a registered Session user, this check is completely skipped.
		// I don't care if it's a valid user, I just care if it's set in general.
		if(\Core\Session::IsKeySet('user')) {
			return true;
		}

		// If there's no device cookie set, also just return.
		if(!filter_has_var( INPUT_COOKIE, COOKIE_USER_DEVICE_NAME)) {
			return true;
		}

		$fullstring = filter_input(INPUT_COOKIE, COOKIE_USER_DEVICE_NAME, FILTER_SANITIZE_URL);

		// This string should match [UUID] : [TOKEN]
		$pos = strpos($fullstring, ':');
		if($pos === false) {
			LogEntry::Init()
				->code('/user/login')
				->type('security')
				->level(LOG_LEVEL_HIGH)
				->tag('cookie', $fullstring)
				->icon('user-times')
				->message('Login via device failed; invalid cookie format')
				->send();

			self::UnsetCookie();

			return true;
		}

		$id = substr($fullstring, 0, $pos);
		$token = substr($fullstring, $pos + 1);

		$model = UserDeviceModel::Construct($id);
		if(!$model->checkExists()) {
			$model->_failedLookup('Login via device failed; device token deleted', ['cookie' => $fullstring]);
			return true;
		}

		/** @var UserModel|null $user */
		$user = $model->getLink('UserModel');

		// The user agent needs to match in order to be a valid user.
		if($model->useragent !== $_SERVER['HTTP_USER_AGENT'] ?? '') {
			$model->_failedLookup('Login via device failed; mismatched user agent', ['cookie' => $fullstring]);
			return true;
		}

		$hash = sha1($token . substr(SECRET_ENCRYPTION_PASSPHRASE, -6));
		if($model->hash !== $hash) {
			$model->_failedLookup('Login via device failed; incorrect verification hash', ['cookie' => $fullstring]);
			return true;
		}

		// Still valid?

		if(!$user) {
			$model->_failedLookup('Login via device failed; user deleted', ['cookie' => $fullstring]);
			return true;
		}

		if($user->get('active') === '0'){
			// The model provides a quick cut-off for active/inactive users.
			// This is the control managed with in the admin.
			$model->_failedLookup('Login via device failed; account not activated yet');
			return true;
		}
		elseif($user->get('active') === '-1'){
			// The model provides a quick cut-off for active/inactive users.
			// This is the control managed with in the admin.
			$model->_failedLookup('Login via device failed; deactivated user account');
			return true;
		}

		// Update this device, resend the cookie, and save the session.
		$model->ip_addr = REMOTE_IP;
		$model->save();
		$model->_setCookie($token);

		// yay...
		$user->set('last_login', DateTime::Now('U', Timezone::TIMEZONE_GMT));
		$user->save();

		\Core\Session::SetUser($user);

		LogEntry::Init()
			->code('/user/login')
			->type('security')
			->level(LOG_LEVEL_INFO)
			->message('Login via device successful')
			->user($user)
			->icon('cookie')
			->send();

		return true;
	}

	/**
	 * Utility to store this device as a remembered device for this user.
	 */
	public static function RememberDevice(UserModel $user) {
		$ua = $_SERVER['HTTP_USER_AGENT'] ?? '';

		if(!$ua) {
			\Core\set_message('Unable to remember devices that do not provide user agent strings!', 'error');
			return;
		}

		$model = new UserDeviceModel();

		$generator = new \Core\Utilities\PasswordGenerator();
		$generator->minLength = 54;
		$generator->capitals = rand(10, 12);
		$generator->numbers = rand(10, 12);
		$generator->symbols = rand(10, 12);
		$token = $generator->generate();

		$model->user = $user->id;
		$model->hash = sha1($token . substr(SECRET_ENCRYPTION_PASSPHRASE, -6));
		$model->useragent = $ua;
		$model->ip_addr = REMOTE_IP;

		$model->save();
		$model->_setCookie($token);
	}

	/**
	 * Get the current user device is one is set via the cookie, does NOT do the verification checks.
	 *
	 * @return UserDeviceModel|null
	 */
	public static function GetCurrentDevice() : ? UserDeviceModel {
		if(!defined('COOKIE_USER_DEVICE_NAME')) {
			define('COOKIE_USER_DEVICE_NAME', 'userdevice');
		}

		// If there's no device cookie set, also just return.
		if(!filter_has_var( INPUT_COOKIE, COOKIE_USER_DEVICE_NAME)) {
			return null;
		}

		$fullstring = filter_input(INPUT_COOKIE, COOKIE_USER_DEVICE_NAME, FILTER_SANITIZE_URL);

		// This string should match [UUID] : [TOKEN]
		$pos = strpos($fullstring, ':');
		if($pos === false) {
			return null;
		}

		$id = substr($fullstring, 0, $pos);

		return UserDeviceModel::Construct($id);
	}

	public static function UnsetCookie() {
		if(!defined('COOKIE_USER_DEVICE_NAME')) {
			define('COOKIE_USER_DEVICE_NAME', 'userdevice');
		}

		$ssl = SSL_MODE === 'allowed' || SSL_MODE === 'required';

		setcookie(COOKIE_USER_DEVICE_NAME, 'DELETED', 1, '/', '', $ssl);
	}
}